/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.bloom;

import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.StagesInitializedSignal;
import ostore.util.ByteUtils;
import ostore.util.Carp;
import ostore.util.Clock;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.NodeId;
import ostore.util.Pair;
import ostore.util.QuickSerializable;
import ostore.util.SecureHash;
import ostore.util.SHA1Hash;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.tapestry.impl.TapestryReadyMsg;
import java.util.Date;
import java.util.LinkedList;
import java.util.HashSet;
import java.util.Set;
import java.util.Map;
import java.util.HashMap;
import java.util.Vector;
import java.util.Iterator;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.StreamTokenizer;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.Reader;
import soss.core.SimulationShutdownEvent;

/**
 * The probabilistic location stage.
 *
 * @author     Sean C. Rhea
 * @version    $Id: StaticClient.java,v 1.8 2003/02/19 10:32:30 hweather Exp $
 */
public class StaticClient implements EventHandlerIF {

    public static class NextReadEvent implements QueueElementIF {
	public String toString () { return "(NextReadEvent)"; }
    }

    private String class_tag = "StaticClient";
    private static final boolean DEBUG = DebugFlags.BLOOM;

    private void dispatch (QueueElementIF item) {
	final String tag = class_tag + ".dispatch";
	if (DEBUG) Debug.printtagln (tag, "dispatching " + item + ".");
	try {
	    _classifier.dispatch(item);
	} catch (SinkException e) {
	    BUG (class_tag + ".dispatch: could not dispatch " + item);
	}
    }

    private static final void BUG (String msg) {
	Carp.die (msg + "  This is a bug in ostore.bloom.StaticClient.  " +
		"Please let Sean know about it ASAP.");
    }
	
    private StreamTokenizer _lru_tok = null;
    private int _lru_lineno = 0;

    public void init_trace (String read_file) {
	// Get the list of files in each object.

	try {
	    // Now get the LRU order to read the objects in.

	    _lru_lineno = 1;
	    {
		FileInputStream is = new FileInputStream (read_file);
		Reader reader = 
		    new BufferedReader (new InputStreamReader (is));
		_lru_tok = new StreamTokenizer (reader);
	    }
	    _lru_tok.parseNumbers ();
	}
	catch (Exception e) {
	    System.err.println ("Exception caught while reading read file:");
	    e.printStackTrace ();
	    System.exit (1);
	}
    }

    public Pair next_read () {
	try {
	    _lru_tok.nextToken ();
	    while (_lru_tok.ttype == StreamTokenizer.TT_EOL) {
		_lru_tok.nextToken ();
		_lru_lineno++;
	    }

	    if (_lru_tok.ttype == StreamTokenizer.TT_EOF)
		return null;

	    if (_lru_tok.ttype != StreamTokenizer.TT_NUMBER) {
		System.err.println ("Expected object on line " + 
			_lru_lineno + ".");
		System.exit (1); 
	    }
	    int obj = (int) _lru_tok.nval;
	    long time = 1000;

	    return new Pair (new Integer (obj), new Long (time));
	}
	catch (Exception e) {
	    System.err.println ("Exception caught while reading SURGE files:");
	    e.printStackTrace ();
	    System.exit (1);
	}
	return null;
    }

    public void init (ConfigDataIF config) throws Exception {
	final String tag = class_tag + ".init";

	if (DEBUG) Debug.turnOn (class_tag);

	ostore.util.Types.registerAllTypes ();

	_pending_reads = new HashSet ();

	_self_node_id = new NodeId (config.getString ("node_id"));
	class_tag = class_tag + ".[" + _self_node_id.port () + "]";

	_classifier = Classifier.getClassifier (_self_node_id);
	_sink = config.getStage().getSink();

	String readfile = config.getString ("read-file");
	init_trace (readfile);

	String [] event_types = {
	    "seda.sandStorm.api.StagesInitializedSignal",
	    "ostore.tapestry.impl.TapestryReadyMsg",
	    "soss.core.SimulationShutdownEvent",
	    "ostore.bloom.StaticClient$NextReadEvent",
	};

	for (int i = 0; i < event_types.length; ++i) {
	    Filter filter = new Filter ();
	    if (! filter.requireType (Class.forName (event_types [i])))
		BUG (tag + ": could not require type " + event_types [i]);
	    _classifier.subscribe (filter, _sink);
	}

	String [] msg_types = {
	    "ostore.bloom.SimulatedReadResp"
	};

	for (int i = 0; i < msg_types.length; ++i) {
	    Filter filter = new Filter ();
	    if (! filter.requireType (Class.forName (msg_types [i])))
		BUG (tag + ": could not require type " + msg_types [i]);
	    if (! filter.requireValue ("inbound", new Boolean (true)))
		BUG (tag + ": could not require inbound = true for " 
			+ msg_types [i]);
	    _classifier.subscribe (filter, _sink);
	}

	dispatch_next_read (30000);
    }

    public void destroy() throws Exception {
	// TODO
    }

    public void handleEvent(QueueElementIF item) 
    throws EventHandlerException {
	final String tag = class_tag + ".handleEvent";

	if (DEBUG) Debug.printtagln (tag, "got " + item + ".");

	if (item instanceof StagesInitializedSignal) {
	    handle_init ((StagesInitializedSignal) item);
	}
	else if (item instanceof TapestryReadyMsg) {
	    handle_tapestry_ready ((TapestryReadyMsg) item);
	}
	else if (item instanceof NextReadEvent) {
	    handle_next_read_event ((NextReadEvent) item);
	}
	else if (item instanceof SimulatedReadResp) {
	    handle_simulated_read_resp ((SimulatedReadResp) item);
	}
	else if (item instanceof SimulationShutdownEvent) {
	    handle_simulation_shutdown_event ();
	}
	else {
	    BUG (tag + ": caught unexpected event " + item + ".");
	}
    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {
	final String tag = class_tag + ".handleEvents";
	if (DEBUG) Debug.printtagln (tag, "called with " + items.length + 
		" events");

	for(int i=0; i<items.length; i++)
	    handleEvent(items[i]);
    }

    public void handle_init (StagesInitializedSignal signal) {
	final String tag = class_tag + ".handle_init";
    }

    public void handle_tapestry_ready (TapestryReadyMsg msg) {
	final String tag = class_tag + ".handle_init";
	_self_guid = msg.node_guid;
	_replica_stage = SimulatedReplica.instance (_self_node_id);
	_simulation_log = SimulationLog.instance (_self_node_id);
    }

    private static Map /* Integer -> SecureHash */ 
	_file_guids = new HashMap ();

    public void handle_next_read_event (NextReadEvent event) {
	final String tag = class_tag + ".handle_next_read_event";

	check (_next_read != null);
	check (_pending_reads.isEmpty ());

	// First put the reads we're going to do in _pending_reads.

	LinkedList files = new LinkedList ();
	files.addLast (_next_read);
	for (Iterator i = files.iterator (); i.hasNext (); ) {
	    
	    Integer file = (Integer) i.next ();
	    SecureHash guid = (SecureHash) _file_guids.get (file);
	    if (guid == null) {
		guid = new SHA1Hash ("file-" + file);
		_file_guids.put (file, guid);
	    }
	    _pending_reads.add (guid);

	    SimulationLog.ReadTransactionInfo xact = 
		_simulation_log.new_xact (_self_node_id, _self_guid, guid);

	    xact.location_start_time = current_time ().getTime ();
	    xact.ideal_location_distance = 
		_replica_stage.closest_replica_lat_ms (guid);
	    xact.total_ideal_lat = 
		_replica_stage.closest_replica_xfer_ms (guid);
	}

	// Then dispatch them.  This funny order is because the reads could
	// be satisfied locally, in which case the dispatch below goes all
	// the way through to handle_simulated_read_resp.  If we dispatch
	// as we add to _pending_reads, then, after the first dispatch,
	// _pending_reads will be empty again if the request is satisfied
	// locally.

	LinkedList reads = new LinkedList (_pending_reads);

	while (! reads.isEmpty ()) {

	    SecureHash guid = (SecureHash) reads.removeFirst ();

	    if (DEBUG) Debug.printtagln (tag, 
		    "starting read for " + guid + ".");

	    dispatch (new SimulatedRead (guid));
	}
    }

    public void handle_simulated_read_resp (SimulatedReadResp resp) {
	final String tag = class_tag + ".handle_simulated_read_resp";
	check (_pending_reads.remove (resp.guid));
	if (DEBUG) Debug.printtagln (tag, "finished read for " + 
		resp.guid + ".");

	SimulationLog.ReadTransactionInfo xact = 
	    _simulation_log.xact (_self_node_id, resp.guid);
	
	xact.routing_finish_time = current_time ().getTime ();
	_simulation_log.xact_finished (_self_node_id, resp.guid);

	if (_pending_reads.isEmpty ())
	    dispatch_next_read (0);
    }

    public void handle_simulation_shutdown_event () {
	if (! _pending_reads.isEmpty ()) {
	    System.err.print ("Node " + _self_node_id.port () + 
		    " received shutdown with the following reads " +
		    "still pending:");
	    for (Iterator i = _pending_reads.iterator (); i.hasNext (); ) {
		SecureHash guid = (SecureHash) i.next ();
		System.err.print (" " + guid);
	    }
	    System.err.println (".");
	}
    }

    private void dispatch_next_read (long extra_delay) {
	final String tag = class_tag + ".dispatch_next_read";

	Pair p = next_read ();
	if (p == null) {
	    if (DEBUG) Debug.printtagln (tag, 
		    "successfully finished all reads.");
	    return;
	}

	Long time = (Long) p.second;
	_next_read = (Integer) p.first;

	if (DEBUG) Debug.printtagln (tag, "dispatching NextReadEvent " +
		time + " ms in the future.");

	_classifier.dispatch_later (new NextReadEvent (), 
		time.longValue () + extra_delay); 
    }

    private Date current_time () {
	return Clock.current_date (_self_node_id);
    }

    private static void check (boolean expr) {
	if (DEBUG && !expr) Carp.die( "BUG");
    }

    private SinkIF _sink;
    private NodeId _self_node_id;
    private SecureHash _self_guid;
    private Classifier _classifier;

    private Integer _next_read;
    private Set _pending_reads;
    private SimulatedReplica _replica_stage;
    private SimulationLog _simulation_log;
}


