/**
 * Resender.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.dispatch;

import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.NodeId;
import ostore.util.Clock;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkFullException;
import java.util.Iterator;
import java.util.TreeSet;
import java.util.HashSet;

/**
 * A utility class that automatically re-dispatches events.
 * <p>
 * If an attempt to <code>dispatch</code> an event throws a
 * <code>SinkFullException</code>, the <code>Resender</code> schedules the
 * event for a later attempt and returns a unique, opaque token.  That
 * token can be used later to confirm or cancel delivery.
 * <p>
 * <b>NOTE:</b> A <code>Resender</code> does not guarantee any delivery
 * order for events which are initially delayed.  If in-order delivery is
 * important, do not use the automatic re-dispatching features, or avoid
 * this class completely.
 * <p>
 * Any stage which uses a <code>Resender</code> must
 * {@link Signal#register} for {@link Signal#ALRM} signals targetted at the
 * name given to the <code>Resender</code> constructor.  The
 * <code>Resender</code> will set its own alarms; the stage must catch the
 * resulting <code>AlarmSignal</code>s and call {@link #process_queues}.
 *
 * @author	Dennis Geels
 * @version $Id: Resender.java,v 1.9 2003/02/07 00:02:38 geels Exp $
 */

public class Resender
{
    ////////////////////       CONSTANTS	////////////////////

    private static final String class_tag = "Resender";
    private static final boolean DEBUG = DebugFlags.DISPATCH_RESENDER;
    private static final boolean DEBUG_RESENDS = true;
    
    ////////////////////    STATIC VARIABLES 	////////////////////

    /**
     *  A monotonically increasing sequence number for tokens.
     */
    private static int seq_counter;
	
    /**
     * A synchronization token.
     */
    private static Object counter_lock;
    
    ////////////////////       VARIABLES	////////////////////

    /**
     * The set of all events to retry at a later date, sorted by that date.
     */
    private TreeSet pending_events;

    /**
     * The set of all events which have expired.  They were retried their
     * maximum number of times (at least once) and never dispatched
     * successfully.
     */
    private HashSet waste_bin;

    /**
     * The <code>NodeId</code> for this machine.
     */
    private NodeId node_id;

    /**
     * The name with which to target alarm signals.
     */
    private String stage_name;
    
    /**
     * The underlying <code>Classifier</code> object through which to
     * dispatch events.
     */
    private Classifier classifier;

    /**
     * The <code>Signal</code> object for this machine.
     */
    private Signal signal;

    /**
     * The earliest deadline of any pending event, or
     * <code>Long.MAX_VALUE</code> if none pending.
     */
    private long earliest_deadline;

    /**
     * The token for the <code>AlarmSignal</code> we expect to arrive at
     * <code>earliest_deadline</code>. 
     */
    private Object alarm_token;

    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {
	if( DEBUG ) Debug.turnOn( class_tag );

	seq_counter = 0;
	counter_lock = new Object();
    }
    
    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Creates a new <code>Resender</code> object that uses the
     * default <code>Classifier</code> for the specified machine.
     *
     * @param node_id	The ID for this machine - needed to locate the
     * correct <code>Classifier</code> and <code>Signal</code>.
     * @param stage_name	The name of the calling stage - needed to
     * address <code>AlarmSignal</code>s properly.
     */
    public Resender( NodeId node_id, String stage_name )
    {
	this.pending_events = new TreeSet();
	this.waste_bin = new HashSet();
	this.node_id = node_id;
	this.stage_name = stage_name;
	this.classifier = Classifier.getClassifier( node_id );
	this.signal = Signal.getSignal( node_id );
	this.earliest_deadline = Long.MAX_VALUE;
	this.alarm_token = null;
    }

    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Processes the queue of pending events, sending any whose deadlines
     * have been reaced.
     * <p>
     * This method should be called each time the stage receives an
     * <code>AlarmSignal</code>. 
     */
    public synchronized void process_queues()
    {
	final String tag = class_tag+".process_queues";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	// Scan the head of the queue until we reach an event which isn't
	// ready yet.
	long now = Clock.current_date( node_id ).getTime();
	if( DEBUG ) Debug.printtagln( tag, "now: " + now + " pending: " +
					  pending_events.size() );

	if( now >= earliest_deadline ) {
	    while( ! pending_events.isEmpty() ) {
		EventState next = (EventState)pending_events.first();

		if( DEBUG ) Debug.printtagln( tag, "next: " + next.deadline );
		if( next.deadline > now ) {
		    // No more events ready yet - all done.
		    earliest_deadline = next.deadline;
		    reset_alarm();
		    break;
		} else {
		    if( DEBUG_RESENDS ) {
			Debug.printtagln( tag, "processing: " + next.event );
		    }
		    pending_events.remove( next );
		    // Try to dispatch the event
		    boolean resend = true;
		    if( next.resend == EventState.NEVER ) {
			if( DEBUG ) Debug.printtagln( tag, "not resending" );
		    } else {
			try {
			    resend = false;	// Assume it will work.
			    classifier.dispatch( next.event );
			    if( DEBUG )
				Debug.printtagln( tag, "dispatch succeeded" );
			} catch( SinkFullException sfe ) {
			    if( DEBUG )
				Debug.printtagln( tag, "dispatch failed" );
			    resend = true;
			}
		    }
		    if( resend || (next.resend == EventState.ALWAYS)) {
			reschedule( next );
		    }
		}
	    }
	    if( pending_events.isEmpty() ) {
		earliest_deadline = Long.MAX_VALUE;
	    }
	} else {
	    if( DEBUG ) Debug.printtagln( tag, "not ready" );
	    reset_alarm();
	}
	
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }

    /**
     * Sets an alarm to wake us before the next event must be sent,
     * clearing any old alarm to save resources. 
     */
    private void reset_alarm()
    {
	final String tag = class_tag+".reset_alarm";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	if( alarm_token != null ) {
	    signal.cancel( alarm_token );
	}
	if( DEBUG ) Debug.printtagln( tag, "earliest_deadline: " +
				      earliest_deadline );
	if( earliest_deadline < Long.MAX_VALUE ) {
	    long now = Clock.current_date( node_id ).getTime();
	    alarm_token = signal.alarm( stage_name, (earliest_deadline-now) );
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }
    
    /**
     * Dispatches an event through the <code>Classifier</code>.
     * <p>
     * If the dispatch fails, the event is stored in a trash pile.  Events
     * on the trash pile can be recovered by the {@link #clear} method.
     *
     * @param event	Any event to dispatch
     *
     * @return a token, used to reference the event later, or
     * <code>null</code> if the object was successfully dispatched.
     *
     * @see #dispatch(seda.sandStorm.api.QueueElementIF, int, long)
     * @see Classifier#dispatch
     */
    public synchronized Object dispatch( QueueElementIF event )
    {
	// Do not retry the dispatch.
	return dispatch( event, 0, 0 );
    }
    
    /**
     * Dispatch the specified event, making at most
     * <code>num_retries</code> additional attempts, waiting
     * <code>millis</code> milliseconds in between attempts.
     * <p>
     * After enough attempts have failed, the event is stored in a trash
     * pile.  Events on the trash pile can be recovered by the
     * {@link #clear} method.
     *
     * @param event	Any event to dispatch
     * @param num_retries	The number (<code>>=0) of times to try
     * dispatching the event if the initial attempt fails.
     * @param millis	The number of milliseconds to wait after each
     * attempt.
     *
     * @return a token, used to reference the event later, or
     * <code>null</code> if the object was successfully dispatched on the
     * first try.
     *
     * @see #dispatch(seda.sandStorm.api.QueueElementIF)
     * @see #resend(seda.sandStorm.api.QueueElementIF, int, long)
     * @see Classifier#dispatch
     */
    public synchronized Object dispatch( QueueElementIF event,
					 int num_retries, long millis )
    {
	final String tag = class_tag+".dispatch";
	if( DEBUG ) Debug.printtagln( tag, "called: " + event + "(" +
				      num_retries + ")(" + millis + ")" );
	EventState token;
	// Call the real method.
	try {
	    classifier.dispatch( event );
	    if( DEBUG ) Debug.printtagln( tag, "dispatch succeeded" );	    
	    token = null;	// <code>null</code> means successful
	} catch( SinkFullException sfe ) {
	    if( DEBUG ) Debug.printtagln( tag, "dispatch failed" );
	    // Dispatch failed.  Store the failed event in an
	    // <code>EventState</code> object, reschedule another try,
	    // and pass the state back to the caller as the event's unique
	    // token. 
	    token = new EventState( event, num_retries, millis,
				    EventState.ON_FAIL /* send only once */);
	    reschedule( token );
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
	return token;
    }

    /**
     * Schedules an event for another attempt.
     * If all attempts have failed, the event is stored in a trash
     * pile.  Events on the trash pile can be recovered by the
     * {@link #clear} method.
     *
     * @param state		The <code>EventState</code> object that 
     * 				stores the event and retry parameters.
     * @see #clear
     */
    private void reschedule( EventState state )
    {
	final String tag = class_tag+".reschedule";
	if( DEBUG )
	    Debug.printtagln( tag, "called: " + state.event + "(" +
			      state.num_retries + ")(" + state.delay + ")" );
	
	if( state.num_retries > 0 ) {
	    long now = Clock.current_date( node_id ).getTime();
	    long next_time = now + state.delay;
	    if( DEBUG ) Debug.printtagln( tag, "rescheduled for " +
					  next_time );
	    if( next_time < earliest_deadline ) {
		earliest_deadline = next_time;
		reset_alarm();
	    }
	    state.num_retries -= 1;
	    state.deadline = next_time;
	    pending_events.add( state );
	} else {
	    // Don't worry about the other state fields.
	    if( DEBUG ) Debug.printtagln( tag, "throwing away" );
	    waste_bin.add( state );
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }

    /**
     * Repeatedly dispatches the specified event, <code>num_retries</code>
     * times, waiting <code>millis</code> milliseconds in between attempts.
     * <p>
     * After enough attempts have failed, the event is stored in a trash
     * pile.  Events on the trash pile can be recovered by the
     * {@link #clear} method.
     * <p>
     * Unlike {@link #dispatch(seda.sandStorm.api.QueueElementIF, int, long)},
     * this method resends the event even if it dispatches correctly.
     * It is therefore slightly more expensive and should be used only when
     * these semantics are required.
     *
     * @param event		Any event to dispatch
     * @param num_retries	The number (<code>>=0) of times to try
     * dispatching the event if the initial attempt fails.
     * @param millis		The number of milliseconds to wait after
     * 				each attempt.
     *
     * @return a token, used to reference the event later.
     *
     * @see #dispatch(seda.sandStorm.api.QueueElementIF, int, long)     
     * @see #cancel(Object)     
     * @see Classifier#dispatch
     */
    public synchronized Object resend( QueueElementIF event,
				       int num_retries, long millis )
    {
	final String tag = class_tag+".resend";
	if( DEBUG ) Debug.printtagln( tag, "called: " + event + "(" +
				      num_retries + ")(" + millis + ")" );
	EventState token;
	// Call the real method.
	classifier.dispatch_lossy( event );
	// Store the event in an <code>EventState</code> object,
	// reschedule another try, and pass the state back to the caller as
	// the event's unique token.
	token = new EventState( event, num_retries, millis,
				EventState.ALWAYS /* always resend */ );
	reschedule( token );

	if( DEBUG ) Debug.printtagln( tag, "done" );
	return token;
    }
    
    /**
     * Dispatches the specified event, waits <code>millis</code>
     * milliseconds, then stores the event in the trash pile.
     * Events on the trash pile can be recovered by the {@link #clear}
     * method.
     * <p>
     * Unlike {@link #dispatch(seda.sandStorm.api.QueueElementIF, int, long)},
     * this method remembers the event even if it dispatches correctly.
     * <p>
     * Unlike {@link #resend(seda.sandStorm.api.QueueElementIF, int, long)},
     * this method does not try to resend the event.
     * Use that method instead if the event may be lost (e.g. a
     * <code>NetworkMessage</code>) and duplicate messages are acceptable.
     *
     * @param event		Any event to dispatch
     * @param millis		The number of milliseconds to wait before
     * 				trashing the event.
     *
     * @return a token, used to reference the event later.
     *
     * @see #resend(seda.sandStorm.api.QueueElementIF, int, long)     
     * @see #cancel(Object)     
     * @see Classifier#dispatch
     */
    public synchronized Object send( QueueElementIF event, long millis )
    {
	final String tag = class_tag+".send";
	if( DEBUG ) Debug.printtagln( tag, "called: " + event + "(" +
				      millis + ")" );
	EventState token;
	// Call the real method.
	classifier.dispatch_lossy( event );
	// Store the event in an <code>EventState</code> object,
	// reschedule another try, and pass the state back to the caller as
	// the event's unique token.
	token = new EventState( event, 1/* time out first time */, millis,
				EventState.NEVER /* do not resend */ );
	reschedule( token );

	if( DEBUG ) Debug.printtagln( tag, "done" );
	return token;
    }

    /**
     * Waits <code>millis</code> milliseconds, then stores the event in the
     * trash pile. Events on the trash pile can be recovered by the
     * {@link #clear}  method.
     * <p>
     * Unlike the other methods, this one never tries to send the event.
     * It is provided as a convenient interface to the internal timeout
     * support. 
     *
     * @param event		Any event to remember
     * @param millis		The number of milliseconds to wait before
     * 				trashing the event.
     *
     * @return a token, used to reference the event later.
     *
     * @see #resend(seda.sandStorm.api.QueueElementIF, int, long)     
     * @see #send(seda.sandStorm.api.QueueElementIF, long)     
     * @see #cancel(Object)     
     * @see Classifier#dispatch
     */
    public synchronized Object remember( QueueElementIF event, long millis )
    {
	final String tag = class_tag+".remember";
	if( DEBUG ) Debug.printtagln( tag, "called: " + event + "(" +
				      millis + ")" );
	EventState token;
	// Store the event in an <code>EventState</code> object,
	// reschedule another try, and pass the state back to the caller as
	// the event's unique token.
	token = new EventState( event, 1/* time out first time */, millis,
				EventState.NEVER /* do not resend */ );
	reschedule( token );

	if( DEBUG ) Debug.printtagln( tag, "done" );
	return token;
    }

    /**
     * Cancels the specified event.
     * All state relating to the pending event is removed from the
     * <code>Resender</code>.  If the event has already dispatched
     * successfully, this method does nothing. 
     *
     * @param token	A token returned by {@link #dispatch}.
     *
     * @return the event, if it has not yet been dispatched
     * successfully, otherwise <code>null</code>.
     *
     * @see #dispatch(seda.sandStorm.api.QueueElementIF, int, long)
     */
    public synchronized QueueElementIF cancel( Object token )
    {
	final String tag = class_tag+".cancel";
	if( DEBUG ) Debug.printtagln( tag, "called: " + token );

	QueueElementIF event;
	boolean still_here = pending_events.remove( token );
	if( ! still_here ) {
	    still_here = waste_bin.remove( token );
	}
	if( still_here ) {
	    if( DEBUG ) Debug.printtagln( tag, "removed" );
	    EventState state = (EventState)token;
	    if( earliest_deadline == state.deadline ) {
		// Reschedule the next alarm.
		if( ! pending_events.isEmpty() ) {
		    state = (EventState)pending_events.first();
		    earliest_deadline = state.deadline;
		} else {
		    earliest_deadline = Long.MAX_VALUE;
		}
		reset_alarm();
	    }
	    event = ((EventState)token).event;
	} else {
	    // indicates that the event was dispatched successfully:
	    event = null;
	}
	
	if( DEBUG ) Debug.printtagln( tag, "done" );
	return event;
    }

    /**
     * Checks whether the specified event was been successfully dispatched.
     *
     * @param token	A token returned by {@link #dispatch}.
     *
     * @return <code>true</code> iff the event has timed out.
     *
     * @see #dispatch(seda.sandStorm.api.QueueElementIF, int, long)
     */
    public synchronized boolean timed_out( Object token )
    {
	final String tag = class_tag+".timed_out";
	if( DEBUG ) Debug.printtagln( tag, "called: " + token );

	boolean dead = waste_bin.contains( token );
	
	if( DEBUG ) Debug.printtagln( tag, "done: " + dead );
	return dead;
    }

    /**
     * Checks whether the specified event is scheduled for further attempts.
     *
     * @param token	A token returned by {@link #dispatch}.
     *
     * @return <code>true</code> iff the event is on the queue for another
     * attempted dispatch.
     *
     * @see #dispatch(seda.sandStorm.api.QueueElementIF, int, long)
     */
    public synchronized boolean pending( Object token )
    {
	final String tag = class_tag+".pending";
	if( DEBUG ) Debug.printtagln( tag, "called: " + token );

	boolean waiting = pending_events.contains( token );
	
	if( DEBUG ) Debug.printtagln( tag, "done: " + waiting );
	return waiting;
    }

    /**
     * Checks whether the specified event has been successfully dispatched.
     * <p>
     * Equivalent to <code>( !pending(token) && !timed_out(token) )</code>.
     *
     * @param token	A token returned by {@link #dispatch}.
     *
     * @return <code>true</code> iff the <code>Resender</code> no longer
     * has any memory of the <code>token</code>, which <emph>implies</emph>
     * that the event was previously dispatched.
     *
     * @see #dispatch(seda.sandStorm.api.QueueElementIF, int, long)
     */
    public synchronized boolean was_sent( Object token )
    {
	final String tag = class_tag+".was_sent";
	if( DEBUG ) Debug.printtagln( tag, "called: " + token );

	boolean still_here = ( pending_events.contains( token ) ||
			       waste_bin.contains( token ) );
	if( DEBUG ) {
	    Debug.printtagln( tag, (still_here?"not sent":"sent") );
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );	
	return ( ! still_here );
    }

    /**
     * Cancels all events which have timed out.  If <code>all ==
     * true</code>, also cancels all pending events.
     *
     * @param all	<code>true</code> iff pending events should also be
     * 			cancelled. 
     *
     * @return an array containing all cancelled events.
     *
     * @see #dispatch(seda.sandStorm.api.QueueElementIF, int, long)
     * @see #cancel
     */
    public synchronized QueueElementIF[] clear( boolean all )
    {
	final String tag = class_tag+".was_sent";
	if( DEBUG ) Debug.printtagln( tag, "clear" );

	// Make an array to hold the dead events
	int len = waste_bin.size();
	if( all ) {	// Include the pending events.
	    len += pending_events.size();
	}
	QueueElementIF[] cancelled = new QueueElementIF[len];

	// Dump events from the waste bin into the array.
	int i = 0;
	Iterator states = waste_bin.iterator();
	while( states.hasNext() ) {
	    cancelled[i] = ((EventState)states.next()).event;
	    i++;
	}
	waste_bin.clear();
	if( all ) {
	    // Repeat for pending events.
	    states = pending_events.iterator();
	    while( states.hasNext() ) {
		cancelled[i] = ((EventState)states.next()).event;
		i++;
	    }	    
	    pending_events.clear();

	    // No events pending:
	    earliest_deadline = Long.MAX_VALUE;
	    reset_alarm();
	}

	if( DEBUG ) Debug.printtagln( tag, "done" );
	return cancelled;
    }
       
    
    ////////////////////     INNER CLASSES	////////////////////

    /**
     * An internal object that stores state for an event which did not
     * dispatch properly.
     */
    private class EventState implements Comparable
    {
	////////////////////     CONSTANTS	////////////////////

	// Possible values for resend:
	public static final int NEVER = 0;
	public static final int ON_FAIL = 1;
	public static final int ALWAYS = 2;
	
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////

	/**
	 * The original event.
	 */
	public QueueElementIF event;

	/**
	 * The number of times to try to dispatch <code>event</code> if the
	 * next attempt fails.
	 */
	public int num_retries;
	
	/**
	 * The number of milliseconds to wait between attempts.
	 */
	public long delay;

	/**
	 * Specifies when to resend the event.
	 * One of <code>NEVER</code>, <code>ON_FAIL</code>, or
	 * <code>ALWAYS</code>. 
	 */
	public int resend;
	
	/**
	 * The time at which to try next.
	 */
	public long deadline;

	/**
	 * A unique (up to wrap-around) sequence number for this token
	 */
	private int seq_num;
	
	////////////////////    CONSTRUCTORS	////////////////////

	public EventState( QueueElementIF event, int num_retries, long delay,
			   int resend )
	{
	    this.event = event;
	    this.num_retries = num_retries;
	    this.delay = delay;
	    this.resend = resend;
	    // this.deadline set in <code>reschedule</code>.
	    synchronized( counter_lock ) {
		this.seq_num = seq_counter;
		seq_counter++;
	    }
	}

	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////

	/**
	 * Impose an order on <code>EventState</code>s based on their
	 * <code.deadline</code>s.
	 */
	public int compareTo( Object o )
	{
	    EventState other = (EventState)o;
	    if( deadline < other.deadline ) {
		return -1;
	    } else if( deadline > other.deadline ) {
		return 1;
	    } else {
		if( seq_num < other.seq_num ) {
		    return -1;
		} else if( seq_num > other.seq_num ) {
		    return 1;
		} else {
		    return 0;
		}
	    }
	}
	/**
	 * Returns a human-readable representation of this
	 * <code>EventState</code>
	 */
	public String toString()
	{
	    return ("<Token for: " + event + ">");
	}
	
	////////////////////   INNER CLASSES	////////////////////
    }
}
