/**
 * TSink.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.dispatch;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.lang.reflect.Field;
import ostore.util.Queue;
import ostore.util.Debug;
import ostore.util.Carp;
import seda.sandStorm.main.Sandstorm;
import seda.sandStorm.api.*;
import seda.sandStorm.core.SimpleSink;

/** 
 * TSink is an efficient implementation of a Sink. TSink sends events 
 * directly to a stage's Sink rather than through a Classifier.
 *
 * @author Aqeel Mahesri
 */

public class TSink extends SimpleSink {

    public static final boolean DEBUG = 
	ostore.util.DebugFlags.DISPATCH_TSINK;
    public static final String tag = "TSink";
    
    private static Vector TSinkV;
    private Sandstorm ss;
    private Hashtable knownTypes; // maps types to Boolean.
    // no entry if type never seen before
    // TRUE if entry is subscribed to by another stage
    // FALSE if entry is not subscribed to by another stage
    private Classifier classifier;
    protected SinkIF mySinkIF;

    static {
	//	TSinkV = new Vector();
    }

    /**
     * Constructs a new <code>TSink</code> for the stage given as an argument.
     * When the requesting stage has the <code>TSink</code> for the 
     * given stage, it will be able to pass events directly instead of
     * invoking this <code>Classifier</code>, which should be more 
     * efficient.
     *
     * @param String stagename   Name of the stage whose TSink is requested
     */
    public TSink(String stagename, Classifier classifier) 
	throws NoSuchStageException 
    {
	ss = Sandstorm.getSandstorm();
	knownTypes = new Hashtable();
	StageIF sif = ss.getManager().getStage(stagename);
	mySinkIF = sif.getSink();
	this.classifier = classifier;
    }

    /**
     * Notifies all <code>TSinks</code> that a <code>SinkIF</code> has 
     * subscribed to a certain type of event via the <code>Classifier</code>. 
     * The <code>TSink</code> class stores the event type so that any events 
     * of this type are enqueued via the <code>Classifier</code> and reach 
     * the stages that need to listen to them.
     *
     * There is one significant limitation.  If someone subscribes to a
     * subclass of a message type, then all of the supertypes of the message
     * type should also be marked as subscribed.  This is currently not the 
     * case; if anyone has a TSink receiving messages with a supertype that
     * is subscribed to, the subscribee will not receive the messages.
     *
     * @param sink	The <code>SinkIF</code> to add
     * @param filter	The <code>Filter</code> to use.
     * @param priority	The <code>Filter</code>'s priority
     */
    public void subscribeNotify( Filter filter, SinkIF sink,
				 Integer priority ) {
	// may occur at same time that enqueue is being called
	// Vector and Hashtable are synchronized, so this is ok
	Class required_type = filter.required_type;
	Enumeration enum = knownTypes.keys();
	while (enum.hasMoreElements()) {
	    Class current_type = (Class)enum.nextElement();
	    if (required_type.isAssignableFrom(current_type)) {
		knownTypes.put(current_type, Boolean.TRUE);
	    }
	}
	
    }

    /**
     * Notifies all <code>TSinks</code> that a <code>SinkIF</code> has been
     * unsubscribed from the <code>Classifier</code>.
     *
     * @param sink  The <code>SinkIF</code> that is being unsubscribed
     */
    public static void unsubscribeNotify( SinkIF sink )
    {
	Carp.error( "This method not yet implemented." );
    }


    /**
     * Enqueues the event in the <code>SinkIF</code> corresponding to this 
     * <code>TSink</code>'s <code>VQueue</code>. It also checks whether any 
     * stages have subscribed to this type of event via the 
     * <code>Classifier</code>; in this case, the event must also be 
     * dispatched by the <code>Classifier</code>.
     *
     * @param event  The event which is to be sent to a queue
     */
    public void enqueue( QueueElementIF event ) throws SinkException {
	Boolean b;
	mySinkIF.enqueue(event);
	if ((b = (Boolean)knownTypes.get(event.getClass())) != null) {
	    if (b.booleanValue()) {
		classifier.dispatch(event);
	    }
	} else {
	    if (DEBUG) Debug.printtagln( tag+".enqueue",
					 "new type " + event.getClass());
	    // rarely happens, so it can run slowly
	    // find out if any sinks are subscribed to this type of event
	    b = classifier.addTSinkType(this, event.getClass());
	    knownTypes.put(event.getClass(), b);
	    if (b.booleanValue()) {
		classifier.dispatch(event);
	    }
	}
    }

    /**
     * Enqueues the event without doing the checks.
     *
     * @param event  The event which is to be sent to a queue
     */
    public void enqueueWOChecks( QueueElementIF event ) throws SinkException {
	mySinkIF.enqueue(event);
    }


}
