/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.monitor;

import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.StagesInitializedSignal;
import seda.sandStorm.lib.aDisk.*;
import seda.sandStorm.core.BufferElement;
import ostore.util.*;
import java.util.*;

/**
 * Test for managed resources.
 *
 * @author  Jeff Pang
 * @version $Id: ManagedTestStage.java,v 1.2 2003/02/27 22:52:01 jeffpang Exp $
 **/
public class ManagedTestStage extends StandardStage
{
    public static final String class_tag = "ManagedTestStage";
    public static final boolean DEBUG = true;

    // Super:
    // protected Classifier classifier;
    // protected NodeId my_node_id;
    // protected SinkIF my_sink;
    // protected String tag;
    // protected String [] event_types = {};
    // protected String [] inb_msg_types = {};
    // protected String [] outb_msg_types = {};

    // ------------------------------------------------------------------------
    // Config
    // ------------------------------------------------------------------------

    // ------------------------------------------------------------------------
    // State
    // ------------------------------------------------------------------------

    /**
     * Constructor
     */
    public ManagedTestStage()
    {
	// events we listen to
	event_types = new String[] {
	    "ostore.monitor.ResourceMonitorReadySignal",
	    "ostore.monitor.RegisterResourceResp",
	    "ostore.monitor.RequestResourceResp",
	    "ostore.monitor.ManagedTestStage$Continue",
	};
    }

    // Super:
    // protected final void BUG (String msg);
    // protected void dispatch (QueueElementIF item);

    /**
     * Sandstorm implementation.
     */
    public void init(ConfigDataIF config) throws Exception 
    {
	final String tag = class_tag+".init";
	if( DEBUG ) Debug.printtagln( tag, "called");
    
	super.init(config);
    }

    //-------------------------------------------------------------------------
    // Stuff
    //-------------------------------------------------------------------------

    public static class Continue implements QueueElementIF
    {
	public int val;
	public Continue(int val) { this.val = val; }
    }

    private static class TestResource extends Resource
    {
	public TestResource(NodeId id, int index)
	{
	    super(id, "TestResource" + index, 0);
	}

	public void log(int val)
	{
	    log((double)val);
	}
    }

    private TestResource rsrc1, rsrc2;
    private boolean phase1_1, phase1_2;

    /**
     * Sandstorm implementation.
     */
    public void handleEvent(QueueElementIF element) 
    {
	final String tag = class_tag+".handleEvent";
	if( DEBUG ) Debug.printtagln( tag, "called: " + element);

	// initialize
	if (element instanceof ResourceMonitorReadySignal) {
	    
	    // (1)
	    rsrc1 = new TestResource(my_node_id, 1);
	    rsrc2 = new TestResource(my_node_id, 2);
	    if (DEBUG) Debug.printtagln(tag, "(1) Registering TestResource1");
	    dispatch(new RegisterResourceReq(rsrc1));
	    return;
	}
	
	// main loop
	if (element instanceof RegisterResourceResp) {
	    RegisterResourceResp resp = (RegisterResourceResp)element;
	    if (resp.getResource() == rsrc1) {

		if (!resp.success()) {
		    Carp.die(tag + ": Register TestResource1 failed");
		}

		// (2)
		if (DEBUG) Debug.printtagln(tag, "(2) Requesting TestResource1");
		dispatch(new RequestResourceReq(rsrc1.getName(), false));
	    } else {

		// (5/6)
		if (DEBUG) Debug.printtagln(tag, "(5/6) Registered TestResource2");
		phase1_1 = true;
		phase2();
	    }
	} else if (element instanceof RequestResourceResp) {
	    RequestResourceResp resp = (RequestResourceResp)element;
	    if (resp.getName().equals(rsrc1.getName())) {
		
		if (!resp.success()) {
		    Carp.die(tag + ": Request TestResource1 failed");
		}

		// (3)
		if (DEBUG) Debug.printtagln(tag, "(3) Requesting blocked " +
					    "TestResource2");
		dispatch(new RequestResourceReq(rsrc2.getName(), true));
		classifier.dispatch_later(new Continue(1), 2000);
	    } else {
		
		// (5/6)
		if (DEBUG) Debug.printtagln(tag, "(5/6) Obtained TestResource2");
		phase1_2 = true;
		phase2();
	    }
	} else if (element instanceof Continue) {
	    Continue c = (Continue)element;
	    if (c.val == 1) {
		
		// (4)
		if (DEBUG) Debug.printtagln(tag, "(4) Registering TestResource2");
		dispatch(new RegisterResourceReq(rsrc2));
	    } else {
		phase2(c.val);
	    }
	} else {
	    Carp.die(tag + ": unknown element type: " + element);
	}
    }

    private void phase2() {
	final String tag = class_tag+".phase2";
	if( DEBUG ) Debug.printtagln( tag, "called");

	if (!phase1_1 || !phase1_2) {
	    return;
	}

	rsrc1.log(1);
	rsrc2.log(1);
	classifier.dispatch_later(new Continue(2), 1000);
    }

    private void phase2(int val)
    {
	switch(val) {
	case 2: {
	    if (rsrc1.getLog().util(0) != 1 || rsrc2.getLog().util(0) != 1) {
		BUG("Bad value");
	    }

	    for (int i=2; i<1000; i++) {
		rsrc1.log(i);
		rsrc2.log(i);
	    }
	    classifier.dispatch_later(new Continue(3), 10000);
	    return;
	}
	case 3: {
	    if (rsrc1.getLog().util(0) != 999 || 
		rsrc2.getLog().util(0) != 999) {
		BUG("Bad value");
	    }

	    rsrc1.log(1000);
	    rsrc2.log(1000);
	    if (rsrc1.getLog().util(0) != 1000 || 
		rsrc2.getLog().util(0) != 1000) {
		BUG("Bad value");
	    }
	    classifier.dispatch_later(new Continue(4), 10000);
	    return;
	}
	case 4: {
	    if (rsrc1.getLog().util(0) != 1000 || 
		rsrc2.getLog().util(0) != 1000) {
		BUG("Bad value");
	    }
	    if (DEBUG) Debug.printtagln(tag, "(7) Done.");
	    System.exit(0);
	}
	}

    }
}
