/**
 * Network.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.network;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StreamTokenizer;
import java.net.InetAddress;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Map;
import java.util.HashSet;
import java.util.Set;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.SingleThreadedEventHandlerIF;
import seda.sandStorm.api.NoSuchStageException;
import seda.sandStorm.api.NoSuchStageException;
import seda.sandStorm.api.ProfilableIF;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkClosedEvent;
import seda.sandStorm.api.SinkClosedException;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkFlushedEvent;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.core.BufferElement;
import seda.sandStorm.internal.SinkProxy;
import seda.sandStorm.lib.aSocket.ATcpClientSocket;
import seda.sandStorm.lib.aSocket.ATcpConnectFailedEvent;
import seda.sandStorm.lib.aSocket.ATcpConnection;
import seda.sandStorm.lib.aSocket.ATcpInPacket;
import seda.sandStorm.lib.aSocket.ATcpListenFailedEvent;
import seda.sandStorm.lib.aSocket.ATcpListenSuccessEvent;
import seda.sandStorm.lib.aSocket.ATcpServerSocket;
import seda.sandStorm.lib.aSocket.aSocketInputStream;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.dispatch.Signal;
import ostore.util.QSException;
import ostore.util.AssertionViolatedException;
import ostore.util.ByteUtils;
import ostore.util.Carp;
import ostore.util.CountBuffer;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.Pair;
import ostore.util.PriorityQueue;
import ostore.util.Queue;
import ostore.util.QuickSerializable;
import ostore.util.QuickSerializable;
import ostore.util.TypeTable;
import soss.network.TransitStub;

/**
 * A stage to tie the OceanStore dispatch model to Sandstorms asyncronous
 * networking functionality.  To send a message on the network, inherit it
 * from NetworkMessage, set inbound = false, and dispatch it.  This stage
 * will catch the event, open a connection (if one is not open already),
 * and send it.  To receive messages, just listen for the correct type
 * conditioned on inbound == true.
 *
 * @author     Sean C. Rhea
 * @version    $Id: Network.java,v 1.80 2003/05/20 01:32:31 hweather Exp $
 */
public class Network 
implements SingleThreadedEventHandlerIF, EventHandlerIF, ProfilableIF {

    private String class_tag = "Network";
    public static final boolean DEBUG = false;
    public static final boolean DEBUG_SERIALIZATION = false;

    /**
     * Attempt to reconnect on ConnectFailureEvent.
     * Note: this will attempt to reconnect continuously until
     * entire JVM process is terminated. */
    private static final boolean INFINITE_CONNECT_RETRY = false;

    private void dispatch (QueueElementIF item) {
	final String tag = class_tag + ".dispatch";
	if (DEBUG) Debug.printtagln (tag, "dispatching " +
		item.getClass () + " event: " + item + ".");
	try {
	    _classifier.dispatch(item);
	} catch (SinkException e) {
	    BUG (class_tag + ".dispatch: could not dispatch " + item +
		 " exception=" + e);
	}
    }

    private static final void BUG (String msg) {
	Carp.die (msg + "  This is a bug in ostore.network.Network.  " +
		"Please let Sean know about it ASAP.");
    }

    private void handle_alarm () {
	final String tag = class_tag + ".handle_alarm";

	if (! _close_me.empty ()) {
	    long now = System.currentTimeMillis ();
	    long top = _close_me.topPriority (false);

	    while (now > top) {
		ATcpConnection conn_to_close = 
		    (ATcpConnection) _close_me.extract (false);

		try {
		    conn_to_close.close (_this_sink);
		}
		catch (SinkClosedException e) {
		    if (DEBUG) Debug.printtagln (tag,
			    "Tried to close already closed "
			    + "client connection.");
		}

		if (_close_me.empty ())
		    break;
		else
		    top = _close_me.topPriority (false);
	    }
	}
	Signal.getSignal (_this_node_id).alarm ("NetworkStage", 10000);
    }

    private void handle_atcp_connection (ATcpConnection connection) {
	final String tag = class_tag + ".handle_atcp_connection";

	NodeId peer = new MyNodeId (connection.getPort (),
		connection.getAddress ());

	if (DEBUG) Debug.printtagln (tag, 
		"Network got ATcpConnection event from " + peer + ".");

	if (_pending.get (peer) != null) {

	    if (DEBUG) Debug.printtagln (tag, 
		    "This connection was established locally so"
		    + " the port number is correct.");

	    // We initiated this connection, so the port number is
	    // correct, but we need to send an IdentityMessage to
	    // let the other side know our server port before we
	    // send any of the pending messages.  We will complete
	    // this when we get the SinkFlushedEvent later.

	    IdentityMessage ident = 
		new IdentityMessage (peer, _this_node_id);
	    ident.sender = _this_node_id;
	    send_message (ident, connection, peer);

	    try {
		connection.flush (_this_sink);
	    }
	    catch (SinkClosedException e) {
		BUG (tag + ": Got SinkClosedException on flush.  "
			+ "(This is bad.)"); 
	    }
	}
	else {

	    if (DEBUG) Debug.printtagln (tag, 
		    "This connection was established remotely so"
		    + " we don't know the server port yet.");

	    // We did not initiate this connection, so the port
	    // number is (possibly) different from the port number
	    // the sender listens on.  We'll hang onto this
	    // connection in anticipation of a message mapping it
	    // to the listening port.  We also store the current
	    // time, so we can timeout if that message never comes.

	    Pair p = new Pair ();
	    p.first = connection;
	    p.second = new Long (System.currentTimeMillis ());

	    _mystery_connections.put (peer, p);
	}

	// Start allowing incoming packets from this connection.

	connection.startReader (_this_sink);
    }

    public void handle_sink_flushed_event (SinkFlushedEvent flush) {
	final String tag = class_tag + ".handle_sink_flushed_event";

	ATcpConnection conn = null;

	try {
	    conn = (ATcpConnection) (flush.sink);
	}
	catch (ClassCastException e) {
	    BUG (tag + ": Got SinkFlushedEvent for a sink which "
		    + "was not an ATcpConnection.");
	}

	NodeId peer = new MyNodeId (conn.getPort (), conn.getAddress ());

	if (DEBUG) Debug.printtagln (tag, 
		"Locally established connection to " + peer 
		+ " flushed.  Sending pending messages.");

	// First off, we need to cache this connection so further
	// outgoing messages can be enqueued on it.  In the process, we
	// check for pending outgoing messages.

	Queue pending = (Queue) _pending.get (peer);
	if (pending == null) {
	    BUG (tag + ": No pending messages!  Confused.");
	}

	// We initiated this connection, so the port number is
	// correct, and we can just go ahead and add it to our
	// list of connections.
	if (_connections.containsKey (peer))
	    throw new AssertionViolatedException (
		    tag + ": connection already exists.");
	if (DEBUG) Debug.printtagln (tag,
		"Adding connection for " + peer + ".");
	_connections.put (peer, conn);

	// Moreover, we need to remove the list of pending
	// messages and send them out.  We remove the list here
	// and send them out below.
	_pending.remove (peer);

	send_pending_messages (pending, conn);
    }

    private void handle_atcp_inpacket (ATcpInPacket packet) {
	final String tag = class_tag + ".handle_atcp_inpacket";

	ATcpConnection conn = packet.getConnection ();
	NodeId peer = new MyNodeId (conn.getPort (), conn.getAddress ());
	ConnectionStream is = 
	    (ConnectionStream) _connection_streams.get (peer);
	if (is == null) {
	    is = new ConnectionStream ();
	    _connection_streams.put (peer, is);
	}

	if (DEBUG) Debug.printtagln (tag, "adding packet " + packet + 
		", seq " + packet.getSequenceNumber ());
	is.add_packet (packet);

	while (is.message_available ()) {

	    int inbound_size = bw_counting ? is.next_message_size () : 0;

	    InputBuffer ib = is.get_next_message_buffer ();
	    if (DEBUG_SERIALIZATION) 
		System.err.println (class_tag + ": " + ib);

	    NetworkMessage msg = null;   
	    try {
		msg = (NetworkMessage) TypeTable.new_object (ib);
	    }
	    catch (Exception e) {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS) {
		    Carp.out (tag + ": " +
			    "Received message from " + peer + 
			    " that caused exception " + e + 
			    " on decode.  Dropping it.  " + 
			    "(Stack trace follows.)"); 
		    e.printStackTrace ();
		}
		return;
	    }

	    if (bw_counting) {
		// the data sent was actually preceeded by four bytes
		// representing the size.
		record_bytes (inbound_bw, peer, inbound_size + 4, msg);
	    }

	    if (_client_port_to_server_port.containsKey (peer)) {
		NodeId server_port = 
		    (NodeId) _client_port_to_server_port.get (peer);
		if (DEBUG) Debug.printtagln (tag, "Translating " + 
			peer + " to " + server_port + ".");
		peer = server_port;
	    }

	    if (DEBUG) Debug.printtagln (tag,
		    "Received a message from " + peer + ".");

	    msg.peer = peer;
	    msg.inbound = true;

	    if (DEBUG) Debug.printtagln (tag, "Message is " + msg + ".");

	    if (msg instanceof IdentityMessage)
		process_identity_message ((IdentityMessage) msg);
	    else 
		dispatch (msg);

	    if (DEBUG) Debug.printtagln (tag,
		    "Message correctly decoded and dispatched.");
	}
    }

    protected static class MsgsBytes {
	MsgsBytes (long m, long b) { msgs = m; bytes = b; }
	long msgs;
	long bytes;
    }

    protected void record_bytes (Map map, NodeId peer, int sz, Object msg) {
	String clazz = msg.getClass ().getName ();
	Map class_to_byte_count = (Map) map.get (peer);
	if (class_to_byte_count == null) { 
	    class_to_byte_count = new HashMap (); 
	    map.put (peer, class_to_byte_count);
	}
	MsgsBytes value = (MsgsBytes) class_to_byte_count.get (clazz);
	if (value == null)
	    class_to_byte_count.put (clazz, new MsgsBytes (1, sz));
	else {
	    ++value.msgs;
	    value.bytes += sz;
	}
    }

    private class DelayWrapper implements QueueElementIF {
	public NetworkMessage outb;
	public DelayWrapper (NetworkMessage outb) {
	    this.outb = outb;
	}
    }

    private LinkedList nat_hack_queue = new LinkedList ();
    private boolean nat_hack_ready;
    private boolean nat_hack_behind_nat;
    private Set nat_hack_outside_nat = new HashSet ();

    public static class NatHackMsg extends NetworkMessage {
	public NatHackMsg (NodeId dest) {
	    super (dest, false);
	}
	public NatHackMsg (InputBuffer buffer) throws QSException {
	    super (buffer);
	}
    }

    private void nat_hack_send_msgs () {
	if (nat_hack_behind_nat) {
	    Iterator i = nat_hack_outside_nat.iterator ();
	    while (i.hasNext ()) {
		NodeId n = (NodeId) i.next ();
		System.out.println ("Sending nat hack msg to " + n);
		NatHackMsg m = new NatHackMsg (n);
		dispatch (m);
	    }
	}
    }

    private void handle_nat_hack_msg (NatHackMsg msg) {
	if (nat_hack_behind_nat) {
	    // Got a pong
	    System.out.println ("Got nat hack resp from " + msg.peer);
	    nat_hack_outside_nat.remove (msg.peer);
	    if (nat_hack_outside_nat.isEmpty ()) {
		// All pongs received.  Send out queued msgs.
		nat_hack_ready = true;
		while (! nat_hack_queue.isEmpty ()) {
		    NetworkMessage outb = 
			(NetworkMessage) nat_hack_queue.removeFirst ();
		    handle_outbound_network_message (outb, false);
		}
	    }
	}
	else {
	    // Send a pong
	    dispatch (new NatHackMsg (msg.peer));
	}
    }

    private void handle_outbound_network_message (
	    NetworkMessage msg, boolean delay) {

	final String tag = class_tag + ".handle_outbound_network_message";

	if (DEBUG) Debug.printtagln (tag, "Network got " + msg 
		+ " to send to " + msg.peer + ".");

	// possibly delay the message

	if (delay && (_ts != null)) {
	    Integer Dst = (Integer) _node_id_to_vertex.get (msg.peer);
	    if (Dst == null) {
		Carp.die (tag + ": got message " + msg + " to " + msg.peer + 
			", who is not in the topology mapping.");
	    }
	    Integer Src = (Integer) _node_id_to_vertex.get (_this_node_id);
	    if (Src == null) {
		Carp.die ("We (" + _this_node_id + ") are somehow " +
			"not in the topology mapping.");
	    }
	    int dst = Dst.intValue ();
	    int src = Src.intValue ();
	    long type_code = TypeTable.type_code (msg);
	    CountBuffer cb = new CountBuffer ();
	    cb.add (0);
	    cb.add (msg);
	    int size = cb.size ();
	    TransitStub.PathInfo pi = _ts.shortest_path (
		    src, dst, size, true, type_code);
	    long delay_ms = Math.round (pi.route_time * 1000.0);
	    //long delay_ms = (Math.round (pi.route_time * 1000.0) / 10) + 1;
	    if (delay_ms != 0) {
		if (DEBUG) Debug.printtagln (tag, 
			"Delaying msg from " + dst + 
			" to " + src + " by " + delay_ms + " ms.");
		_classifier.dispatch_later (new DelayWrapper (msg), delay_ms);
		return;
	    }
	}

	if (nat_hack_behind_nat && (! nat_hack_ready) && 
	    ! (msg instanceof NatHackMsg)) {
	    // delay == false
	    nat_hack_queue.addLast (msg);
	    return;
	}

	// If this is the first time we have sent a message to a node,
	// we need to open up a connection to it first.

	ATcpConnection conn = null;

	if ((! _connections.containsKey (msg.peer)) || 
		_closing.containsKey (msg.peer)) {
	    // If we either do not have a connection to this peer as a
	    // remote server, or if we do have such a connection, but
	    // it is being closed, then see if we have a connection
	    // from it as a remote client.

	    if (_server_port_to_client_port.containsKey (msg.peer)) {
		NodeId client_port = 
		    (NodeId) _server_port_to_client_port.get (msg.peer);
		conn = (ATcpConnection) _connections.get (client_port);
	    }
	}
	else {
	    // Otherwise, we can use the connection to this peer as a
	    // remote server.

	    conn = (ATcpConnection) _connections.get (msg.peer);
	}

	if (conn == null) {
	    // If we still have not found a connection, we must
	    // establish one.  If there is already a pending connection
	    // open, add this message to it's queue.  If there is no
	    // pending connection, start one.

	    Queue pending = (Queue) _pending.get (msg.peer);
	    if (pending == null) {

		// Create a queue for pending messages
		pending = new Queue ();
		_pending.put (msg.peer, pending);

		if (DEBUG) Debug.printtagln (tag, 
			"Openning new socket to " + msg.peer 
			+ ".");
		// Open the socket
		try {
		new ATcpClientSocket (
			msg.peer.address (), 
			msg.peer.port (), 
			_this_sink);
		} catch(NullPointerException npe) {
		    Carp.die(tag + ": caught NullPointerException=" + npe +
			     " while attempting to create an ATcpClientSocket"+
			     " msg=" + msg + " msg.peer=" + msg.peer +
			     " msg.peer.addresss=" + (msg.peer==null?null:msg.peer.address()) +
			     " msg.peer.port=" + (msg.peer==null?-1:msg.peer.port()));
		}
	    }
	    else if (DEBUG) {
		Debug.printtagln (tag, 
			"Socket exists but is not yet open.");
	    }

	    if (DEBUG) Debug.printtagln (tag, "Queuing message for now.");
	    pending.enqueue (msg);
	}
	else {

	    // If there is already a connection open to this
	    // destination, enqueue the message to it.

	    send_message (msg, conn, msg.peer);
	}
    }

    protected void send_message (NetworkMessage msg, 
	    ATcpConnection conn, NodeId peer) {

	final String tag = class_tag + ".send_message";

	if (DEBUG) Debug.printtagln (
		tag, "Sending message out on network.");

	int outbound_bytes = 0;
	try {
	    NetworkOutputBuffer buffer = new NetworkOutputBuffer ();
	    buffer.add ((int) 0);
	    buffer.add (msg);
	    outbound_bytes = buffer.size ();
	    buffer.overwrite (0, outbound_bytes - 4);
	    if (DEBUG_SERIALIZATION) 
		System.err.println (class_tag + ": " + msg + ", " + buffer);
	    if (DEBUG) Debug.printtagln (
		    tag, "Size of message is " + buffer.size());

	    buffer.enqueue (conn);
	}
	catch (SinkException e) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": " +
		    "Got SinkException on enqueue of " + 
		    "outgoing message. e=" + e + " Dropping it.");
	}
	catch (Exception e) {
	    Carp.out(tag + ": to_bytes error on type " +
		    msg.getClass () + ".");
	    e.printStackTrace(System.err);
	    Carp.die(e);
	    return;
	}

	if (bw_counting)
	    record_bytes (outbound_bw, msg.peer, outbound_bytes, msg);

    }

    
    private void handle_sink_closed_event (SinkClosedEvent sce) {
	final String tag = class_tag + ".handle_sink_closed_event";

	if (sce.sink instanceof ATcpConnection) {
	    ATcpConnection conn = (ATcpConnection) sce.sink;
	    NodeId peer = new MyNodeId (conn.getPort (),
		    conn.getAddress ());

	    ATcpConnection conn2 = 
		(ATcpConnection) _connections.get (peer);
	    if (conn2 == null) {
		BUG (tag + ":got SinkClosedEvent for non-existant "
			+ "connection: " + peer);
	    }
	    else if (conn2 == conn) {

		if (DEBUG) Debug.printtagln (tag, 
			"Removing connection " + peer + " because "
			+ "of SinkClosedEvent.");

		_connections.remove (peer);
		_connection_streams.remove (peer);
		_closing.remove (peer);

		/** 
		 * TODO: This has been commented out because
		 * it breaks things in the case of simultaneous
		 * connection opening, where one of the
		 * connections is then closed.  Is there a safe
		 * place to remove these translations?
		if (_server_port_to_client_port.containsKey (peer)) {
		    NodeId image = (NodeId)
			_server_port_to_client_port.remove (peer);
		    _client_port_to_server_port.remove (image);
		}
		else 
		    _client_port_to_server_port.remove (peer);
		*/

		if (_pending.containsKey (peer))
		    throw new AssertionViolatedException (
			    tag + ": should not be a pending conn.");
	    }
	    else {
		// We closed this connection ourselves, probably
		// because it was duplicated.  Do nothing.
		BUG (tag + ": duplicate connection to " + peer + ".");
	    }
	}
	else {
	    BUG (tag + ": got SinkClosedEvent for: " + sce.sink);
	}
    }

    private void handle_atcp_connect_failed_event (
	    ATcpConnectFailedEvent failure) {
	final String tag = class_tag + ".handle_atcp_connect_failed_event";

	ATcpClientSocket socket = failure.getSocket ();

	NodeId peer = new MyNodeId (socket.getPort (), socket.getAddress ());
	boolean sensible = false;

	if (_pending.containsKey (peer)) {
	    Queue pending = (Queue)_pending.remove (peer);
	    sensible = true;
	    if (DEBUG) Debug.printtagln (tag, 
		    "Removing pending connection " + peer + " because "
		    + "of ATcpConnectFailedEvent.");
	    if( DebugFlags.LOG_BAD_NETWORK_EVENTS ) {
		StringBuffer msg_sbuf = new StringBuffer(tag);
		msg_sbuf.append(": could not connect to ");
		msg_sbuf.append(socket.getAddress());
		msg_sbuf.append(":");
		msg_sbuf.append(socket.getPort());

		msg_sbuf.append(INFINITE_CONNECT_RETRY ? "resending" : "dropping");
		msg_sbuf.append(": (  ");

		while (pending != null && ! pending.empty ()) {
		    NetworkMessage msg = (NetworkMessage) pending.dequeue ();
		    if(INFINITE_CONNECT_RETRY)
		    {
			msg.inbound = false;
			_classifier.dispatch_later(msg, 5000);
		    }
		    msg_sbuf.append( msg );
		    msg_sbuf.append(", ");
		}
		msg_sbuf.setCharAt( msg_sbuf.length()-2, ')' );

		// TODO: Should be an error in most cases, but not at end of
		// successful experiments.  For now, don't print "ERROR".
		Carp.out( tag + ": " + msg_sbuf.toString() );
	    }
	}

	if (_connections.containsKey (peer)) {
	    if (DEBUG) Debug.printtagln (tag, 
		    "Removing existing connection " + peer + " because "
		    + "of ATcpConnectFailedEvent.");
	    _connections.remove (peer);
	    _connection_streams.remove (peer);
	    if (_closing.containsKey (peer))
		throw new AssertionViolatedException (
			tag + ": should not be in _closing.");
	    sensible = true;
	}

	if (_server_port_to_client_port.containsKey (peer)) {
	    _server_port_to_client_port.remove (peer);
	    if (DEBUG) Debug.printtagln (tag, 
		    "Removing server->client map " + peer + " because "
		    + "of ATcpConnectFailedEvent.");
	    sensible = true;
	}

	if (_client_port_to_server_port.containsKey (peer)) {
	    _client_port_to_server_port.remove (peer);
	    if (DEBUG) Debug.printtagln (tag, 
		    "Removing client->server " + peer + " because "
		    + "of ATcpConnectFailedEvent.");
	    sensible = true;
	}

	/* TODO: this doesn't seem to cause any problems... */
	{
	  String str = null;
	  {
	    str = failure.toString();
	    String errorStr = new String("error");
	    int index = str.lastIndexOf(errorStr);
	    if(index >= 0)
	    {
		String failStr = new String("fail ");
		byte[] bytes = str.getBytes();
		byte[] failBytes = failStr.getBytes();
		for(int i = 0; i < failBytes.length; i++)
		    bytes[index + i] = failBytes[i];
		str = new String(bytes);
	    }
	  }
	  
	  System.err.println(tag + ":could NOT form "
			     + "connection to " + peer + 
			     " because " + str + "."  );
	  System.err.flush();
	}
	if (! sensible ) {
	    BUG (tag + ":Got ATcpConnectFailedEvent for unknown "
		    + "connection to " + peer + ".");
	}
    }

    public void handleEvent (QueueElementIF item) 
    throws EventHandlerException {
      try 
      {
	final String tag = class_tag + ".handleEvent";

	if (DEBUG)
	{
	    String errorStr = new String("error");
	    String str = item.toString();
	    int index = str.lastIndexOf(errorStr);
	    if(index >= 0)
	    {
		String failStr = new String("fail ");
		byte[] bytes = str.getBytes();
		byte[] failBytes = failStr.getBytes();
		for(int i = 0; i < failBytes.length; i++)
		    bytes[index + i] = failBytes[i];
		str = new String(bytes);
	    }
	    if (DEBUG) Debug.printtagln (tag, 
		    "handleEvent called with " + str + " event.");
	}

	if (item instanceof ostore.dispatch.Signal.AlarmSignal) {
	    handle_alarm ();
	}
	else if (item instanceof ATcpConnection) {
	    handle_atcp_connection ((ATcpConnection) item);
	}
	else if (item instanceof SinkFlushedEvent) {
	    handle_sink_flushed_event ((SinkFlushedEvent) item);
	}
	else if ((item instanceof IdentityMessage) && 
		(((IdentityMessage) item).inbound == true)) {
	    process_identity_message ((IdentityMessage) item);
	}
	else if (item instanceof ATcpInPacket) {
	    handle_atcp_inpacket ((ATcpInPacket) item);
	}
	else if (item instanceof DelayWrapper) {
	    handle_outbound_network_message (((DelayWrapper) item).outb, 
		    false /* msg has already been delayed */);
	}
	else if ((item instanceof NetworkMessage) && 
		(((NetworkMessage) item).inbound == false)) {
	    handle_outbound_network_message ((NetworkMessage) item, 
		    true /* allow delay if needed */);
	}
	else if (item instanceof SinkClosedEvent) {
	    handle_sink_closed_event ((SinkClosedEvent) item);
	}
	else if (item instanceof NatHackMsg) {
	    handle_nat_hack_msg ((NatHackMsg) item);
	}
	else if (item instanceof ATcpConnectFailedEvent) {
	    handle_atcp_connect_failed_event ((ATcpConnectFailedEvent) item);
	}
	else if (item instanceof ATcpListenSuccessEvent) {
	    ATcpListenSuccessEvent e = (ATcpListenSuccessEvent) item;
	    ATcpServerSocket s = e.getSocket ();
	    int port = s.getPort ();
	    System.out.println (class_tag + " now listening on port " + 
		    port + ".");
	}
	else if (item instanceof ATcpListenFailedEvent) {
	    ATcpListenFailedEvent e = (ATcpListenFailedEvent) item;
	    ATcpServerSocket s = e.getSocket ();
	    int port = s.getPort ();
	    System.err.println ("ERROR: Network could not listen on " + 
		    "port " + port + ".  There is probably another " + 
		    "program using the port.  Dying...");
	    System.exit (1);
	}
	else if (item instanceof ClearBandwidthCountsReq) {
	    ClearBandwidthCountsReq req = (ClearBandwidthCountsReq) item;
	    print_bw_numbers (req.user_data);
	    inbound_bw = new HashMap (); 
	    outbound_bw = new HashMap ();
	    dispatch (new ClearBandwidthCountsResp (req.user_data));
	}
	else if (item instanceof RecordBytesMsg) {
	    RecordBytesMsg rbm = (RecordBytesMsg) item;
	    Map bwmap;
	    if( rbm.msg.inbound ) {
		bwmap = inbound_bw;
	    } else {
		bwmap = outbound_bw;
	    }
	    record_bytes( bwmap, rbm.msg.peer, (int) rbm.bytes, rbm.msg );
	}
	else {
	    Carp.die (tag + ": got unknown event of type " + 
		    item.getClass ().getName () + ": " + item + ".");
	}
      }
      catch(Exception e)
      {
        Carp.dieGracefully(e, _this_node_id); 
      }
    }

    protected void print_bw_numbers (Object user_data) {
	Set neighbors = new HashSet ();
	neighbors.addAll (inbound_bw.keySet ());
	neighbors.addAll (outbound_bw.keySet ());
	Iterator i = neighbors.iterator ();
	while (i.hasNext ()) { 
	    NodeId neighbor = (NodeId) i.next ();
	    Map inbound_map = (Map) inbound_bw.get (neighbor);
	    Map outbound_map = (Map) outbound_bw.get (neighbor);
	    Set classes = new HashSet ();
	    if (inbound_map != null)
		classes.addAll (inbound_map.keySet ());
	    if (outbound_map != null)
		classes.addAll (outbound_map.keySet ());
	    Iterator j = classes.iterator ();
	    while (j.hasNext ()) {
		String clazz = (String) j.next ();
		long inm = 0, outm = 0;
		long inb = 0, outb = 0;
		if (inbound_map != null) {
		    MsgsBytes In = (MsgsBytes) inbound_map.get (clazz);
		    if (In != null) {
			inb += In.bytes;
			inm += In.msgs;
		    }
		}
		if (outbound_map != null) {
		    MsgsBytes Out = (MsgsBytes) outbound_map.get (clazz);
		    if (Out != null) {
			outb += Out.bytes;
			outm += Out.msgs;
		    }
		}
		System.out.println (user_data + "\t" + neighbor + "\t" + 
			clazz + "\t" + inm + "\t" + outm +
			"\t" + inb + "\t" + outb);
	    }
	}
    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {

	final String tag = class_tag + ".handleEvents";

	if (DEBUG) Debug.printtagln (tag, 
		"handleEvents called with " + items.length + " events.");

	for(int i=0; i<items.length; i++)
	    handleEvent(items[i]);
    }

    public void init (ConfigDataIF config) throws Exception {
	
	String stageName = config.getStage().getName();
	class_tag = stageName;
	final String tag = class_tag + ".init";

	ostore.util.Types.registerAllTypes ();
        if (DEBUG) Debug.turnOn (class_tag);

	_connections = new HashMap ();
	_connection_streams = new HashMap ();
	_mystery_connections = new HashMap ();
	_server_port_to_client_port = new HashMap ();
	_client_port_to_server_port = new HashMap ();
	_pending = new HashMap ();
	_closing = new HashMap ();
	_close_me = new PriorityQueue (10);

	bw_counting = config.getBoolean ("bw_counting");
	String topology_graph = config.getString ("topology_graph");
	String topology_mapping = config.getString ("topology_mapping");

	if ((topology_graph == null) ^ (topology_mapping == null)) {
	    Carp.die (tag + ": must specify topology_graph " +
		    " and topology_mapping together.");
	}

	if (topology_graph != null) {

	     // Open the graph file.
	    try {
		_ts = new TransitStub (topology_graph);
	    }
	    catch (IOException e) {
		Carp.die (tag + ": caught exception when creating " +
			"graph: " + e + ".");
	    }

	    // Read the IP:Port to Vertex mapping
	    _node_id_to_vertex = new HashMap ();
	    StreamTokenizer tok = null;
	    try {
		FileInputStream is = new FileInputStream (topology_mapping);
		Reader reader = 
		    new BufferedReader (new InputStreamReader (is));
		tok = new StreamTokenizer (reader);
	    }
	    catch (Exception e) {
		Carp.die (tag + ": caught exception " + e + " trying " +
			"to read topology mapping " + topology_mapping);
	    }

	    tok.resetSyntax ();
	    tok.whitespaceChars ((int) ' ', (int) ' ');
	    tok.wordChars ((int) 'a', (int) 'z');
	    tok.wordChars ((int) 'A', (int) 'Z');
	    tok.wordChars ((int) '0', (int) '9');
	    tok.wordChars ((int) ':', (int) ':');
	    tok.wordChars ((int) '.', (int) '.');
	    tok.wordChars ((int) '-', (int) '-');

	    int lineno = 1;
	    tok.nextToken ();
	    while (tok.ttype == StreamTokenizer.TT_EOL) {
		tok.nextToken ();
		lineno++;
	    }

	    while (tok.ttype != StreamTokenizer.TT_EOF) {

		// Read the IP:Port
		if (tok.ttype != StreamTokenizer.TT_WORD) {
		    Carp.die (tag + ": expected IP:Port on line " + 
			    lineno + ": " + tok);
		}
		NodeId nid = null;
		try {
		    nid = new NodeId (tok.sval);
		}
		catch (Exception e) {
		    Carp.die (tag + ": Bad IP:Port string on " +
			    "line " + lineno + ": " + e);
		}

		// Read the Vertex
		tok.nextToken ();
		if (tok.ttype != StreamTokenizer.TT_WORD) {
		    Carp.die (tag + ": expected IP:Port on line " + 
			    lineno + ": " + tok);
		}
		Integer vertex = null;
		try {
		    vertex = new Integer (Integer.parseInt (tok.sval));
		}
		catch (NumberFormatException e) {
		    Carp.die (tag + ": number format error reading " +
			    "vertex on line " + lineno + ": " + e);
		}

		// Put them in the mapping.
		if (_node_id_to_vertex.containsKey (nid)) {
		    Carp.die (tag + ": duplicate node id " + nid +
			    " in topology mapping " + topology_mapping);
		}
		_node_id_to_vertex.put (nid, vertex);

		tok.nextToken ();
		while (tok.ttype == StreamTokenizer.TT_EOL) {
		    tok.nextToken ();
		    lineno++;
		}
	    }
	}
    
	_this_node_id = new NodeId (config.getString ("node_id"));	
  	int port = _this_node_id.port();
	_classifier = Classifier.getClassifier (_this_node_id);

	if (DEBUG) Debug.printtagln (tag, "Starting on port " + port + ".");

	try 
	{
	    if(config.getBoolean("Profile"))
	    {
		config.getManager().getProfiler().add("ostore.network.Network", this);
		SinkIF sink =
		    config.getManager().getStage(stageName).getSink();
		_this_sink = new SinkProxy(sink,
			      config.getManager(),
			      config.getStage().getWrapper());
	    }
	    else
		_this_sink =
		    config.getManager().getStage(stageName).getSink();
	} 
	catch (NoSuchStageException e) 
        {
	    BUG (tag + ": could not get own sink.");
	}
		
	Filter filter = new Filter ();
	if (! filter.requireType (
		    Class.forName ("ostore.network.IdentityMessage")))
	    BUG (tag + ": could not require type.");
	if (! filter.requireValue ("inbound", new Boolean (true)))
	    BUG (tag + ": could not require outbound.");
	_classifier.subscribe (filter, _this_sink);


	TypeTable.register_type ("ostore.network.Network$NatHackMsg");
	filter = new Filter ();
	if (! filter.requireType (
		    Class.forName ("ostore.network.Network$NatHackMsg")))
	    BUG (tag + ": could not require type.");
	if (! filter.requireValue ("inbound", new Boolean (true)))
	    BUG (tag + ": could not require outbound.");
	_classifier.subscribe (filter, _this_sink);

	filter = new Filter ();
	if (! filter.requireType (
		    Class.forName ("ostore.network.NetworkMessage")))
	    BUG (tag + ": could not require type.");
	if (! filter.requireValue ("inbound", new Boolean (false)))
	    BUG (tag + ": could not require outbound.");
	_classifier.subscribe (filter, _this_sink);

	filter = new Filter ();
	if (! filter.requireType (
		    Class.forName ("ostore.network.ClearBandwidthCountsReq")))
	    BUG (tag + ": could not require type.");
	_classifier.subscribe (filter, _this_sink);

	filter = new Filter ();
	if (! filter.requireType (
		    Class.forName ("ostore.network.Network$DelayWrapper")))
	    BUG (tag + ": could not require type.");
	_classifier.subscribe (filter, _this_sink);

	if( bw_counting ) {
	    filter = new Filter ();
	    if (! filter.requireType (
		Class.forName ("ostore.network.RecordBytesMsg")))
		BUG (tag + ": could not require type.");
	    _classifier.subscribe (filter, _this_sink);

	}

	Signal.getSignal (_this_node_id).register (
		Signal.ALRM, "NetworkStage", _this_sink);
	Signal.getSignal (_this_node_id).alarm ("NetworkStage", 10000);
	
	// Open the server socket
	if (DEBUG) Debug.printtagln (tag, 
		"Openning server socket on port " + port + ".");
	serversock = new ATcpServerSocket (port, _this_sink);

	// TODO: start timeout thread for _mystery_connections

 	nat_hack_behind_nat = config.getBoolean ("nat_hack_behind_nat");
	if (nat_hack_behind_nat) {
	    int m = config.getInt ("nat_hack_outside_hosts");
	    for (int i = 0; i < m; ++i) {
		NodeId h = new NodeId (config.getString (
			    "nat_hack_outside_host_" + i));
		nat_hack_outside_nat.add (h);
	    }
	    nat_hack_send_msgs (); 
	}

   }

    public void destroy() throws Exception {
	serversock.close();
	// TODO
    }

    //  methods for ProfilableIF
    public int profileSize()
    {
	return _this_sink.size();
    }
    //  end methods for ProfilableIF

    protected NodeId process_identity_message (IdentityMessage ident) {

	final String tag = class_tag + ".process_identity_message";
	NodeId result = null;

	if (DEBUG) Debug.printtagln (tag, "got " + ident);

	if (ident.node_ids.size () != 1)
	    throw new AssertionViolatedException (
		    "can't handle multiple node ids in Identity message.");

	// Look up the mystery connection that this message corresponds
	// to.
	Pair p = (Pair) _mystery_connections.remove (ident.peer);
	if (p == null) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": " +
		    "Got unexpected IdentityMessage from " 
		    + ident.peer + 
		    ".  (Maybe it's running MultiplexedNetwork...)");
	    return result;
	}

	_connections.put (ident.peer, p.first);

	result = (NodeId) ident.node_ids.getFirst ();

	// TODO: authenticate it

	ATcpConnection conn_to_remote_server = 
	    (ATcpConnection) _connections.get (result);

	if (conn_to_remote_server != null) {
	    // We have openned this connection from both sides
	    // simultaneously, so we need to pick a client and
	    // server from TCP's point of view.  Let the node with
	    // the lowest NodeId be the server.

	    if (_this_node_id.less_than (result)) {
		// We are the server.  Close our outgoing connection to the
		// remote server and use the incoming one instead.

		if (DEBUG) Debug.printtagln (tag,
			"We are the server.  Closing old conn.");

		_closing.put (result, conn_to_remote_server);
		_close_me.insert (conn_to_remote_server,
			System.currentTimeMillis () + 10000, false);

		if (DEBUG) Debug.printtagln (tag,
			"Adding connection for " + ident.peer + ".");
		if (DEBUG) Debug.printtagln (tag,
			"Adding translation " + ident.peer + " to " 
			+ result + ".");
		_client_port_to_server_port.put (ident.peer, result);
		_server_port_to_client_port.put (result, ident.peer);
	    }
	    else {
		// We are the client.  Close our incoming connection and
		// use the outgoing one instead.

		if (DEBUG) Debug.printtagln (tag,
			"We are the client.  Closing incoming conn.");

		// Even though we are closing this connection, we may have
		// already received some other messages from it, so store
		// the mapping until it is really closed.

		_closing.put (ident.peer, p.first);
		_client_port_to_server_port.put (ident.peer, result);
	    }
	}
	else {
	    // This connection was not yet open, so we are cool.

	    if (DEBUG) Debug.printtagln (tag,
			"Adding connection for " + ident.peer + ".");
	    if (DEBUG) Debug.printtagln (tag,
		    "Adding translation " + ident.peer + " to " 
		    + result + ".");
	    _client_port_to_server_port.put (ident.peer, result);
	    _server_port_to_client_port.put (result, ident.peer);
	}

	return result;
    }

    private void send_pending_messages (Queue pending, ATcpConnection conn) {
	final String tag = class_tag + ".send_pending_messages";

	if (pending.size () <= 0)
	    throw new AssertionViolatedException (
		    tag + ": pending queue bug.");

	while (! pending.empty ()) {
	    NetworkMessage msg = (NetworkMessage) pending.dequeue ();
	    send_message (msg, conn, msg.peer);
	}
    }

    private NodeId _this_node_id;

    private Classifier _classifier;

    private HashMap _connections;
    private HashMap _connection_streams;
    private HashMap _mystery_connections;
    private HashMap _client_port_to_server_port;
    private HashMap _server_port_to_client_port;
    private HashMap _pending;
    private HashMap _closing;
    private SinkIF _this_sink;
    private PriorityQueue _close_me;
    private Map _node_id_to_vertex;
    private TransitStub _ts;

    // NodeId -> Class Name (String) -> long
    protected Map inbound_bw = new HashMap (), outbound_bw = new HashMap ();
    protected boolean bw_counting;
    private ATcpServerSocket serversock;
}

