/**
 * NetworkInputBuffer.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.network;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.SortedMap;
import java.util.TreeMap;
import seda.sandStorm.lib.aSocket.ATcpInPacket;
import ostore.util.AssertionViolatedException;
import ostore.util.ByteUtils;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.InputBuffer;
import ostore.util.InputBufferImpl;
import ostore.util.OutputBuffer;

/**
 * Implements the InputBuffer interface as a linked list of ATcpInPacket
 * objects.
 */
public class NetworkInputBuffer extends InputBufferImpl {

    private static final boolean DEBUG = DebugFlags.NETWORK_NETWORK;

    private LinkedList in_order_packets = new LinkedList ();
    private long next_seq_num;
    private int first_packet_offset;
    private int bytes_available;

    public Object clone () throws CloneNotSupportedException {
	NetworkInputBuffer result = (NetworkInputBuffer) super.clone ();
	result.next_seq_num = next_seq_num;
	result.first_packet_offset = first_packet_offset;
	result.bytes_available = bytes_available;
	result.in_order_packets = new LinkedList ();
	Iterator i = in_order_packets.iterator ();
	while (i.hasNext ()) result.in_order_packets.addLast (i.next ());
	return result;
    }

    public void convertToBytes (byte out_buffer [][], int out_offset [][]) {
	if (bytes_available > 0) {
	    out_buffer [0] = new byte [bytes_available];
	    out_offset [0][0] = 0;

	    Iterator i = in_order_packets.iterator ();
	    int offset = first_packet_offset;
	    int remaining = bytes_available;
	    while (remaining > 0) {
		ATcpInPacket packet = (ATcpInPacket) i.next ();
		int toread = Math.min (packet.size () - offset, remaining);
		System.arraycopy (packet.getBytes (), offset,
			out_buffer [0], out_offset [0][0], toread);
		out_offset [0][0] += toread;
		remaining -= toread;
		offset = 0;
	    }

	    out_offset [0][0] = 0;
	}
	else {
	    out_buffer [0] = new byte [0];
	    out_offset [0][0] = 0;
	}
    }

    public NetworkInputBuffer (long next_seq_num) {
	this.next_seq_num = next_seq_num;
    }

    void set_available (int value) {
	if (value > bytes_available)
	    throw new AssertionViolatedException (
		    "can only decrease or stay same");
	bytes_available = value;
    }

    void set_first_packet_offset (int value) {
	first_packet_offset = value;
    }

    int peek_int () {
	if (bytes_available < 4) throw new AssertionViolatedException (
		"must have at least four bytes to peek_int");
	ATcpInPacket packet = (ATcpInPacket) in_order_packets.getFirst ();
	if (packet.size () - first_packet_offset >= 4) {
	    int [] offset = {first_packet_offset};
	    int res = ByteUtils.bytesToInt (packet.getBytes (), offset);
	    return res;
	}
	else {
	    byte [] bytes = new byte [4];
	    int rem = 4;
	    int boffset = 0;
	    int poffset = first_packet_offset;
	    Iterator i = in_order_packets.iterator ();
	    while (rem > 0) {
		packet = (ATcpInPacket) i.next ();
		byte [] pbytes = packet.getBytes ();
		int toread = Math.min (pbytes.length - poffset, rem);
		System.arraycopy (pbytes, poffset, bytes, boffset, toread);
		poffset = 0;
		boffset += toread;
		rem -= toread;
	    }
	    int [] offset = {0};
	    return ByteUtils.bytesToInt (bytes, offset);
	}
    }

    public byte nextByte () {
	if (bytes_available < 1) 
	    throw new AssertionViolatedException ("no byte available");
	ATcpInPacket packet = (ATcpInPacket) in_order_packets.getFirst ();
	byte result = packet.getBytes () [first_packet_offset];
	first_packet_offset += 1;
	if (first_packet_offset == packet.size ()) {
	    in_order_packets.removeFirst ();
	    first_packet_offset = 0;
	}
	--bytes_available;
	return result;
    }

    public void nextBytes (byte [] out_buffer, int out_offset, int count) {

	if (count > bytes_available)
	    throw new AssertionViolatedException (
		    "count (" + count + ") > bytes_available (" +
		    bytes_available + ")");

	ATcpInPacket packet = (ATcpInPacket) in_order_packets.getFirst ();
	int remaining = count;
	while (remaining > 0) {
	    int toread = Math.min (packet.size () - first_packet_offset, 
		    remaining);
	    System.arraycopy (packet.getBytes (), first_packet_offset,
		    out_buffer, out_offset, toread);
	    out_offset += toread;
	    first_packet_offset += toread;
	    remaining -= toread;

	    if (first_packet_offset == packet.size ()) {
		in_order_packets.removeFirst ();
		if (remaining != 0)
		    packet = (ATcpInPacket) in_order_packets.getFirst ();
		first_packet_offset = 0;
	    }
	}

	bytes_available -= count;
    }

    public void skipBytes (int count) {

	if (count > bytes_available)
	    throw new AssertionViolatedException ("count > bytes_available");
	if (count == 0)
	    return;

	ATcpInPacket packet = (ATcpInPacket) in_order_packets.getFirst ();
	int remaining = count;
	while (remaining > 0) {
	    int toread = Math.min (packet.size () - first_packet_offset, 
		    remaining);
	    first_packet_offset += toread;
	    remaining -= toread;

	    if (first_packet_offset == packet.size ()) {
		in_order_packets.removeFirst ();
		if (remaining != 0)
		    packet = (ATcpInPacket) in_order_packets.getFirst ();
		first_packet_offset = 0;
	    }
	}

	bytes_available -= count;
    }

    public void nextBytes (OutputBuffer output) {

        if (bytes_available == 0)
	    return;

	ATcpInPacket packet = (ATcpInPacket) in_order_packets.getFirst ();
	int remaining = bytes_available;
	while (remaining > 0) {
	    int toread = Math.min (packet.size () - first_packet_offset, 
		    remaining);
	    output.add (packet.getBytes (), first_packet_offset, toread);
	    first_packet_offset += toread;
	    remaining -= toread;

	    if (first_packet_offset == packet.size ()) {
		in_order_packets.removeFirst ();
		if (remaining != 0)
		    packet = (ATcpInPacket) in_order_packets.getFirst ();
		first_packet_offset = 0;
	    }
	}

	bytes_available = 0;
    }

    public InputBuffer subBuffer (int count) {

	if (count > bytes_available)
	    throw new AssertionViolatedException (
		    "count (" + count + ") > bytes_available (" +
		    bytes_available + ")");

	NetworkInputBuffer result = new NetworkInputBuffer (next_seq_num);
	result.first_packet_offset = first_packet_offset;
	result.bytes_available = count;

	ATcpInPacket packet = (ATcpInPacket) in_order_packets.getFirst ();
	int remaining = count;
	while (remaining > 0) {
	    int toread = Math.min (packet.size () - first_packet_offset, 
		    remaining);
	    result.in_order_packets.addLast (packet);
	    first_packet_offset += toread;
	    remaining -= toread;

	    if (first_packet_offset == packet.size ()) {
		in_order_packets.removeFirst ();
		if (remaining != 0)
		    packet = (ATcpInPacket) in_order_packets.getFirst ();
		first_packet_offset = 0;
	    }
	}

	bytes_available -= count;
	return result;
    }

    public void add_packet (ATcpInPacket packet) {
	long seq = packet.getSequenceNumber ();
	if (seq == next_seq_num) {
	    in_order_packets.addLast (packet);
	    ++next_seq_num;
	    bytes_available += packet.size ();
	}
	else {
	    throw new AssertionViolatedException ("must be in order");
	}
    }

    public String toString () {
	String result = "(NetworkInputBuffer ";
	Iterator i = in_order_packets.iterator ();
	int count = 0, start = first_packet_offset;
	while (i.hasNext ()) {
	    ATcpInPacket packet = (ATcpInPacket) i.next ();
	    byte [] data = packet.getBytes ();
	    int toprint = Math.min (
		    data.length - start, 
		    bytes_available - count);
	    result += ostore.util.ByteUtils.print_bytes (data, start, toprint);
	    start = 0;
	}
	return result + ")";
    }
}














