/**
 * MonitorMsg.java
 *
 * @author Jeremy Stribling
 * @version    $Id: MonitorMsg.java,v 1.5 2003/05/08 01:55:05 strib Exp $
 *
 * Copyright (c) 2003 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;
import ostore.util.NodeId;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;

/**
 *
 * A <code>MonitorMsg</code> is sent from a stage interested in monitoring
 * statistics to certain <code>NodeId</code>s.  This message tells the
 * Patchwork stage to begin monitoring this node at a certain level of
 * probing, until further notice.  It can also serve to change the level
 * of probing of a node that's already being monitored.  Since the specific
 * constructors for details.  See <code>Patchwork</code> for a description 
 * of what "probing level" means.
 *
 * @see Patchwork
 *
 * @author Jeremy Stribling
 * @version    $Id: MonitorMsg.java,v 1.5 2003/05/08 01:55:05 strib Exp $
 */
public class MonitorMsg implements QueueElementIF {

    public NodeId[] peers;
    public int[] levels;
    public long[] values;
    public byte type;
    public Object userData;
    public SinkIF sink;

    private MonitorMsg (NodeId peer, Object userData, SinkIF sink) {
	NodeId[] tmp = {peer}; 
	peers = tmp;
	this.userData = userData;
	this.sink = sink;
	levels = null;
	values = null;
    }

    /**
     * Tell Patchwork to start monitoring a node at a given level.
     *
     * @param peer   The node to monitor
     * @param level   The specified probing level
     */
    public MonitorMsg (NodeId peer, int level, Object userData, SinkIF sink) {
	this( peer, userData, sink );
	int[] tmp = {level};
	levels = tmp;
    }

    /**
     * Tell Patchwork to start monitoring a node, given a value for
     * a statistic as a starting point.  From this value, Patchwork
     * will determine what the probing level of the node should be.
     *
     * @param peer   The node to monitor
     * @param value  The given initial value for some statistic 
     *                   (also determines the probing level)
     * @param type   The type of the above value (see <code>Patch</code>)
     *
     * @see Patch
     */
    public MonitorMsg (NodeId peer, long value, byte type, Object userData, 
		       SinkIF sink) {
	this( peer, userData, sink );
	long[] tmp = {value};
	values = tmp;
	this.type = type;
    }

    /**
     * Tell Patchwork to start monitoring a node, given a value for
     * a statistic as a starting point.  Also given is the desired probing
     * level of that node.
     *
     * @param peer   The node to monitor
     * @param level   The specified probing level
     * @param value  The given initial value for some statistic 
     * @param type   The type of the above value (see <code>Patch</code>)
     *
     * @see Patch
     */
    public MonitorMsg (NodeId peer, int level, long value, byte type, 
		       Object userData, SinkIF sink) {
	this( peer, value, type, userData, sink );
	int[] tmp = {level}; 
	levels = tmp;
    }

    private MonitorMsg (NodeId[] peers, Object userData, SinkIF sink) {
	this.peers = peers;
	this.userData = userData;
	this.sink = sink;
	levels = null;
	values = null;
    }

    /**
     * Tell Patchwork to start monitoring some nodes at given levels.
     *
     * @param peers   The nodes to monitor
     * @param levels   The specified probing levels, corresponding to each node
     */
    public MonitorMsg (NodeId[] peers, int[] levels, Object userData, 
		       SinkIF sink) {
	this( peers, userData, sink );
	this.levels = levels; 
    }

    /**
     * Tell Patchwork to start monitoring some nodes, given values for
     * a statistic as a starting point.  From these values, Patchwork
     * will determine what the probing levels of the nodes should be.
     *
     * @param peers   The nodes to monitor
     * @param values  The given initial values for some statistic 
     *                   (also determines the probing levels)
     * @param type   The type of the above values (see <code>Patch</code>)
     *
     * @see Patch
     */
    public MonitorMsg (NodeId[] peers, long[] values, byte type, 
		       Object userData, SinkIF sink) {
	this( peers, userData, sink );
	this.values = values; 
	this.type = type;
    }

    /**
     * Tell Patchwork to start monitoring some nodes, given values for
     * a statistic as a starting point.  Also given is the desired probing
     * levels of those nodes.
     *
     * @param peers   The nodes to monitor
     * @param levels   The specified probing levels, corresponding to each node
     * @param values  The given initial values for some statistic 
     * @param type   The type of the above values (see <code>Patch</code>)
     *
     * @see Patch
     */
    public MonitorMsg (NodeId[] peers, int[] levels, long[] values, byte type, 
		       Object userData, SinkIF sink) {
	this( peers, values, type, userData, sink );
	this.levels = levels; 
    }

    public Object clone () throws CloneNotSupportedException {
	MonitorMsg newone = (MonitorMsg) super.clone ();
	newone.peers = peers;
	newone.levels = levels;
	newone.values = values;
	newone.userData = userData;
	newone.sink = sink;
	return newone;
    }

    public String toString () {
	String s = "";

	s += "(MonitorMsg, peers: {";
	for( int i = 0; i < peers.length; i++ ) {
	    s += peers[i] + " ";
	}
	s += "}, levels {"; 
	if( levels == null ) {
	    s+= "null";
	} else {
	    for( int i = 0; i < levels.length; i++ ) {
		s += levels[i] + " ";
	    }
	}
	s += "}, values {";
	if( values == null ) {
	    s += "null";
	} else {
	    for( int i = 0; i < values.length; i++ ) {
		s += values[i] + " ";
	    }
	}
	s += "}. userData: " + userData + ")";
	return s;
    }
}

