/**
 * Array.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.oil;

import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Vector;
import java.util.TreeMap;
import java.text.NumberFormat;
import ostore.util.QuickSerializable;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QSException;
import ostore.util.ByteUtils;
import ostore.util.TypeTable;

/**
 * A sparse array of <code>double</code>s.  Elements are indexed by
 * arbitrary <code>QuickSerializable</code> objects.
 * <p>
 * The <code>Array</code> objects understand the concept of a
 * <i>default</i> element, which is returned for unknown IDs.  Using
 * this functionality, an <code>Array</code> could explicitly store only
 * <i>interesting</i> elements, yet appear to store a full array.
 *
 * This class is synchronized.
 *
 * @author	Dennis Geels
 * @version	$Id: Array.java,v 1.16 2002/07/20 19:38:03 srhea Exp $
 */

public class Array implements Cloneable, QuickSerializable
{
    ////////////////////       CONSTANTS	////////////////////

    private static final boolean DEBUG = DebugFlags.OIL_ARRAY;

    /**
     * The default value for <code>default_elt</code>.
     * Currently: <code>Double.Nan</code>
     */
    public static final double DEFAULT_DEFAULT = Double.NaN;

    /**
     * Used in <code>toString</code> for uniform number lengths.
     */
    private static final int num_frac_digits = 4;
    private static final double ln_10 = Math.log(10);
    
    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * The value to return for unknown IDs.
     */
    protected double default_elt;

    /**
     * The internal storage mechanism.
     */
    protected HashMap map;
    
    ////////////////////  STATIC CONSTRUCTORS	////////////////////
    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Construct a new <code>Array</code> with the default default.
     */
    public Array()
    {
	this( DEFAULT_DEFAULT );
    }

    /**
     * Construct a new <code>Array</code> with the specified default.
     *
     * @param default_elt	The value to be returned for any unknown IDs
     */
    public Array( double default_elt )
    {
	this.default_elt = default_elt;
	this.map = new HashMap();
    }
    
    /**
     * Constructs a <code>Array</code> from its
     * <code>serialize</code>d form.
     */
    public Array( InputBuffer buffer )
	throws QSException
    {
	this.default_elt = buffer.nextDouble();
	int size = buffer.nextInt();
	this.map = new HashMap( size );

	long type;
	QuickSerializable id;
	double value;
	for( int i=0; i<size; i++ ) {
	    // Deserialize the id
	    id = (QuickSerializable)buffer.nextObject();

	    // Now deserialize the value
	    value = buffer.nextDouble();

	    // Finally, add the mapping to the <code>Array</code>.
	    put( id, value );
	}
    }
    
    /**
     * Constructs a <code>Array</code> from its
     * <code>serialize</code>d form.
     *
     * @param decompression_map	An <code>Integer->QuickSerializable</code> mapping
     * 				for decompression. 
     */
    public Array( InputBuffer buffer, HashMap decompression_map )
	throws QSException
    {
	this.default_elt = buffer.nextDouble();
	int size = buffer.nextInt();
	this.map = new HashMap( size );

	int mapped_id;
	QuickSerializable id;
	double value;
	for( int i=0; i<size; i++ ) {
	    // Deserialize the id, using the decompression mapping
	    mapped_id = buffer.nextInt();
	    id = (QuickSerializable)
		decompression_map.get( new Integer(mapped_id) );
	    if( id == null ) {
		throw new QSException("Array: corrupted cecompression map");
	    }

	    // Now deserialize the value
	    value = buffer.nextDouble();

	    // Finally, add the mapping to the <code>Array</code>.
	    put( id, value );
	}
    }

    ////////////////////     STATIC METHODS 	////////////////////

    /**
     * Returns a human-readable representation of the specified array.
     * Formatting assumes small, positive array values.  Other arrays may
     * not look as pretty.
     *
     * @param array	the array to print
     *
     * @return a human-readable representation of the specified array.
     */
    public static String toString( double[] array )
    {
	NumberFormat nf = NumberFormat.getInstance();
	// Set the formatter for values.
	nf.setMinimumIntegerDigits( 1 );
	nf.setMaximumFractionDigits( num_frac_digits );
	nf.setMinimumFractionDigits( num_frac_digits );

	// Estimate the buffer length (ok to miss).
	// Assume average of at most two-digit ints for values.
	int len = (array.length * (num_frac_digits+4));
	StringBuffer sbuf = new StringBuffer( len );
	
	sbuf.append('{');
	for( int i=0; i<array.length; i++ ) {
	    sbuf.append( nf.format( array[i] ) );
	    sbuf.append(',');
	}
	sbuf.setCharAt( sbuf.length()-1, '}' );

	return sbuf.toString();
    }

    /**
     * Returns a human-readable representation of the specified array, with
     * indices.  This method should be used to print sub-arrays, or simply
     * because the indices are helpful
     * Formatting assumes small, positive array values.  Other arrays may
     * not look as pretty.
     *
     * @param array	the array to print
     * @param indices	the indices, in order, of the elements to print
     *
     * @exception IllegalArgumentException if <code>indices.length >
     * array.length</code>.
     *
     * @see #max
     *
     * @return a human-readable representation of the specified array.
     */
    public static String toString( double[] array, int[] indices )
    {
	if( indices.length > array.length ) {
	    throw new
		IllegalArgumentException( "Cannot find " + indices.length +
					  " elements in array of length " +
					  array.length );
	}
	NumberFormat nf = NumberFormat.getInstance();
	// Tell the formatter to print indices at the same length.
	int num_int_digits =
	    (int)(Math.ceil(Math.log(array.length)/ln_10));
		
	// Estimate the buffer length (ok to miss).
	// Assume average of at most two-digit ints for values.
	int element_len = (num_frac_digits+num_int_digits+7);
	int len = (indices.length * element_len);
	StringBuffer sbuf = new StringBuffer( len );
	
	sbuf.append('{');
	for( int i=0; i<indices.length; i++ ) {
	    sbuf.append( '<' );
	    // Set the formatter for indices.
	    nf.setMinimumIntegerDigits( num_int_digits );
	    nf.setMaximumFractionDigits( 0 );
	    nf.setMinimumFractionDigits( 0 );
	    sbuf.append( nf.format( indices[i] ) );
	    sbuf.append( ':' );
	    // Set the formatter for values.
	    nf.setMinimumIntegerDigits( 1 );
	    nf.setMaximumFractionDigits( num_frac_digits );
	    nf.setMinimumFractionDigits( num_frac_digits );
	    sbuf.append( nf.format( array[indices[i]] ) );
	    sbuf.append( ">" );
	    sbuf.append(',');
	}
	sbuf.setCharAt( sbuf.length()-1, '}' );

	return sbuf.toString();
    }

    /**
     * Finds the indices of the largest elements of the specified array.
     * The length of <code>indices</code> determines the number of top
     * elements to find; it must be greater than <code>array.length</code>.
     *
     * @param array	the array to check
     * @param indices	the array into which to put the (sorted) indices of
     * the largest elements of <code>array</code>.
     *
     * @exception IllegalArgumentException if <code>indices.length >
     * array.length</code>. 
     */
    public static void max( double[] array, int[] indices )
    {
	if( indices.length > array.length ) {
	    throw new
		IllegalArgumentException( "Cannot find " + indices.length +
					  " elements in array of length " +
					  array.length );
	}
	// Loop through <code>array</code>, using Insertion Sort to
	// maintain <code>indices</code>.
	// If <code>indices.length</code> is of the same order of magnitude
	// as <code>array.length</code>, this method would be more
	// efficiently implemented using a heap.
	for( int i=0; i<array.length; i++ ) {
	    int nominee = i;
	    // Only loop over the entries we've already filled
	    for( int j=0; j<=i && j<indices.length; j++ ) {
		if( j == i ) {
		    // First time here - nothing to compare
		    indices[j] = nominee;
		} else if( array[nominee] > array[indices[j]] ) {
		    // Insert the nominee here and bubble down.
		    int new_nominee = indices[j];
		    indices[j] = nominee;
		    nominee = new_nominee;
		}
	    }
	}
    }

    //////////////////// 	METHODS		////////////////////

    /**
     * Returns the value that this <code>Array</code> should return for
     * unknown IDs.
     */
    public synchronized double get_default()
    {
	return default_elt;
    }

    /**
     * Defines the value that this <code>Array</code> should return for
     * unknown IDs.
     *
     * @param default_elt	The value to return for unknown IDs.
     */
    public synchronized void set_default( double default_elt )
    {
	this.default_elt = default_elt;
    }

    /**
     * Returns the size of the <code>Array</code>.
     * 
     * @return the number of elements explicitly stored in this
     * <code>Array</code>. 
     */
    public synchronized int size()
    {
	return map.size();
    }
    
    /**
     * Returns an <code>Iterator</code> over the IDs stored explicitly in
     * this <code>Array</code>.
     *
     * @return a <code>HashIterator</code> over the backing map's keys.
     */
    public synchronized Iterator ids(){
	return map.keySet().iterator();
    }

    /**
     * Shows whether the specified ID is explicitly stored in this
     * <code>Array</code>.
     *
     * @param id	The unique ID for the element to check.
     * @return <code>true</code> iff <code>id</code> is explicitly mapped
     * to a value.
     */
    public synchronized boolean contains( QuickSerializable id )
    {
	return map.containsKey( id );
    }

    /**
     * Returns the value indexed by the specified ID, or the default
     * element if no such value exists.
     *
     * @param id	The unique ID for an element of this
     * 			<code>Array</code>.
     * @return the value stored under <code>id</code>, or the default
     * element.
     */
    public synchronized double get( QuickSerializable id )
    {
	Object val = map.get( id );
	return ((val==null) ? default_elt : ((Double)val).doubleValue());
    }

    /**
     * Places the specified mapping into this <code>Array</code>.  If
     * <code>id</code> already maps to a value, the old value is
     * overwritten and returned.
     *
     * @param id	The index under which to store this element.
     * @param value	The value to store for this index.
     * @return the value previously mapped by this index, or the default
     * element if there was no previous mapping.
     */
    public synchronized double put( QuickSerializable id, double value )
    {
	Object val = map.put( id, new Double(value) );
	return ((val==null) ? default_elt : ((Double)val).doubleValue());
    }

    /**
     * Removes the value indexed by the specified ID
     *
     * @param id	The unique ID for an element of this
     * 			<code>Array</code>.
     * @return the value stored under <code>id</code>, or the default
     * element.
     */
    public synchronized double remove( QuickSerializable id )
    {
	Object val = map.remove( id );
	return ((val==null) ? default_elt : ((Double)val).doubleValue());
    }

    /**
     * Removes all elements from this <code>Array</code>.
     * This method does not reset the default return value.
     */
    public synchronized void clear()
    {
	map.clear();
    }

    /**
     * Adds the contents of the specified <code>Array</code> to this one.
     * Values for known IDs are added to their current values.
     * Unknown IDs are added directly (default elements are not considered).
     *
     * @param addend	The <code>Array</code> whose contents to add
     */
    public synchronized void add( Array addend )
    {
	Iterator ids = addend.ids();
	QuickSerializable id;
	
	while( ids.hasNext() ) {
	    id = (QuickSerializable)ids.next();
	    if( map.containsKey( id ) ) {
		put( id, (addend.get( id ) + get( id )) );
	    } else {
		map.put( id, addend.map.get( id ) );
	    }
	}
    }

    /**
     * Adds the specified value to this <code>Array</code>.
     * If the ID is known, the value is added to the current value.
     * Otherwise it is inserted as in
     * {@link #put(ostore.util.QuickSerializable, double)}.
     *
     * @param id	The unique ID for <code>value</code>.
     * @param value	The value to add.
     */
    public synchronized void add( QuickSerializable id, double value )
    {
	if( map.containsKey( id ) ) {
	    put( id, (value + get( id )) );	    
	} else {
	    put( id, value );
	}
    }

        //FIXME!!
//      void sparsify( double threshold ){}

    /**
     * Shrinks this <code>Array</code> to <code>new_size</code>,
     * keeping only the largest elements.  Ties are broken in an arbitrary,
     * non-deterministic fashion.
     *
     * @param new_size	The number of elements to keep.
     */
    public synchronized void sparsify( int new_size )
    {
	if( new_size >= size() ) return;

	// Use a <code>TreeMap</code> as a priority queue to store the ids
	// of the largest elements.
	TreeMap pqueue = new TreeMap();
	Vector to_drop = new Vector( size() - new_size );
	Iterator ids = ids();
	QuickSerializable id;
	// Populate the queue with the first <code>new_size</code> elements.
	for( int i=0; i<new_size; i++ ) {
	    id = (QuickSerializable)ids.next();
	    Double val = (Double)map.get(id);
	    Object old = pqueue.put( val, id );	// inverse mapping, now.
	    if( old != null ) {
		// Identical values - map the value to a
		// <code>Vector</code> which holds all the matching ids.
		Vector set;
		if( old instanceof Vector ) {
		    set = (Vector)old;
		} else {
		    set = new Vector();
		    set.add( old );
		}
		set.add( id );
		pqueue.put( val, set );
	    }		
	}
	// <code>pqueue<code> now contains <code>new_size</code> ids.  Scan
	// across the rest of the elements in the <code>Array</code>,
	// retaining at all times the largest <code>new_size</code> elements.
	Double lowest = (Double)pqueue.firstKey();
	
	while( ids.hasNext() ) {
	    id = (QuickSerializable)ids.next();
	    Double val = (Double)map.get(id);

	    if( lowest.compareTo(val) < 0 ) {
		// swap in this element
		Object old = pqueue.put( val, id );
		if( old != null ) {
		    // Identical values - map the value to a
		    // <code>Vector</code> which holds all the matching ids.
		    Vector set;
		    if( old instanceof Vector ) {
			set = (Vector)old;
		    } else {
			set = new Vector();
			set.add( old );
		    }
		    set.add( id );
		    pqueue.put( val, set );
		}		
		// swap out the lowest element
		Object lowest_id = pqueue.remove(lowest);
		if( lowest_id instanceof Vector ) {
		    // Many ids for this value - just remove one.
		    Vector set = (Vector)lowest_id;
		    int size = set.size();
		    if( size == 1 ) {
			to_drop.add( set.get(0) );
		    } else {
			Object last_id = set.remove( size-1 );
			to_drop.add( last_id );
			pqueue.put( lowest, set );
		    }
		} else {
		    // Single element - drop it and get the new lowest key
		    to_drop.add( lowest_id );		    
		    lowest = (Double)pqueue.firstKey();
		}
	    } else {
		to_drop.add( id );
	    }
	}
	if( DEBUG ) {
	    if( to_drop.size() != (size()-new_size) ) {
		Carp.die( "Failed to sparsify to correct size" );
	    }
	}
	// The <code>new_size</code> elements remaining in
	// <code>pqueue</code> at this point are the largest overall.
	// All other elements have been stored in <code>to_drop</code>.
	// Remove the latter from this <code>Array</code>.
	for( int i=0; i<to_drop.size(); i++ ) {
	    remove( (QuickSerializable)to_drop.get(i) );
	}
	return;
    }
    
    // FIXME
    // subtract( Array subtrahend );

    /**
     * Performs a slightly deeper copy than <code>Object.clone</code>.
     * The mapping is cloned, so the new <code>Array</code> can be modified
     * independently of this one.  Neither the IDs nor the values are
     * themselves cloned.
     *
     * @return a new <code>Array</code> with the same set of elements as
     * this one.
     */
    public synchronized Object clone(){
	Array cloned = new Array( default_elt );
	cloned.map = (HashMap)map.clone();
	return cloned;
    }

    /**
     * Produce a human-readable version of this <code>Array</code>.
     *
     * @return a <code>String</code> of  &lt id:value &gt pairs.
     */
    public synchronized String toString()
    {
	NumberFormat nf = NumberFormat.getInstance();
	StringBuffer sbuf = new StringBuffer( "(" );
	sbuf.append( nf.format(default_elt) );
	sbuf.append( ')' );

	Iterator ids = ids();
	QuickSerializable id;
	while( ids.hasNext() ) {
	    id = (QuickSerializable)ids.next();
	    sbuf.append( '<' );
	    sbuf.append( id );
	    sbuf.append( ':' );
	    // Set the formatter for values.
	    nf.setMinimumIntegerDigits( 1 );
	    nf.setMaximumFractionDigits( num_frac_digits );
	    nf.setMinimumFractionDigits( num_frac_digits );
	    sbuf.append( nf.format(get(id)) );
	    sbuf.append( ">\n" );
	}
	return sbuf.toString();
    }
    
    /* Specified by ostore.util.QuickSerializable */
    public void serialize( OutputBuffer buffer )
    {
	buffer.add( default_elt );
	buffer.add( map.size() );

	Iterator ids = ids();
	QuickSerializable id;
	
	while( ids.hasNext() ) {
	    id = (QuickSerializable)ids.next();

	    // First store the id
	    buffer.add( id );
	    
	    // Now store the value
	    buffer.add( get(id) );
	}
    }

    /**
     * Serializes the <code>Array</code> using a compression map.
     */
    public void serialize( OutputBuffer buffer, HashMap compression_map )
    {
	buffer.add( default_elt );
	buffer.add( map.size() );

	Iterator ids = ids();
	QuickSerializable id;
	Integer mapped_id;
	
	while( ids.hasNext() ) {
	    id = (QuickSerializable)ids.next();
	    
	    // First store the id
	    mapped_id = (Integer)compression_map.get( id );
	    if( mapped_id == null ) {
		// Assume that the mapped indices are 0..size().
		mapped_id = new Integer( compression_map.size() );
		if( mapped_id.intValue() >= Integer.MAX_VALUE ) {
		    throw new IndexOutOfBoundsException( "Array.serialize:"
						 + " compression map too large " );
		}
		compression_map.put( id, mapped_id );
	    }
	    buffer.add( mapped_id.intValue() );
	    
	    // Now store the value
	    buffer.add( get(id) );
	}
    }

}
