/**
 * Screen.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.oil;

import java.util.HashMap;
import java.util.Iterator;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.QuickSerializable;
import ostore.util.QSException;
import ostore.util.QSVector;
import ostore.util.Types;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;

/**
 * A model that filters a stream of <code>QuickSerializable</code>
 * objects.  The <code>Screen</code> counts the number of times each of a
 * specified set of keys occurs in the stream.
 * <p>
 * Internally <code>HashMap</code>s are used to store the set of
 * interesting keys.  It is therefore not advised to apply this class to
 * objects whose <code>hashCode</code> or <code>equals</code> methods are
 * expensive to compute.
 * <p>
 * This class is synchronized.
 *
 * @see Model
 *
 * @author	Dennis Geels
 * @version	$Id: Screen.java,v 1.10 2002/07/20 18:44:40 srhea Exp $
 */

public class Screen implements Model
{
    ////////////////////       CONSTANTS	////////////////////

    private static final boolean DEBUG = DebugFlags.OIL_SCREEN;
    private static final String class_tag = "Screen";

    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /*
     * The internal key->index storage;
     */
    private HashMap map;

    /*
     * Store the counts as <code>int</code>s, not <code>Integer</code>s,
     * because they may be highly dynamic.
     */
    private int[] counts;

    /*
     * The count of all matches ever, including for <code>remove</code>d keys.
     */
    private int total;

    /*
     * <code>true</code> iff the <code>Screen</code> should remember the
     * counts for keys that are <code>remove</code>d.
     */
    private boolean remember_all;

    /*
     * Key->count mappings for old counts.  (name from CVS)
     */
    private HashMap attic;

    /*
     * Holds <code>record</code>ed observations until they are processed.
     */
    private QSVector sequence;

    ////////////////////  STATIC CONSTRUCTORS	////////////////////
    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Construct a new <code>Screen</code>.
     */
    public Screen()
    {
	this( true );
    }
    
    /**
     * Constructs a <code>Screen</code> from its
     * <code>serialize</code>d form.
     */
    public Screen( InputBuffer buffer ) throws QSException
    {
	if( DEBUG ) Debug.printtagln( class_tag, "deserializing..." );
	
	if( DEBUG ) Debug.printtagln( class_tag, "done" );

	throw new NoSuchMethodError();
	// FIXME
    }    
    
    /**
     * Construct a new <code>Screen</code>.
     *
     * @param remember <code>true</code> iff the <code>Screen</code> should
     * remember the counts for keys that are <code>remove</code>d.
     */
    public Screen( boolean remember )
    {
	this.map = new HashMap();
	this.counts = new int[10];
	this.total = 0;
	this.remember_all = remember;
	if( remember_all )
	    this.attic = new HashMap();
	this.sequence = new QSVector();
    }
    
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Adds a key to screened.
     * If this <code>Screen</code> remembers a previous count for the
     * specified key, that count shall be restored.
     *
     * @param key	The object to look for in the stream.
     */
    public synchronized void watch( QuickSerializable key )
    {
	final String tag = class_tag+".watch";
	if( DEBUG ) Debug.printtagln( tag, "called:\t" + key );
	
	if( key == null ) {
	    throw new
		NullPointerException( "Screen.watch called with null key" );
	} else if( map.containsKey( key ) ) {
	    if( DEBUG ) Debug.printtagln( tag, "already watching: " + key );
	} else {
	    // Make sure counts[] is big enough
	    int size = map.size();
	    int len = counts.length;
	    if( size >= len ) {
		int new_len = len*2;
		if( DEBUG ) Debug.printtagln( tag, "growing counts to " +
					      new_len );
		int[] new_counts = new int[ new_len ];
		for( int i=0; i<len; i++ ) {
		    new_counts[i] = counts[i];
		}
		counts = new_counts;
	    }
	    // Now add the mapping
	    map.put( key, new Integer( size ) );
	    if( remember_all && attic.containsKey( key ) ) {
		if( DEBUG ) Debug.printtagln( tag, "restoring " + key +
					      " from attic" );
		counts[ size ] = ((Integer)attic.remove( key )).intValue();
	    } else {
		counts[ size ] = 0;
	    }
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }

    /**
     * Removes a key from the screened set.
     * If this <code>Screen</code> has been told to remember old keys, the
     * key and its current count will be stored for later salvaging.
     *
     * @param key	The object to stop looking for.
     */
    public synchronized void remove( QuickSerializable key )
    {
	final String tag = class_tag+".remove";
	if( DEBUG ) Debug.printtagln( tag, "called:\t" + key );
	
	Integer index = (Integer)map.remove( key );
	if( index == null ) {
	    if( DEBUG ) Debug.printtagln( tag, "wasn\'t watching: " + key );	    
	} else if( remember_all ) {
	    if( DEBUG ) Debug.printtagln( tag, "moved to attic: " + key +
					  " -> " + counts[index.intValue()] );
	    attic.put( key, new Integer( counts[index.intValue()] ) );
	} else {
	    if( DEBUG ) Debug.printtagln( tag, "removed: " + key );
	}
	// Reclaim the space in <code>counts</code>.
	int hole = index.intValue();
	int num_to_shift = map.size()-hole;
	if( num_to_shift > 0 ) 
	    System.arraycopy( counts, hole+1, counts, hole, num_to_shift );
			  
	if( DEBUG ) Debug.printtagln( tag, "done" );	    
    }
    
    /**
     * Removes all keys from the screened set.
     * If this <code>Screen</code> has been told to remember old keys, the
     * keys and their current counts will be stored for later salvaging.
     */
    public synchronized void remove_all()
    {
	final String tag = class_tag+".remove_all";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	if( remember_all ) {
	    Iterator keys = map.keySet().iterator();
	    while( keys.hasNext() ) {
		QuickSerializable key = (QuickSerializable)keys.next();
		Integer index = (Integer)map.get(key);
		if( DEBUG ) Debug.printtagln( tag, "moved to attic: " + key +
					      " -> " + counts[index.intValue()] );
		attic.put( key, new Integer( counts[index.intValue()] ) );
	    }
	}
	// Now just remove all mappings.
	map.clear();
	
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }


    /**
     * Returns the current count for the specified key.
     *
     * @param key	The key whose count to return.
     * 
     * @return the number of times <code>key</code> has been processed.
     * If the key was never <code>watch</code>ed, or has since been
     * forgotten, returns 0.
     */
    public synchronized int get_count( QuickSerializable key )
    {
	final String tag = class_tag+".get_count";
	if( DEBUG ) Debug.printtagln( tag, "called:\t" + key );

	int count = 0;
	Integer index = (Integer)map.get( key );
	if( index != null ) {
	    count = counts[index.intValue()];
	    if( DEBUG ) Debug.printtagln( tag, key + "->" + count );
	} else if( remember_all ) {
	    Integer c = (Integer)attic.get( key );
	    if( c != null )
		count = c.intValue();
	    if( DEBUG ) Debug.printtagln( tag, "(attic): " + count );
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
	return count;
    }
    
    /**
     * Returns the current sum of all counts.
     *
     * @return the number of times any currently watched key has been
     * matched.
     */
    public synchronized int get_all_counts()
    {
	final String tag = class_tag+".get_all_counts";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	// The <code>counts</code> array only contains counts for currently
	// watched keys.
	int sum=0;
	int len = map.size();
	for( int i=0; i<len; i++ ) {
	    sum += counts[i];
	}
	
	if( DEBUG ) Debug.printtagln( tag, "done: " + sum );
	return sum;
    }
    
    /**
     * Returns the current total count.
     *
     * @return the total number of observations processed.
     */
    public synchronized int total()
    {
	final String tag = class_tag+".total";
	if( DEBUG ) Debug.printtagln( tag, "called, done" );
	return total;
    }
    
    /**
     * Records an object from the event stream (does not process).
     *
     * @param observation	The object to record
     */
    public synchronized void record( QuickSerializable observation )
    {
	final String tag = class_tag+".record";
	if( DEBUG ) Debug.printtagln( tag, "called:\t" + observation );

	sequence.add( observation );
	    
	if( DEBUG ) Debug.printtagln( tag, "done" );	    
    }
    
    /**
     * Processes a group of objects from the event stream.
     * If an object matches any key in the watched set, increments the
     * proper counters.
     *
     * @param observations	The objects to record
     */
    public synchronized void record( QuickSerializable[] observations )
    {
	final String tag = class_tag+".record";
	if( DEBUG ) Debug.printtagln( tag, "called with " +
				      observations.length + " observations" );

	sequence.add( observations );
	
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }

    /* Specified by ostore.oil.Model interface */
    public synchronized void clear( int num )
    {
	final String tag = class_tag+".clear";
	if( DEBUG ) Debug.printtagln( tag, "called ( " + num + ")" );

	sequence.removeRange( 0, num );
	
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }
    
    /* Specified by ostore.oil.Model interface */
    public synchronized void clear()
    {
	final String tag = class_tag+".clear";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	sequence.removeAllElements();
	
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }

    /* Specified by ostore.oil.Model interface */    
    public synchronized double loglikelihood()
    {
	final String tag = class_tag+".loglikelihood";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	double ll = 0;
	
	// Estimate the probability of each element in the sequence
	int len = sequence.size();
	for( int i=0; i<len; i++ ) {
	    QuickSerializable observation =
		(QuickSerializable)sequence.get(i);
	    int count = get_count( observation );
	    double likelihood = ((double)count)/((double)total);
	    ll += Math.log(likelihood);
	}

	if( DEBUG ) Debug.printtagln( tag, "done: " + ll );
	return ll;
    }

    /**
     * Processes the recent event stream.
     * If an object matches any key in the watched set, this method
     * increments the proper counters.
     */
    public synchronized Model.Delta recalculate()
    {
	final String tag = class_tag+".recalculate";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	// Process each element in the sequence, looking for matches.
	int len = sequence.size();
	for( int i=0; i<len; i++ ) {
	    QuickSerializable observation =
		(QuickSerializable)sequence.get(i);
	    
	    Integer index = (Integer)map.get( observation );
	    if( index != null ) {
		counts[index.intValue()]++;
	    }
	}
	// Update the global (match & miss) count
	total += len;
	
	// Forget the processed events.
	clear();
	
	if( DEBUG ) {
	    // Print out the current counts.
	    Iterator keys = map.keySet().iterator();
	    while( keys.hasNext() ) {
		QuickSerializable key = (QuickSerializable)keys.next();
		int count = get_count(key);
		Debug.printtagln( tag, "count[" + key + "]: " + count );
	    }
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
	
	// FIXME
	//produce a delta?
	return null;
    }
    

    
    /* Specified by ostore.oil.Model interface */    
    public synchronized Model.Prediction predict( int horizon )
    {
	final String tag = class_tag+".predict";
	if( DEBUG ) Debug.printtagln( tag, "called" );


	if( DEBUG ) Debug.printtagln( tag, "done" );
	// FIXME!!
	return null;
    }
      
    /* Specified by ostore.oil.Model interface */    
    public synchronized QuickSerializable[] outliers()
    {
	final String tag = class_tag+".outliers";
	if( DEBUG ) Debug.printtagln( tag, "called" );


	if( DEBUG ) Debug.printtagln( tag, "done" );
	// FIXME!!
	return null;
    }
    /* Specified by ostore.oil.Model interface */
    public synchronized void add_delta( Model.Delta d )
    {
	final String tag = class_tag+".add_delta";
	if( DEBUG ) Debug.printtagln( tag, "called" );


	if( DEBUG ) Debug.printtagln( tag, "done" );
	// FIXME!!
    }

    
    /* Specified by ostore.util.QuickSerializable */
    public synchronized void serialize (OutputBuffer buffer)
    {
	throw new NoSuchMethodError();
	// FIXME
    }
    
    ////////////////////     INNER CLASSES	////////////////////
}
