/**
 * ReplicaQuery.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.replica;

import ostore.util.SecureHash;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.ByteUtils;
import ostore.util.QSException;
import ostore.util.QSInt;
import ostore.tapestry.api.TapestryTag;
import ostore.tapestry.api.TapestryQuery;
import ostore.tapestry.api.TapestryQueryResult;
import ostore.tapestry.api.TapestryQueryResultMatch;
import ostore.tapestry.api.TapestryQueryResultNoMatch;
import ostore.tapestry.api.TapestryQueryState;

/**
 * The Tapestry query sent to find object replicas.
 *
 * @author	Dennis Geels
 * @version	$Id: ReplicaQuery.java,v 1.18 2003/02/23 23:11:41 hweather Exp $
 */

public class ReplicaQuery implements TapestryQuery
{
    ////////////////////       CONSTANTS	////////////////////
    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * The maximum acceptable depth a replica can be from the inner
     * ring, or <code>null</code> if any will do.
     */
    private QSInt max_depth;
    
    /**
     * The maximum acceptable distance (milliseconds) replica can be from
     * the inner ring, or <code>null</code> if any will do.
     */
    private Double max_distance;
    
    ////////////////////  STATIC CONSTRUCTORS	////////////////////
    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Creates a new <code>ReplicaQuery</code>.
     */
    public ReplicaQuery()
    {
	this( (QSInt)null, (Double)null );
    }

    /**
     * Creates a new <code>ReplicaQuery</code> with an upper bound on
     * the depth of replicas in the tree.
     */
    public ReplicaQuery( QSInt max_depth, Double max_distance )
    {
	this.max_depth = max_depth;
	this.max_distance = max_distance;
    }
    
    /**
     * Constructs a <code>ReplicaQuery</code> from its
     * <code>serialize</code>d form.
     */
    public ReplicaQuery( InputBuffer buffer ) throws QSException
    {
	boolean present = buffer.nextBoolean();
	if( present ) {
	    max_depth = (QSInt)buffer.nextObject();
	} else {
	    max_depth = null;
	}
	present = buffer.nextBoolean();
	if( present ) {
	    max_distance = new Double(buffer.nextDouble());
	} else {
	    max_distance = null;
	}
    }
    
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /* Specified by ostore.util.QuickSerializable */
    public void serialize( OutputBuffer buffer )
    {
	if( max_depth == null ) {
	    buffer.add( false );
	} else {
	    buffer.add( true );
	    buffer.add( max_depth );
	}
	if( max_distance == null ) {
	    buffer.add( false );
	} else {
	    buffer.add( true );
	    buffer.add( max_distance.doubleValue() );
	}
    }
    
    /* Specified by ostore.tapestry.api.TapestryQuery */
    public TapestryQueryResult query( SecureHash node_guid,
                                      SecureHash bp_guid,
				      TapestryTag tag,
				      TapestryQueryState state )
    {
	if( tag instanceof ReplicaTag ) {
	    // FIXME  - include distance, fix tags
	    
	    if( (max_depth == null) ||
		(max_depth.intValue() >=
		 ((ReplicaTag)tag).depth.intValue()) ) {

		SecureHash[] m = { bp_guid };
		return new TapestryQueryResultMatch (m, true /* stop */);
	    }
	}
	return new TapestryQueryResultNoMatch ();
    }

    /** Specified by java.lang.Object */
    public String toString()
    {
	String str = new String();

	str += "<ReplicaQuery";
	str += " max_depth=" + max_depth;
	str += " max_distance=" + max_distance;
	str += " " + super.toString();
	str += ">";

	return str;
    }
    
    ////////////////////     INNER CLASSES	////////////////////
}
