/*
 * DynamicTClient.java
 *
 *  Copyright (c) 2002 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl;
import java.util.*;
import java.net.InetAddress;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.SingleThreadedEventHandlerIF;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.StagesInitializedSignal;
import seda.sandStorm.api.ProfilableIF;
import seda.sandStorm.api.NoSuchStageException;
import seda.sandStorm.internal.SinkProxy;
import ostore.network.patchwork.MonitorMsg;
import ostore.network.patchwork.MonitorStopMsg;
import ostore.network.patchwork.Patch;
import ostore.network.patchwork.QuickStatReq;
import ostore.network.patchwork.QuickStatResp;
import ostore.network.patchwork.NodeStatusMsg;
import ostore.util.AckStage;
import ostore.util.Carp;
import ostore.util.Clock;
import ostore.util.AssertionViolatedException;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.SecureHash;
import ostore.util.NodeId;
import ostore.util.NonceAckMsg;
import ostore.util.NonceMsg;
import ostore.util.Pair;
import ostore.util.TypeTable;
import ostore.util.QSException;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.dispatch.Signal;
import ostore.tapestry.api.TapestryNodeDetachMsg;
import ostore.tapestry.api.TapestryDetachConfirm;
import ostore.tapestry.api.TapestryIntegrateMsg;
import ostore.tapestry.api.TapestryNeighborInfo;
import ostore.tapestry.api.TapestryRoutingTableChanged;
import ostore.tapestry.api.TapestryReversePointersChanged;

/**
 * The stage that handles all dynamic behavior of a Tapestry node. Pretty
 * much everything done in Tapestry is here, with the exception of core
 * routing of messages.  This handles node insertion, deletion, requests for
 * diagnostic or debugging info, introduction service, self-repair, and
 * failover to backup routes.
 * 
 * @author     Ben Y. Zhao
 * @version    $Id: DynamicTClient.java,v 1.154 2003/06/06 00:08:58 strib Exp $
 */
public class DynamicTClient extends AckStage
    implements SingleThreadedEventHandlerIF, EventHandlerIF, ProfilableIF {

    //  methods for ProfilableIF
    /* Specified by seda.sandStorm.api.ProfilableIF */
    public int profileSize()
    {
	return my_sink.size();
    }
    //  end methods for ProfilableIF

    private static final String classname = "DynamicTClient";
    private String class_tag = "";
    private static final boolean DEBUG_FT = false;

    public static final int NOT_ATTACHED = 0;
    public static final int INSERT_MCAST_MODE = 1;
    public static final int INSERT_NEARESTN_MODE = 2;
    public static final int ATTACHED = 3;
    public static final int DETACH_ROUTINGOBJ = 4;
    public static final int DETACH_FINALACK = 5;

    private static final double UNKNOWN_PING_VAL = 1000.0;

    private static final Integer APING_REINIT_NN = new Integer(0);
    private static final Integer APING_MAO = new Integer(1);

    private static final long SHORT_TIMEOUT = 30000;
    private static final long LONG_TIMEOUT = 60000;
    private static final long PING_TIMEOUT = 10000;
    private static final long REPAIRREQ_TIMEOUT = 60000;

    //times you will try to contact the gateway node after timeout
    private static final int GATEWAY_TIMEOUT_RETRIES = 30;

    // user data constants
    private static final Integer USERDATA_GATEWAY = new Integer(1000);
    private static final Integer USERDATA_PTR = new Integer(1001);
    private static final Integer USERDATA_NOTIFY = new Integer(1002);
    private static final Integer USERDATA_SEED = new Integer(1003);
    private static class DMCastUserData {
	public int level,digit,entry;
	public NodeId peer;
	public SecureHash newId;
	public DynamicDMCastMsg msg;
	public int ack_nonce;
	public DMCastUserData( DynamicDMCastMsg msg, int level,
			       int digit, int entry ) {
	    this.level = level;
	    this.digit = digit;
	    this.entry = entry;
	    this.msg = msg;
	    this.peer = msg.peer;
	    this.newId = msg.newGuid;
	}
	public DMCastUserData( NodeId peer, SecureHash newId, int level,
			       int digit, int entry ) {
	    
	    this.peer = peer;
	    this.newId = newId;
	    this.level = level;
	    this.digit = digit;
	    this.entry = entry;
	}
	public boolean equals( Object other ) {
	    if( !(other instanceof DMCastUserData) ) {
		return false;
	    }
	    DMCastUserData rhs = (DMCastUserData) other;
	    return rhs.peer.equals(peer) && 
		rhs.newId.equals(newId);
	}
	public int hashCode() {
	    return peer.hashCode() ^ newId.hashCode();
	}
	public String toString() {
	    return "(DMCastUserData " + peer + "/" + newId + ")"; 
	}
    }

    private class NonceMsgDelay implements QueueElementIF {
	public NonceMsg nm;
	public Object ud;
	public long timeout;
	public NonceMsgDelay( NonceMsg nm, long timeout, Object ud ) {
	    this.nm = nm;
	    this.ud = ud;
	    this.timeout = timeout;
	}
    }

    /**
     * Dispatch a list of messages
     */
    private void dispatch (LinkedList messages) {
	final String tag = class_tag + ".dispatch";
	ListIterator list = messages.listIterator(0);
	QueueElementIF item;

	while (list.hasNext()) {
	    item = (QueueElementIF) list.next();
	    dispatch( item );
	}
    }

    public DynamicTClient() throws Exception {
	DEBUG = DebugFlags.TAPESTRY_DYNAMIC_TCLIENT; 
	tag = "DynamicTClient";

	String [] event_types = {
	    "seda.sandStorm.api.StagesInitializedSignal",
	    "ostore.tapestry.api.TapestryNodeDetachMsg",
	    "ostore.tapestry.api.TapestryIntegrateMsg",
	    "ostore.tapestry.impl.NodeGuid",
	    "ostore.tapestry.impl.FRLSChangeMsg",
	    "ostore.network.patchwork.QuickStatResp",
	    "ostore.network.patchwork.NodeStatusMsg",
	    "ostore.tapestry.impl.PeriodicReminder",
	    "ostore.tapestry.impl.DynamicTClient$NonceMsgDelay",
	    "ostore.tapestry.impl.ModifyObjLevelsMsg"
	};
	this.event_types = event_types;
	String [] inb_msg_types = {
	    //Dynamic Messages follow
	    "ostore.tapestry.impl.DynamicDMAckMsg",
	    "ostore.tapestry.impl.DynamicDMCastMsg",
	    "ostore.tapestry.impl.DynamicGatewayMsg",
	    "ostore.tapestry.impl.DynamicMapMsg",
	    "ostore.tapestry.impl.DynamicMCReqMsg",
	    "ostore.tapestry.impl.DynamicNotifyMsg",
	    "ostore.tapestry.impl.DynamicObjMapMsg",
	    "ostore.tapestry.impl.DynamicObjReqMsg",
	    "ostore.tapestry.impl.DynamicSeedMapMsg",
	    "ostore.tapestry.impl.DynamicSeedMsg",
	    "ostore.tapestry.impl.DynamicSurrogateMsg",
	    "ostore.tapestry.impl.DynamicBeginDelete",
	    "ostore.tapestry.impl.DynamicEndDelete",
	    "ostore.tapestry.impl.DynamicPtrMsg",
	    "ostore.tapestry.impl.PartialRepublish",
	    "ostore.tapestry.impl.DebugMsg",
	    "ostore.tapestry.impl.DynamicRepairReqMsg",
	    "ostore.tapestry.impl.DynamicRepairRespMsg",
	    "ostore.tapestry.impl.PeriodicRoutePatch"
	};
	this.inb_msg_types = inb_msg_types;
	String [] outb_msg_types = {
	};
	this.outb_msg_types = outb_msg_types;
    }

    /**
     * Initialization
     */
    public void init (ConfigDataIF config) throws Exception {

	super.init( config );

	String stageName = config.getStage().getName();

	try {
	    if(config.getBoolean("Profile")) {
		config.getManager().getProfiler().add("ostore.tapestry." + 
						      "impl.DynamicTClient", 
						      this);
		SinkIF sink =
		    config.getManager().getStage(stageName).getSink();
		my_sink = new SinkProxy(sink, config.getManager(),
				      config.getStage().getWrapper());
	    } 
	} catch (NoSuchStageException e) {
	    Carp.die ("DynamicTClient.init" + ": could not get own sink.");
	}

	class_tag = classname+"("+my_node_id+")";
	this.tag = class_tag;
	final String tag = class_tag + ".init";
	Debug.turnOn (class_tag);

	if(config.getBoolean("die_gracefully"))
	    Carp.setThrowException(true);

	boolean disablemonitor = config.getBoolean("disable_monitor");
	if (disablemonitor) 
	  MONITORING = false;
	boolean ignore_status_msgs = config.getBoolean("ignore_status_msgs");
	if (ignore_status_msgs)
	  LISTEN_STATUS = false;
	boolean disableperiodicrepair = 
	  config.getBoolean("disable_periodic_repair");
	if (!disableperiodicrepair) {
	  classifier.dispatch_later(new PeriodicReminder(my_node_id,
							 PeriodicReminder.ONE),
				    100000);
	}

	_dynamic_route = config.getString("dynamic_route");
	String[] gateway = config.getStringList("gateway");
	if (gateway != null) {
	    _gate_node_id = new NodeId[gateway.length];
	    for( int i = 0; i < gateway.length; i++ ) {
		if( DEBUG ) {
		    Debug.printtagln( tag, "Found gateway: " + gateway[i] );
		}
		_gate_node_id[i] = new NodeId(gateway[i]);
	    }
	}

	deadEntries = new HashMap();

	// make the ping cache before all_pairs_pings
	_ping_cache = new HashMap();
	_fulltest = config.getBoolean("fulltestmember");
	allpairspingsfile = config.getString("all_pairs_pings_file");
	simulator = config.getBoolean("simulator_running");
	if (simulator) {
	    net = soss.network.Network.instance();
	    allpairspingsfile = null;
	    //if using simulator, then ignore all pairs ping file
	}

	if (allpairspingsfile != null)
	    parse_all_pings(allpairspingsfile);

	Filter f = new Filter ();

	if (_dynamic_route.equals("static")) {
	    f = new Filter();
	    String rtaname = "ostore.tapestry.impl.RoutingTable$Available";
	    if (!f.requireType(Class.forName (rtaname)))
		BUG (tag + ": could not require type "+
		     "ostore.tapestry.impl.RoutingTable$Available");
	    classifier.subscribe (f, my_sink);
	}

	if (_dynamic_route == null) {
	    BUG (tag+ ": missing dynamic_route config argument");
	} else if (_dynamic_route.equals("static")) {
	    f = new Filter ();
	    String romname = "ostore.tapestry.impl.Router$ObjectMap";
	    if (! f.requireType(Class.forName(romname)))
		BUG (tag + ": could not require type " + 
		     "ostore.tapestry.impl.Router$ObjectMap");
	    classifier.subscribe (f, my_sink);
	}	  
	
	// make this here, in case someone tries using it early
	_movingObjs = new HashMap();

    }

    /**
     * A single node was removed from the routing table
     */
    private void send_routing_table_remove( NodeId node, SecureHash guid,
					    double rtt ) {
	TapestryNeighborInfo[] removed = { new TapestryNeighborInfo( node, 
								     guid, 
								     rtt ) };
	dispatch( new TapestryRoutingTableChanged( null, removed ) );
    }

    /**
     * A single node was added to the reverse pointers
     */
    private void send_reverse_ptr_add( NodeId node, SecureHash guid,
				       double rtt ) {
	TapestryNeighborInfo[] added = { new TapestryNeighborInfo( node, 
								     guid, 
								     rtt ) };
	dispatch( new TapestryReversePointersChanged( added, null ) );
    }

    /**
     * A single node was removed from the reverse pointers
     */
    private void send_reverse_ptr_remove( NodeId node, SecureHash guid,
					  double rtt ) {
	TapestryNeighborInfo[] removed = { new TapestryNeighborInfo( node, 
								     guid, 
								     rtt ) };
	dispatch( new TapestryReversePointersChanged( null, removed ) );
    }

    /**
     * Got a StagesInitSignal
     * Do nothing for now, wait for NodeGuid
     */
    protected void handle_stages_init_signal () {
	final String tag = class_tag + ".handle_stages_init_signal";

	if (DEBUG) Debug.printtagln (tag,
				     "processing StagesInitializedSignal. " +
				     " do nothing until NodeGuid recv'd.");
    }

    /**
     * Someone else (probably the router) told us our NodeGuid.  Now we
     * can do some real initialization and start integration, if we 
     * are dynamic.
     */
    protected void handle_node_guid (NodeGuid local_node_guid) {
	final String tag = class_tag + ".handle_node_guid";

	if (DEBUG) Debug.printtagln (tag,
				     "processing NodeGuid" );

	_self_guid = local_node_guid.node_guid;
	_self_digits = RoutingTable.digits (_self_guid);
	if(DEBUG) Debug.printtagln(tag, "called integration algorithm" +
				   " for node " + 
				   RoutingTable.printGuid(_self_guid) + " " + 
				   my_node_id);
	if (_dynamic_route.equals("dynamic")) {
	    handle_integrate( new TapestryIntegrateMsg() );
	} else {
	    // want to initialize these, even in the static case
	    _objlevels = new HashMap();
	    _waiting_notifies = new HashMap();
	    _out_times = new HashMap();
	    _currentList_pings = new HashMap();
	}
    }

    /**
     * Start the dynamic integration process by reinitializing data structures
     * and sending a message to the gateway.
     */
    private void handle_integrate(TapestryIntegrateMsg msg) {
	final String tag = class_tag +".handle_integrate";
	if (DEBUG) Debug.printtagln(tag, "Processing integrate, gateway="+
				    msg.gateway);
	_dynamic_route = "dynamic";
      
	if (msg.gateway != null)
	    _gate_node_id = msg.gateway;
	if (_gate_node_id == null)
	    Carp.die("Unexpected error: NO valid gateway set!");

	//reinitialize the data structures for insertion process
	//	_waiting_pings = new HashMap();
	_currentList_pings = new HashMap();
	sync = false;
	initList = new TreeSet();
      
	//reinitialize useful result data structures
	_routing_tbl = new RoutingTable(my_node_id, _self_guid, classifier);
	_routing_tbl.setSink( my_sink );
	_routing_tbl.addSelf(_self_guid);

	_route_bitmap = new BitSet(_routing_tbl.digit_values());
	_backpointers = new HashMap();
	_revpointers = new HashMap();

	//reinitializing other variables
	status = NOT_ATTACHED;
	_objlevels = new HashMap();
	_waiting_notifies = new HashMap();
	_out_times = new HashMap();

	// ping_cache = new HashMap(); cache doesn't need cleaning out
	timer = Clock.current_date(my_node_id).getTime();
	status = INSERT_MCAST_MODE;

	if( !_gate_node_id[0].equals( my_node_id ) ) {
	    send_gateway_msg(0);
	} else {
	    // if we are the gateway, we're pretty much done
	    // because there's no one to contact
	    _routing_tbl.addSelf( _self_guid );
	    dispatch(new Router.ObjectMap(_backpointers));
	    finished_integration();
	}
    }

    /**
     * Send a DynamicGatewayMsg.
     *
     * @param position which gateay to use (mod the number of gateways)
     */
    private void send_gateway_msg( int position ) {
	NodeId gateway = _gate_node_id[position % _gate_node_id.length];
	DynamicGatewayMsg dgm = new DynamicGatewayMsg(gateway, 
						      _self_guid);
	// need to keep track of which surrogate we sent it to,
	// as well as the number of times around the list we've gone
	Pair userData = new Pair( USERDATA_GATEWAY, new Integer(position) );
	// expect a response from the surrogate
	dgm.surr_nonce = forceRegisterNonce( LONG_TIMEOUT, userData );
	sendNonceMsg( dgm, SHORT_TIMEOUT, userData );
    }

    /**
     * Got the routing table, update data structures
     */
    private void handle_routing_table_available (RoutingTable.Available msg) {

	final String tag = class_tag + ".handle_routing_table_available";
	if (DEBUG) {
	    Debug.printtagln (tag, "processing RoutingTable.Available" );
	    if (_dynamic_route.equals("dynamic")) {
		Debug.printtagln(tag, "unexpected RoutingTable.Available");
	    }
	}

	//if we sent this out (reintegration), then ignore
	if (_dynamic_route.equals("dynamic")) return;

	if (_routing_tbl != null) 
	    throw new AssertionViolatedException(tag + 
				       ": got two versions of routing table.");

	_routing_tbl = msg.table;
	_routing_tbl.setSink( my_sink );
	_revpointers = msg.revpointers;
	_objlevels = msg.objlevels;
	if (_dynamic_route.equals("static")) {
	    System.out.println("Tapestry: ready.\n");
	    System.out.flush();
	}
	send_monitor_msgs();
    }

    /**
     * Got the object map, update _backpointers
     */
    private void handle_object_map(Router.ObjectMap msg) {
	final String tag = class_tag+".handle_object_map";
	if (DEBUG) Debug.printtagln(tag, "processing ObjectMap "+msg);

	//if we sent this out for reintegration, then ignore
	if (_dynamic_route.equals("dynamic")) return;
	if(_backpointers != null) 
	    throw new AssertionViolatedException(tag+
					    ": got 2 versions of object map.");
	_backpointers = msg.objmap;
    }

    /**
     * Take responsibility to be someone's gateway, and contact the surrogate
     */
    private void handle_gateway( DynamicGatewayMsg msg ) {

	final String tag = class_tag + ".handling DynamicGatewayMsg";

	if (DEBUG) {
	    Debug.printtagln(tag, "Handling DynamicGatewayMsg: "+msg);
	}
	
	// if this is an error, we are using a different base from 
	// our gateway
	if( msg.error ) {
	    Carp.die( "We are using a different routing table base " +
		      " from our gateway " + msg.peer + ": (" +
		      msg.digits_per_guid + ", " + msg.digit_values + 
		      ", " + msg.bits_per_digit + ")" );
	}
	
	// if the routing table base of the new guy doesn't match,
	// send an error back.
	if( msg.digits_per_guid != RoutingTable.DIGITS_PER_GUID ||
	    msg.digit_values != RoutingTable.DIGIT_VALUES ||
	    msg.bits_per_digit != RoutingTable.BITS_PER_DIGIT ) {
	    
	    if( DEBUG ) {
		Debug.printtagln( tag, "Rejecting Gateway msg because " +
				  " the routing table base doesn't match." );
	    }
	    DynamicGatewayMsg dgm = new DynamicGatewayMsg( msg.peer, 
							   _self_guid );
	    dgm.error = true;
	    // we don't need this one to be reliable, so just dispatch
	    dispatch( dgm );
	    return;
	}
	
	DynamicSurrogateMsg dsm = 
	    new DynamicSurrogateMsg(msg.guid, msg.peer, false);
	dsm.nonce = msg.surr_nonce;
	dsm.stage_id = msg.getStageId();
	dispatch(dsm);

    }
    
    /**
     * Send DynamicMapMsgs out to everyone below a specified level
     */
    private boolean send_dynamic_map_msgs( NodeId from, int maxLevel ) {
	final String tag = class_tag + ".send_dynamic_map_msgs";
	ArrayList al;
	NodeId entry;
	DynamicMapMsg route_level;
	LinkedList messages = new LinkedList();
	boolean stop;
	boolean anysent = false;
	int j;
	for (j=0; j<=maxLevel; j++) {
	    //go thru routing table, and send all interesting levels to new guy
	    stop = true;
	    al = new ArrayList(_routing_tbl.digit_values());
	    for (int i=0; i<RoutingTable.DIGIT_VALUES; i++) {
		entry = _routing_tbl.read (j, i);
		if (DEBUG) Debug.printtagln(tag, "hop="+j+
					    ", add GA entry="+entry);
		if (entry == null) al.add(new GuidAddr(null, my_node_id));
		else {
		    al.add(new GuidAddr(_routing_tbl.node_id_to_guid (entry), 
					entry));
		    if (!entry.equals(my_node_id)) stop = false;
		}
	    }
	    if (stop) break;
	    messages.add(new DynamicMapMsg(from, _self_guid, al, j, 0));
	    anysent = true;
	}
	for (int i=0; i<messages.size(); i++) {
	    ((DynamicMapMsg) messages.get(i)).total = j;
	}
	dispatch(messages);
	return anysent;
    }

    /**
     * Send a DynamicNotify during a multicast
     */
    private void send_mc_dynamic_notify( NodeId from, SecureHash newID, 
					 WatchList wl, int alpha ) {

	//create the watchlist, then use it to send first order msgs to new guy
	DynamicNotifyMsg dnm;
	TreeSet ll = wl.checkAndSend(newID, _routing_tbl);
	wl.prune();  //prune out levels which are all filled out
	if (!ll.contains(my_node_id))
	    ll.add(my_node_id);
	if (!ll.isEmpty()) {
	    //now want to send DynamicNotify to tell new guy to fill in hole
	    //HERE  should it be msg.alpha -1 instead?
	    dnm = new DynamicNotifyMsg(from, _self_guid, newID,
				       DynamicNotifyMsg.FILLHOLE_TYPE, 
				       alpha-1);
	    dnm.list = new GuidAddr [ll.size()];
	    SecureHash guid;
	    NodeId nid;
	    Iterator li = ll.iterator();
	    for (int i=0; li.hasNext(); i++) {
		nid = (NodeId) li.next();
		guid = (SecureHash) _routing_tbl.node_id_to_guid(nid);
		dnm.list[i] = new GuidAddr(guid, nid);
	    }
	    sendNonceMsg(dnm, SHORT_TIMEOUT, null);
	    //note: sending FILLHOLE_TYPE also builds the initList
	} else {
	    dnm = new DynamicNotifyMsg(from, _self_guid, newID,
				       DynamicNotifyMsg.DMCAST, alpha-1);
	    sendNonceMsg(dnm, SHORT_TIMEOUT, null);
	}

    }

    /**
     * Add a new node (with unknown alpha) to the routing table.
     */
    private void add_node_to_rt( NodeId newAddr, SecureHash newID, 
				 double latency ) {
	add_node_to_rt( newAddr, newID, 0, latency );
    }

    /**
     * Add a new node (with unknown ping time) to the routing table.
     */
    private void add_node_to_rt( NodeId newAddr, SecureHash newID, 
				 int alpha ) {
	add_node_to_rt( newAddr, newID, alpha, UNKNOWN_PING_VAL );
    }

    /**
     * Add a new node (with known ping time) to the routing table.
     */
    private void add_node_to_rt( NodeId newAddr, SecureHash newID, 
				 int alpha, double latency ) {

	LinkedList changes = _routing_tbl.optimize(latency, newAddr, newID, 
						   alpha);
	int [] digits = RoutingTable.digits(newID);
	send_remove_pointers();
	int changed_level;
	if (changes != null && !changes.isEmpty()) {
	    for (Iterator it = changes.iterator(); it.hasNext(); ) {
		//adding this guy to our routing tables, he needs to know us
		//as a revpointer, so send him a DynamicPtrMsg
		changed_level = ((Integer) it.next()).intValue();
		if (DEBUG_FT)
		  Debug.printtagln(tag, _routing_tbl.toShortString()+
				   "add_node_to_rt");
		send_dynamic_ptr( newAddr, changed_level,
				  digits[changed_level]);
	    }
	}

    }

    private Object getLockIndex( SecureHash newID ) {
	int maxmatch = Router.guidCompare( _self_guid, newID );
	int firstdiff;
	if( maxmatch != Router.LOCALROOTOBJS ) {
	    firstdiff = RoutingTable.digits(newID)[maxmatch];
	} else {
	    firstdiff = -1;
	}
	return new Pair( new Integer(maxmatch), 
			 new Integer(firstdiff) );
    }

    /**
     * Do all the multicast stuff, return whether or not any multicast 
     * messages were sent
     */
    private boolean do_multicast( NodeId newAddr, SecureHash newID, 
				  NodeId parent, int alpha, WatchList wl,
				  boolean setNotified ) {
	
	final String tag = class_tag + ".do_multicast";
	
	//String lockindex = (alpha+1)+":"+
	//    RoutingTable.digits(newID)[alpha+1];
	Object lockindex = getLockIndex( newID );
	if (DEBUG) 
	    Debug.printtagln(tag, "entered multicast with alpha="+alpha+
			     " guid="+newID);

	//add our node state for this entry
	NodeState ns =
	    _mcast.addNodeState(newID, alpha, parent, newAddr, 
				new TreeSet());
	if (DEBUG)
	    Debug.printtagln(tag, "Just added nodestate in surr for guid="
			     + newID);

	//take the current list of required locks for this entry
	TreeSet templocks = (TreeSet) _mcast.indexToLocks.get(lockindex);
	if (templocks != null && !templocks.isEmpty())
	    _mcast.addLocks(newID, (TreeSet) templocks.clone());
      
	//TODO: check the order of these actions to make sure it's correct
	//add this node to the list of locks required for this level
	if (templocks == null) {
	    templocks = new TreeSet();
	    _mcast.indexToLocks.put(lockindex, templocks);
	}
	templocks.add(newID);
	if( DEBUG ) {
	    Debug.printtagln( tag, "added " + newID + 
			      " to lock index " + lockindex );
	}

	if( setNotified ) {
	    //we do not need to set this, since we've made sure we send out 
	    //ourselves as part of the checkAndSend dynamicNotify.FILLHOLE, 
	    //and the dnm handler recognizes us as a special node and replies 
	    //with a dynamicping.INSERTION, which then causes its handler to 
	    //set this. BUT we keep this here, to allow faster insertion, 
	    //and make sure handler for dynamicping.FILLHOLE ignores if 
	    //nodestate is null
	    ns.notified = true;
	}
      
	boolean sentMCast = false;
	NodeId addr;
	DynamicDMCastMsg dmcm;
	for (int j = alpha; j<RoutingTable.DIGITS_PER_GUID; j++) {
	    for (int i=0; i<RoutingTable.DIGIT_VALUES; i++) {
		addr = _routing_tbl.read(j, i);
		if (addr != null && !addr.equals(newAddr) &&
		    !addr.equals(my_node_id)) {
		    //only send mc to actual non-null entries not equal to us
		    dmcm = 
			new DynamicDMCastMsg(addr, _self_guid, my_node_id,
					     newID, newAddr, wl, j+1, 
					     false);
		    //false to indicate it's NOT from a lock
		    _mcast.addMCast(newID, addr);
		    ns.addMCast(addr);
		    if (DEBUG) {
			Debug.printtagln(tag, "J="+j+", I="+i+
					 " Sending out DMCast: "+dmcm);
		    }
		    sentMCast = true;
		    DMCastUserData ud = new DMCastUserData( dmcm, j, i, 0 );
		    // now, let it know you expect an DynamicDMAck back later
		    ud.ack_nonce = forceRegisterNonce( LONG_TIMEOUT, ud );
		    sendNonceMsg(dmcm, SHORT_TIMEOUT, ud);
		}
	    }
	}

	if (!sentMCast) {
	    //BUGFIX
	    //means we didn't actually send out any mcasts, 
	    //so no acks are coming
	    //remove the lock now
	    templocks.remove(newID);
	}


	//now go through list of locks and multicast to them
	DynamicDMCastMsg lockdmm;
	for (Iterator it=templocks.iterator(); it.hasNext(); ) {
	    SecureHash itguid = (SecureHash) it.next();
	    if (itguid.equals(newID))
		continue;
	    if( DEBUG ) {
		Debug.printtagln( tag, "Getting nodestate for " + itguid );
	    }
	    NodeState nsit = _mcast.getNodeState(itguid);
	    if( nsit == null ) {
		if( DEBUG ) {
		    Debug.printtagln( tag, "No nodestate for " + itguid + 
				      " when lock-iterating, ignoring." );
		}
		// remove from the locks list and go to the next one
		it.remove();
		continue;
	    }
	    NodeId itaddr = nsit.addr;
	    lockdmm = new DynamicDMCastMsg(itaddr, _self_guid, my_node_id, 
					   newID, newAddr, wl, alpha+1, 
					   true);
	    //true to indicate this comes from a lock
	    if (DEBUG)
		Debug.printtagln(tag, "Sending mcast on lock, msg="+lockdmm);
	    DMCastUserData ud = new DMCastUserData( lockdmm, 
						    alpha, -1, 0 );
	    ud.ack_nonce = forceRegisterNonce( LONG_TIMEOUT, ud );
	    sendNonceMsg(lockdmm, SHORT_TIMEOUT, ud);
	    sentMCast = true;
	}

	return sentMCast;
    }

    private void mc_move_objects( NodeId newAddr, SecureHash newID ) {

	final String tag = class_tag + ".mc_move_objects";

	//now that we've sent the multicasts, we need to look at the locally
	//rooted objects to consider which ones need to be moved over
	
	//go through list of objects and examine which need to
	//be moved to new surrogate, then put them into
	//movingObjs with the new node's guid as index
	int [] olddigits = RoutingTable.digits(_self_guid);
	int [] newdigits = RoutingTable.digits(newID);
	SecureHash objguid;
	DynamicObjMapMsg domm;
	if (DEBUG) {
	    if (_backpointers == null)
		Debug.printtagln(tag, "_backpointers is null");
	    else if (_backpointers.keySet() == null) 
		Debug.printtagln(tag, "_backpointers.keySet() is null");
	}

	HashSet rootobjs = null;
	if (_objlevels != null) {
	    rootobjs=
		(HashSet) _objlevels.get(_local_ptr_index);
	}
	if ((rootobjs != null) && !rootobjs.isEmpty() &&
	    _backpointers != null && (!_backpointers.isEmpty())) {
	    //only look to move objects IFF backpointers is not null
	    TreeSet l = new TreeSet();
	    for(Iterator it = rootobjs.iterator(); it.hasNext();) {
		objguid = (SecureHash) it.next();
		if (moveObj(olddigits, newdigits, objguid)) {
		    l.add(objguid);
		}
	    }
	    //make sure we put in data structure before sending msgs
	    if (!l.isEmpty())
		_movingObjs.put(newAddr, l);
	    int c=1;
	    for (Iterator it = l.iterator(); it.hasNext();) {
		objguid = (SecureHash) it.next();
		domm = new 
		    DynamicObjMapMsg(newAddr, my_node_id, objguid,
				     (LinkedList) _backpointers.get(objguid),
				     l.size(), c);
		dispatch(domm);
		if (DEBUG)
		    Debug.printtagln(tag, "moving object " + objguid + 
				     " to new node" + newAddr + ".");
		c++;
	    }
	}

    }

    /**
     * You are someone's surrogate.  Start integration (multicast, etc.)
     */
    private void handle_dynamic_surrogate_msg(DynamicSurrogateMsg msg) {
	//need to send back entire routing level plus backpointers
	//we make a copy of the revpointers map, then go level by level, and
	//send pointers while deleting duplicates from the revpointers map
	final String tag = class_tag + ".handle_dynamic_surrogate_msg";

	// ack the surrogate message (not automatic since it's 
	// a RouteMsg
	forceSendAck( msg.from, msg.nonce, msg.stage_id );

	// could happen, especially with timeouts and retries
	if( _self_guid.equals( msg.newID ) ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Got a surrogate message for myself." );
	    }
	    return;
	}

	//after we've sent back a routingtable to the new guy, we
	//should start the acknowledged multicast
	//we will be the multicast root
	if (_mcast == null) 
	    _mcast = new MulticastState(_self_guid);


	if( _mcast.getNodeState( msg.newID ) != null ) {
	    if (DEBUG) {
		Debug.printtagln(tag, "Already doing an mcast for " + 
				 msg.newID + ", ignoring." );
	    }
	    return;
	}

	// send the dynamic map msgs
	send_dynamic_map_msgs( msg.from, Router.guidCompare(_self_guid, 
							    msg.newID));

	int alpha = _mcast.calcAlpha(msg.newID, _self_guid);
	if( DEBUG ) {
	    Debug.printtagln( tag, "Alpha between " + msg.newID + 
			      " and " + _self_guid + " is " + alpha ); 
	}
	

	//add hash entry from new node's guid to this nodeid
	_mcast.addMCAddr(msg.newID, msg.from);

	//the first time we run the watchlist, it should start from
	//the first routing level, and go all the way up to include
	//alpha
	WatchList wl = new WatchList(msg.newID, 0, alpha+1);

	// send a multicast dynamic notify
	send_mc_dynamic_notify( msg.from, msg.newID, wl, alpha );

	//want to add this node to our routing table
	add_node_to_rt( msg.from, msg.newID, alpha );

	boolean sentMCast = do_multicast( msg.from, msg.newID, null, 
					  alpha, wl, true );
	mc_move_objects( msg.from, msg.newID );

	//BUGFIX
	if (!sentMCast) {
	    if (DEBUG) Debug.printtagln(tag, "Taking shortcut, no " + 
					"DMCast messages sent. Insertion " + 
					"DONE, RoutingTable="+_routing_tbl);
	    //whenever we send up the ack, we should also make sure that 
	    //we check to add this guy to our routing table. We do this
	    //by faking a notifymsg (type=GENERAL_TYPE) sent from the new node.
	    DynamicNotifyMsg newdnm = 
		new DynamicNotifyMsg(msg.from, msg.newID, _self_guid,
				     DynamicNotifyMsg.GENERAL_TYPE, alpha-1);
	    handle_dynamic_notify(newdnm, true);
	    //no mcasts were sent, just shortcut to MCReq msg
	    if (DEBUG)
		Debug.printtagln(tag, "Cleaning slate in DMSurr for guid:"+
				 msg.newID);
	    _mcast.cleanSlate(msg.from, msg.newID);
	    /**
	    classifier.dispatch_later(new DynamicMCReqMsg(msg.from, 
							   _self_guid,
						  DynamicMCReqMsg.CONFIRM_TYPE,
							   alpha), 50);
	    */
	    NonceMsgDelay delay = 
		new NonceMsgDelay( new DynamicMCReqMsg(msg.from, 
						       _self_guid,
						  DynamicMCReqMsg.CONFIRM_TYPE,
						       alpha),
				   SHORT_TIMEOUT, new Integer(0) );
	    classifier.dispatch_later( delay, 50 );
	}
      
    }

    /**
     * Got back a list of nodes in a map, take care of it
     */
    private void handle_dynamic_map( DynamicMapMsg msg ) {
	final String tag = class_tag + ".handling DynamicMapMsg";
	
	if (DEBUG) Debug.printtagln(tag, "Handling DynamicMapMsg: "+ msg);
	
	ArrayList al = msg.route_level;
	GuidAddr ga;
	//add the entries from this message into our routing table
	SecureHash guid;
	for (int i=0; i<al.size(); i++) {
	    ga = (GuidAddr) al.get(i);
	    guid = ga.getGuid();
	    if (guid != null) {
		boolean written =
		    _routing_tbl.write(msg.sequence, i, ga.getAddr(), 
				       ga.getGuid(), 
				       new Double(UNKNOWN_PING_VAL));
		if (written) {
		    send_remove_pointers();
		    if (DEBUG_FT)
		      Debug.printtagln(tag, _routing_tbl.toShortString()+
				   "handle_dynamic_map");
		    send_dynamic_ptr( ga.getAddr(), msg.sequence, i );
		}
	    }
	}
    }

    /**
     * Got a confirmation that send objects were received
     */
    private void handle_dynamic_obj_req( DynamicObjReqMsg msg ) {
	final String tag = class_tag + ".handling DynamicObjReqMsg";

	if (DEBUG) Debug.printtagln(tag, "Handling DynamicObjReqMsg: "+ msg);
	
	if (msg.msgtype == DynamicObjReqMsg.CONFIRM) {
	    //msg to confirm that sent objects were received
	    if (_movingObjs.containsKey(msg.peer)) {
		TreeSet objs = (TreeSet) _movingObjs.get(msg.peer);
		//if _movingObjs contains a list of objects for this address
		//then we try to remove this acked object. if the resulting
		//list is empty, then we remove it from _movingObjs
		if (objs == null) {
		    if (DEBUG) 
			Debug.printtagln(tag, "No moveObjs list found for: "
					 +msg.peer+
					 "ignore DynamicObjReqMsg.CONFIRM.");
		} else if (objs.remove(msg.obj)) {
		    if (objs.isEmpty()) {
			_movingObjs.remove(msg.peer);
		    }

		    // remove it from the local root level and
		    // move it to more reasonable levels
		    HashSet level = (HashSet) _objlevels.get(_local_ptr_index);
		    if (level != null) 
			level.remove(msg.obj);

		    Integer i = 
			new Integer( Router.guidCompare( msg.src, 
							 _self_guid ) );
		    modify_objlevels( msg.obj, i, true );
		    if( i.intValue() != -1 ) {
			int j = Router.guidCompare(_self_guid, msg.obj );
			if( j != i.intValue() ) {
			    for( ; j < i.intValue(); j++ ) {
				modify_objlevels( msg.obj, new Integer(j), 
						  true );
			    }
			}
		    }
		    /**
		     * Don't remove the object (TODO: soft state should 
		     * time it out eventually)
		    //TODO: removed obj, need to tell router to be aware
		    //first check if this is a local object, 
		    //if so, then keep a copy of it
		    LinkedList ll = (LinkedList) _backpointers.remove(msg.obj);
		    
		    if (ll != null) {
			PublishInfo keepcopy = null;
			for (Iterator it = ll.iterator(); it.hasNext(); ) {
			    PublishInfo pi = (PublishInfo) it.next();
			    if (pi.src.equals(_self_guid))
				keepcopy = pi;
			}
			if (keepcopy != null) {
			    ll = new LinkedList();
			    ll.add(keepcopy);
			    _backpointers.put(msg.obj, ll);
			} else {
			    HashSet level;
			    for (int k=0; k<RoutingTable.DIGITS_PER_GUID; 
				 k++) {
				level = 
				    (HashSet) _objlevels.get(new Integer(k));
				if (level != null) {
				    level.remove(msg.obj);
				}
			    }
			    level = (HashSet) _objlevels.get(_local_ptr_index);
			    if (level != null) 
				level.remove(msg.obj);
			}
		    } else if (DEBUG) {
			Debug.printtagln(tag, "Moved object confirmation " + 
					 "on obj already moved elsewhere. " + 
					 "Happens in parallel insertion.");
		    }
		    */    
		} else if (DEBUG)
		    Debug.printtagln(tag, "Moved object confirmation " + 
				     "on object we don't have!!");
		
	    } else {
		if (DEBUG) Debug.printtagln(tag, "Moved object confirmation " 
					    + "on object we don't have!!");
	    }
	} else {
	    //we currently do not support any other types of DynamicObjReq
	    //yet, so output an error message
	    System.err.println("ERROR: unexpected DynamicObjReqMsg received.");
	}
	
    }

    /**
     * Put in or remove a guid form a certain level of the objlevels 
     * data structure
     */
    private void modify_objlevels( SecureHash guid, Integer level, 
				   boolean publish ) {

	HashSet levelset = (HashSet) _objlevels.get(level);
	if (publish) {
	  if (levelset == null) {
	    levelset = new HashSet();
	    _objlevels.put(level, levelset);
	  }
	  if (DEBUG) {
	      Debug.printtagln(tag, "Publishing obj, level="+level+
			       ", adding guid:"+
			       RoutingTable.printGuid(guid)+
			       " to objlevel"); 	  
	  }
          levelset.add(guid);
	} else {
	  if (levelset !=null && levelset.contains(guid))
	    levelset.remove(guid);
	  else if (DEBUG)
	    Debug.printtagln(tag, "Trying to do unpublish with a"+
			     " nonexistent objguid="+guid+", level="+level);
	}
    }

    /**
     * Add new publishinfo for this guid to the backpointers table
     */
    private void add_guid_to_backpointers( SecureHash objguid, 
					   LinkedList pinfo ) {
	final String tag = class_tag + ".add_guid_to_backpointers";

	LinkedList ll;
	if( DEBUG ) {
	    Debug.printtagln( tag, "Adding in " + objguid + " with pinfo "
			      + pinfo );
	}

	synchronized( _backpointers ) {
	    if (_backpointers.containsKey(objguid)) {
		ll = (LinkedList) _backpointers.get(objguid);
		//ll.addAll(pinfo);
		synchronized( ll ) {
		    for( int i = 0; i < pinfo.size(); i++ ) {
			if( !ll.contains(pinfo.get(i)) ) {
			    ll.add( pinfo.get(i) );
			}
		    }
		}
		/**
		 * TODO: This check is whack, I don't think it's right.  Is it?
		 * Possibly only valid on a DynamicObjMapMsg.INSERTION
		 if (DEBUG) {
		 HashSet level = 
		 (HashSet)_objlevels.get(_local_ptr_index);
		 if(level==null || !level.contains(objguid))
		 BUG("level=" + level + 
		 " does not contain msg.objguid=" +
		 objguid + " and it should.");
		 }
		*/
	    } else {
		_backpointers.put(objguid, pinfo);
		HashSet level = 
		    (HashSet) _objlevels.get(_local_ptr_index);
		if (level == null) {
		    level = new HashSet();
		    _objlevels.put(_local_ptr_index, level);
		}
		level.add(objguid);
	    }
	}
    }

    /**
     * Handle message regarding obj pointer information for another node
     */
    private void handle_dynamic_obj_map( DynamicObjMapMsg msg ) {
	final String tag = class_tag + ".handling DynamicObjMapMsg";

	if (DEBUG) Debug.printtagln(tag, "Handling DynamicObjMapMsg: "+ msg);
	if (msg.type == DynamicObjMapMsg.INSERTION) {
	    if (msg.objcount > 0) {  //in other words, not just a null msg
		add_guid_to_backpointers( msg.objguid, msg.pinfo );
		
		//got object from msg, now reply and confirm
		DynamicObjReqMsg dorm = 
		    new DynamicObjReqMsg(msg.peer, _self_guid, msg.objguid,
					 DynamicObjReqMsg.CONFIRM);
		dispatch(dorm);
	    }
	} else if (msg.type == DynamicObjMapMsg.DELETION) {
	    if (msg.objcount <= 0) { //a null msg
		//this should never happen
		if (DEBUG) 
		    Debug.printtagln(tag, 
				     "Unexpected deletion with no objects.");
	    } else {
		//now need to figure out if this object should stay or
		//go to the next surrogate node
		//HERE
		int [] digits = RoutingTable.digits(msg.objguid);
		int [] selfdigits = RoutingTable.digits(_self_guid);
		int i;
		NodeId next = null;
		for (i=0 ; digits[i] == selfdigits[i] && 
			 i<RoutingTable.DIGITS_PER_GUID; i++);
		if (i == RoutingTable.DIGITS_PER_GUID) 
		    //every single digit matches us (highly unlikely)
		    next = my_node_id;
		else {
		    //we need to examine every single digit of the 
		    //object ID and self_guid, and make sure there are no
		    //better roots anywhere down the road.
		    RouteEntry re;
		    NodeId temp;
		    for (; i<RoutingTable.DIGITS_PER_GUID; i++) {
			//look to see if there's a better root at this level
			//next=null right now
			for (int j = digits[i]; ; j++) {
			    if (j==RoutingTable.DIGIT_VALUES) j=0;
			    re = _routing_tbl.getEntry(i, j);
			    if (re!=null) {
				temp = re.getFirst();
				if (temp == null) continue;
				else if (re.getFirst() != null &&
					 !re.isLeavingAtIndex(0)) {
				    //if it's valid, doesn't matter if it's 
				    //pointing at us or a different node, 
				    //we stop.
				    next = temp;
				    break;
				}
			    }
			}
			if (next != null) {
			    if (next.equals(my_node_id))
				next = null;
			    else break;
			}
		    }
		}
		if (next == null || next.equals(my_node_id)) {
		    //this is the last hop, integrate this object
		    //but we need to check to see if we're leaving too
		    if (_routing_tbl.isAttached()) {  //not leaving
			add_guid_to_backpointers( msg.objguid, msg.pinfo );
			//now tell the leaving node this guy is staying here
			DynamicObjMapMsg domm=
			    new DynamicObjMapMsg(msg.src, msg.src, msg.objguid,
						 new LinkedList(), msg.seq,
						 msg.seq, 
						 DynamicObjMapMsg.DELETIONACK);
			sendNonceMsg (domm, SHORT_TIMEOUT, null);
		    } else {  // !isAttached(), we're leaving too
			//now need to find next best object surrogate
			//then tell leaving node about it
			NodeId nextSurrogate = 
			    _routing_tbl.nextSurrogate(msg.ignoreNodes, 
						       _self_guid);
			DynamicObjMapMsg nack =
			    new DynamicObjMapMsg(msg.src, my_node_id, 
						 msg.objguid, msg.pinfo, 
						 msg.objcount, msg.seq,
						 msg.DELETION_NACK);
			nack.addIgnoreNodes(msg.ignoreNodes);
			nack.addNextSurrogate(nextSurrogate);
			sendNonceMsg(nack, SHORT_TIMEOUT, null);
		    }
		} else {
				//need to route to the next node instead
		    DynamicObjMapMsg forward =
			new DynamicObjMapMsg(next, msg.src, msg.objguid,
					     msg.pinfo, msg.objcount, msg.seq,
					     msg.DELETION);
		    sendNonceMsg(forward, SHORT_TIMEOUT, null);
		}
	    }
	} else if (msg.type == DynamicObjMapMsg.DELETIONACK) {
	    //we're the node that's leaving, so check to see if we've got
	    //all the objects acked from their new surrogate nodes
	    if (DEBUG)
		Debug.printtagln(tag, "Handing DELETION ACK for obj:"+
				 msg.objguid+" from node: "+msg.peer);

	    try {
		Integer deletion = new Integer( DynamicObjMapMsg.DELETION );
		int nonce = findAssociatedNonce( new Pair( msg.objguid,
							   deletion ) );
		NonceAckMsg nam = new NonceAckMsg( msg.peer );
		nam.addNonceAck( nonce );
		if( DEBUG ) {
		    Debug.printtagln( tag, "Using " + msg + 
				      " as an acknowledgement." );
		}
		receivedAck( nam );
	    } catch( NonceNotFoundException e ) {
		if( DEBUG ) {
		    Debug.printtagln( tag, "Could not find nonce: " + e 
				      + " Probably timed out.  Ignoring." );
		}
		return;
	    }
	    if (_leavingobjs.containsKey(msg.objguid)) {
		_leavingobjs.remove(msg.objguid);
		if (_leavingobjs.isEmpty()) {
		    //we're DONE! now send out EndDelete to all backpointers
		    send_end_deletes();
		}
	    } else { 
		if (DEBUG)
		    Debug.printtagln(tag, "Unexpected ObjMapMsg.DELETIONACK!");
	    }
	} else if (msg.type == DynamicObjMapMsg.DELETION_NACK) {
	    //our request hit another leaving node somewhere, and he has
	    //nacked with a suggested next surrogate
	    if (DEBUG)
		Debug.printtagln(tag, "Handling DELETION NACK for obj:" +
				 msg.objguid+" from node: "+msg.peer);
	    //if we hit 4 leaving nodes, then we just republish from source
	    if (msg.ignoreNodes != null && msg.ignoreNodes.size() > 3) {
		for (Iterator it = msg.pinfo.iterator(); it.hasNext(); ) {
		    dispatch(new RepublishRequestMsg(msg.objguid,
						     (SecureHash) it.next(),
						     _self_guid));
		}
		return;
	    }
	    DynamicObjMapMsg nextTry = 
		new DynamicObjMapMsg(msg.nextSurrogate, my_node_id, 
				     msg.objguid, msg.pinfo, msg.objcount,
				     msg.seq, msg.DELETION);
	    msg.ignoreNodes.add(msg.peer);
	    nextTry.addIgnoreNodes(msg.ignoreNodes);
	    dispatch(nextTry);
	}
    }

    private void ack_multicast( SecureHash newGuid, NodeId notifier ) {

	final String tag = class_tag + ".ack_multicast";

	//unlock the pointers:
	NodeState ns = _mcast.getNodeState(newGuid);
	//String lockindex = (ns.alpha+1)+":"+
	//    RoutingTable.digits(newGuid)[ns.alpha+1];
	Object lockindex = getLockIndex( newGuid );
	TreeSet hs = (TreeSet) _mcast.indexToLocks.get(lockindex);
	if (hs == null) {
	    //this should never happen, we add ourselves when we
	    //started our multicast at this node
	    if (DEBUG)
		Debug.printtagln(tag, "Error: missing " + 
				 "lock for ourselves at lockindex " +
				 lockindex + " for guid:"+newGuid);
	} else {
	    hs.remove(newGuid);
	    if( DEBUG ) {
		Debug.printtagln( tag, "removed " + newGuid + 
				  " from lock index " + lockindex );
	    }
	}
	
	//whenever we send up the ack, we should also make sure that 
	//we check to add this guy to our routing table. We do this
	//by faking a notifymsg (GENERAL_TYPE) sent from the new node.
	//we need to do this BEFORE sending the ack and cleaning state
	DynamicNotifyMsg dnm = 
	    new DynamicNotifyMsg(notifier, newGuid, _self_guid,
				 DynamicNotifyMsg.GENERAL_TYPE, 
				 _mcast.getAlpha(newGuid)-1);
	handle_dynamic_notify(dnm, true);
	
	NodeId parent = _mcast.getMCParent(newGuid);
	if (parent == null) {
	    //if we have no parent (we're the root of the mcast)
	    //we need to send ready msg to new node to indicate end 
	    //of mc
	    
	    DynamicMCReqMsg dmcrm = 
		new DynamicMCReqMsg(_mcast.getMCAddr(newGuid),
				    _self_guid, 
				    DynamicMCReqMsg.CONFIRM_TYPE,
				    _mcast.getAlpha(newGuid));
	    //now clean up the state for this particular multicast tree
	    if (DEBUG)
		Debug.printtagln(tag, "Cleaned slate in ping " + 
				 "(as surr) guid:" + newGuid);
	    _mcast.cleanSlate(notifier, newGuid);
	    sendNonceMsg(dmcrm, SHORT_TIMEOUT, new Integer(0));
	} else {
	    DynamicDMAckMsg dmam = 
		new DynamicDMAckMsg(parent, newGuid, _self_guid);
	    //now clean up the state for this particular multicast tree
	    if (DEBUG)
		Debug.printtagln(tag, "Cleaned slate in ping " + 
				 "(as dmm) guid:"+
				 newGuid);
	    _mcast.cleanSlate(notifier, newGuid);
	    dispatch(dmam);
	}
    }

    /**
     * informing this current node about new node with nodeID
     * notifier and guid notGuid
     */
    private void handle_dynamic_notify(DynamicNotifyMsg msg, 
				       boolean internal) {
	final String tag = class_tag + ".handle_dynamic_notify";
	
	if (msg.type == DynamicNotifyMsg.RESPONSE_TYPE) {

	    if (_mcast == null) 
		_mcast = new MulticastState(_self_guid);

	    if (_mcast.setNotify(msg.src)) {

		//we're ready to send up ack
		//all acks for this guid at this level are back
		//we need to return an ack to our parent, and unlock this guy

		ack_multicast( msg.src, msg.peer );
	
	    }
	    return;
	}

	//if the node is added, either to fill a hole or as an optimization,
	//we need to notify the node to add us to its reverse link list
	//we do this by pinging the node and putting this msg in 
	//_waiting_notifies
	if (DEBUG) Debug.printtagln(tag, "Handling DynamicNotifyMsg: "+ msg);
	if (my_node_id.equals(msg.peer)) {
	    if (DEBUG)
		Debug.printtagln(tag, "Received notify from self, ignoring.");
	    return;
	}
      
	if (msg.type == DynamicNotifyMsg.FILLHOLE_TYPE ||
	    msg.type == DynamicNotifyMsg.DMCAST) {
	    //we need to keep this guy as part of initial nearest neighbor
	    GuidAddr ga = new GuidAddr(msg.src, msg.peer);
	    if (!initList.contains(ga))
		initList.add(ga);
	}
	if (_waiting_notifies.containsKey(msg.peer)) {
	    //this is useful for cases where we get multiple notifymsgs
	    //via different routes for the same guy
	    if (DEBUG) Debug.printtagln(tag, "Receiving notify msg for a"+
					" new node already pinged. Ignored.");
	    return;
	}
	double latency;
	int changed_level;
	if (msg.type == DynamicNotifyMsg.FILLHOLE_TYPE) {
	    Vector apings = new Vector();
	    for (int i=0; i<msg.list.length; i++) {
		NodeId nid = msg.list[i].getAddr();
		latency = getPingTime(nid);
		if( latency == -1 ) {
		    // couldn't get a ping.  do it asynchronously
		    apings.add( nid );
		} else {
		    boolean sendNotify = nid.equals(msg.peer);
		    finish_do_notify( nid, msg.list[i].getGuid(), latency, 
				      msg.minlevel, sendNotify );
		}

	    }
	    if( !apings.isEmpty() ) {
		NodeId[] nids = new NodeId[ apings.size() ];
		for( int i = 0; i < nids.length; i++ ) {
		    nids[i] = (NodeId) apings.get(i);
		}
		startAPings( nids, msg );
	    }
	    return;
	}
	
	//if msg is not of FILLHOLE_TYPE
	latency = getPingTime(msg.peer);
	if( latency == -1 ) {
	    NodeId[] nids = {msg.peer};
	    msg.internal = internal;
	    startAPings( nids, msg );
	} else {
	    finish_do_notify( msg.peer, msg.src, latency, 
			      msg.minlevel, !internal );
	}
    }

    /**
     * Asynchronous pings have finished for the DynamicNotifyMessage
     */
    private void finish_do_notify( NodeId nid, SecureHash guid, 
				   double latency, int minlevel,
				   boolean sendNotify ) {

	add_node_to_rt( nid, guid, minlevel, latency );
	_currentList_pings.put(nid, new Double(latency));
	
	if (sendNotify) {
	    sendNonceMsg(new DynamicNotifyMsg(nid, _self_guid, 
					      guid,
					      DynamicNotifyMsg.RESPONSE_TYPE),
			 SHORT_TIMEOUT, null);
	    //send out response to notify msg to set .notified
	}
	
    }

    /**
     * Function that sends out a directed multicast to the relevant 
     * set of nodes which should know about its insertion into the network.
     * The node goes through all higher-level nodes in the routing table
     * starting with level equal to argument (hop) and sends a DMCastMsg to
     * them.  All of them are added to the MulticastState.  Then we go through
     * the objects on this node, and move the correct ones to the new node.
     * Then we look at whether we should send back an ack.
     */
    private void handle_dynamic_dmcast(DynamicDMCastMsg msg) {
	final String tag = class_tag + ".handle_dynamic_dmcast";

	if (DEBUG) Debug.printtagln(tag, "Handling DynamicDMCastMsg: "+ msg);

	if (_mcast == null) _mcast = new MulticastState(_self_guid);

	if( !msg.fromLock && _mcast.getNodeState( msg.newGuid ) != null ) {
	    if (DEBUG) {
		Debug.printtagln(tag, "Already doing an mcast for " + 
				 msg.newGuid + ", ignoring." );
	    }
	    return;
	}

	send_mc_dynamic_notify( msg.newAddr, msg.newGuid, msg.watchlist, 
				msg.alpha );

	//want to add this node to our routing table
	add_node_to_rt( msg.newAddr, msg.newGuid, msg.alpha );

	//now, we need to send multicast messages to each node we should
	//then we should execute the transfer data obj process
	//and also wait for all the acks to come back
	//BUT NOT IF fromLock flag is SET
	boolean sentMCast = false;
	NodeState ns = null;   //have to declare this here for scoping reasons
	if (!msg.fromLock) {

	    sentMCast = do_multicast( msg.newAddr, msg.newGuid, msg.peer,
				      msg.alpha, msg.watchlist, false );
	    ns = _mcast.getNodeState( msg.newGuid );
	}  //end if (!msg.fromLock)

	mc_move_objects( msg.newAddr, msg.newGuid );

	//if we're a leaf of the mcast tree, then ack now back to the parent
	//BUT IF ONLY msg is not from a lock
	if (!msg.fromLock) {
	    if (!sentMCast && ns.notified) {
		//whenever we send up the ack, we should also make sure that 
		//we check to add this guy to our routing table. We do this
		//by faking a notifymsg (type=GENERAL_TYPE) sent from the 
		//new node. we need to do this BEFORE sending the ack and 
		//cleaning state
		DynamicNotifyMsg newdnm = 
		    new DynamicNotifyMsg(msg.newAddr, msg.newGuid, _self_guid,
					 DynamicNotifyMsg.GENERAL_TYPE, 
					 msg.alpha-1);
		handle_dynamic_notify(newdnm, true);
	
		DynamicDMAckMsg ack = 
		    new DynamicDMAckMsg(msg.peer, msg.newGuid, _self_guid);
		if (DEBUG) {
		    Debug.printtagln(tag, "Acking DMCast: "+msg+
				     " by sending MCastAck to "+ msg.peer);
		    Debug.printtagln(tag, "Cleaning slate in dm_cast for guid:"
				     + msg.newGuid);
		}
		//now clean up the state for this particular multicast tree
		_mcast.cleanSlate(msg.newAddr, msg.newGuid);
		dispatch(ack);
	    }
	} else {  //end if !msg.fromLock
	    DynamicDMAckMsg ack = 
		new DynamicDMAckMsg(msg.peer, msg.newGuid, _self_guid);
	    if (DEBUG)
		Debug.printtagln(tag, "Acking DMCast for lockedDMC: "+msg+
				 " by sending MCastAck to "+msg.peer);
	    dispatch(ack);
	}
    }

    /**
     * Handler for nodes which are already inserted to handle 
     * DynamicSeedMsg from a new node as part of the nearest neighbor
     * algorithm.  The seedmsg is a request for all backpointers in
     * _revpointers pertaining to a certain routing level.
     */     
    private void handle_dynamic_seed (DynamicSeedMsg msg) {
	final String tag = class_tag + ".handle_dynamic_seed";
	if (DEBUG) {
	    Router.dumpRevpointers(_revpointers, msg.level);
	}
	HashMap level = (HashMap) _revpointers.get(new Integer(msg.level));
	Map.Entry entry;
	ArrayList al = new ArrayList();
	if (level == null) {
	    if (DEBUG)
		Debug.printtagln( tag, "Relatively rare occurence of " + 
				  "empty revpointers");
	} else 
	    for (Iterator i = level.entrySet().iterator(); i.hasNext(); ) {
		entry = (Map.Entry) i.next();
		al.add(new GuidAddr((SecureHash) entry.getKey(), 
				    (NodeId) entry.getValue()));
	    }
	//always send yourself back as seed for next round.
	al.add(new GuidAddr(_self_guid, my_node_id));
	DynamicSeedMapMsg dsmm =
	    new DynamicSeedMapMsg(msg.peer, _self_guid, al, msg.level);
	if (DEBUG)
	    Debug.printtagln( tag, "Replying to seed request from " +
			      msg.peer + ", level=" + msg.level);
	// set the stage id, just in case
	dsmm.setStageId( msg.getStageId() );
	// allow it to serve as an ack for the DynamicSeedMsg
	sendNonceAckMsg(dsmm);
	// now see if this guy should be added to the routing table
	GuidAddr[] gaar = { new GuidAddr( msg.src, msg.peer ) };
	measure_and_optimize( gaar );
    }

    /**
     * Handler for responses from nodes who've received DynamicSeedMsg, and
     * have responded with their backpointers.
     */
    private void handle_dynamic_seedmap (DynamicSeedMapMsg msg) {
	final String tag = class_tag + ".handle_dynamic_seedmap";
	if( !seedmap_nodes.contains( msg.peer ) ) {
	    if (DEBUG) {
		Debug.printtagln(tag, "Duplicate seed map msg for " + 
				 msg.peer + ", ignoring." );
	    }
	    return;
	}
	seedmap_nodes.remove( msg.peer );
	GuidAddr ga;
	if (DEBUG)
	    Debug.printtagln(tag, "Handling seed reply, new seedcount=" + 
			     seedmap_nodes.size());

	//add revpointers from this guy into our list    
	for (Iterator i=msg.revpointers.iterator(); i.hasNext(); ) {
	    ga = (GuidAddr) i.next();
	    if (!initList.contains(ga) && 
		(!my_node_id.equals(ga.getAddr()))) {
		initList.add(ga);
		if (DEBUG)
		    Debug.printtagln(tag, "Adding Guid to initList: " + ga);
	    }
	}

	if (seedmap_nodes.isEmpty()) {
	    reinitialize_nn();
	}
    }

    /**
     * Start over with nearest neighbor algorithm
     */
    private void reinitialize_nn() {
	final String tag = classname+".reinitialize_nn";	
	double pingdist;
	GuidAddr ga;
	Vector apings = new Vector();
	for (Iterator i = initList.iterator(); i.hasNext(); ) {
	    ga = (GuidAddr) i.next();
	    pingdist = getPingTime(ga.getAddr());
	    if (DEBUG) Debug.printtagln(tag, "pingdist for " + ga + ": " +
					pingdist );
	    if( pingdist == -1 ) {
		apings.add( ga );
	    } else {
		_currentList_pings.put(ga.getAddr(), new Double(pingdist));
	    }
	}
	// if you couldn't get any pings, do asynchronous pings
	if( !apings.isEmpty() ) {
	    NodeId[] nids = new NodeId[ apings.size() ];
	    for( int j = 0; j < apings.size(); j++ ) {
		GuidAddr gaddr = (GuidAddr) apings.get(j);
		nids[j] = gaddr.getAddr();
		if (DEBUG) Debug.printtagln(tag, "starting APing for " +
					    gaddr );
	    } 
	    startAPings( nids, APING_REINIT_NN );
	    return;
	}
	_nearest_neighbors = new TreeMap();
	requestSeeds();
    }

    /**
     * Integration is done, let the world know
     */
    private void finished_integration() {
	final String tag = classname+".finished_integration";

	timer = Clock.current_date(my_node_id).getTime() - timer;
	System.err.println("Complete Integration time in ms: "+timer+
			   " for node:" + my_node_id + _self_guid);
	if (DEBUG)
	    Debug.printtagln(tag, "Nearest neighbor completed.");
	
	status = ATTACHED;
	_routing_tbl.node_status = status;
	dispatch(new RoutingTable.Available(_routing_tbl, _revpointers,
					    _objlevels));
	send_monitor_msgs();
	TapestryReadyMsg ready = new TapestryReadyMsg(_self_guid, timer);
	dispatch(ready);
    }

    /**
     * Function that gets called after a round of ping results have been 
     * gathered. Requests seeds from the current list of neighbors, or 
     * quits insertion if we've finished the final level. Gets called at the
     * end of handle_dynamic_mcreq and handle_dynamic_seedmap
     */
    private void requestSeeds() {
	//these are the results of the initList;
	final String tag = classname+".requestSeeds";
	if (DEBUG)
	    Debug.printtagln(tag, "current initList all ponged back."
			     +" Level = "+insertinglevel);
	Set nextRoundSeeds;
	nextRoundSeeds = nearest_level(insertinglevel, initList,
				       _currentList_pings);
	//should be a change of level here
	insertinglevel --;
	if (insertinglevel < 0) {  //STOP, we're DONE!!!!
	    finished_integration();
	    return;
	}
	initList = new TreeSet();
    
	DynamicSeedMsg seed;
	NodeId seedAddr;
	seedmap_nodes = new TreeSet();
	Iterator it = nextRoundSeeds.iterator();
	for (int i=0; i < 16 && it.hasNext(); i++) {
	    seedAddr = ((GuidAddr) 
			((Map.Entry) it.next()).getValue()).getAddr();
	    seed = new DynamicSeedMsg(seedAddr, _self_guid, insertinglevel);
	    if (DEBUG)
		Debug.printtagln(tag, "Sending out requests for next " +
				 "round seeds to " + seedAddr);
	    sendNonceMsg(seed, SHORT_TIMEOUT, 
			 new Pair( USERDATA_SEED, new Integer(0)));
	    seedmap_nodes.add( seedAddr );
	} //end for
	if (DEBUG) {
	    Debug.printtagln(tag, "Sent out total of " + seedmap_nodes.size() 
			     + " seed reqs.");
	}
    }

    /**
     * Function to take in a bunch of latencies from ping results of
     * the last round of nearest neighbor, and use the latencies to fill up
     * this particular routing level entries.  This function scans through
     * the entire list, filling up a sorted routingtable level and closest
     * node list.  It returns the closest node list which can then be used
     * to generate the next list of candidates.
     * @param level The level of routing table we're trying to fill
     * @param list The list of nodes (guid-->nodeid) to be considered
     * @param pings The results of pings for the entire set
     * @return Nearest neighbors in a set, in sorted order
     */
    private Set nearest_level(int level, TreeSet list, HashMap pings) {
	SecureHash guid;
	NodeId addr;
	GuidAddr ga;
	int [] self_digits = RoutingTable.digits(_self_guid);
	int [] guid_digits;
	boolean written;

	for (Iterator i=list.iterator(); i.hasNext(); ) {
	    ga = (GuidAddr) i.next();
	    guid_digits = RoutingTable.digits(ga.getGuid());
	    addr = ga.getAddr();
	    Double lat = (Double) pings.get(addr);
	    if( lat == null ) {
		lat = new Double( UNKNOWN_PING_VAL );
	    }
	    written = _routing_tbl.write(level, guid_digits[level], addr,
					 ga.getGuid(), lat);
	    if( written ) {
		send_remove_pointers();
	    }

	    if (written) {
	      if (DEBUG_FT)
		Debug.printtagln(tag, _routing_tbl.toShortString()+
				 "nearest_level");
	      send_dynamic_ptr( addr, level, guid_digits[level] );
	    }
	    //now insert into sorted list as candidate for next round seed node
	    _nearest_neighbors.put((Double) pings.get(addr), ga);
	}
    
	//now return the closest neighbors in sorted order as a set
	return _nearest_neighbors.entrySet();
    }

    /**
     * Function to handle the dynamicMCReqMsg.  Currently, this is only sent
     * by the surrogate node to the new inserting node to indicate that the 
     * directed multicast has succeeded and all nodes have acked.  By then,
     * everything should be done, except for running the nearest neighbor 
     * algorithm to get the routing tables done.
     */
    private void handle_dynamic_mcreq(DynamicMCReqMsg msg) {
	//start with initList
	final String tag = class_tag +".handle_dynamic_mcreq";

	if (status != INSERT_MCAST_MODE) {
	    if (DEBUG)
		Debug.printtagln(tag, "Unexpected DynamicMCReqMsg, ignoring.");
	    return;
	}

	status = INSERT_NEARESTN_MODE;  //done with MCAST_MODE now
	if (_routing_tbl != null) _routing_tbl.node_status = status;

	if (msg.type != DynamicMCReqMsg.CONFIRM_TYPE) {
	    if (DEBUG)
		Carp.die("ERROR: unexpected DynamicMCReqMsg received.");
	}

	//done with multicast, meaning all object transfers should be done
	//therefore we send out the list of obj pointers to the router
	dispatch(new Router.ObjectMap(_backpointers));

	GuidAddr ga;
	insertinglevel = msg.insertinglevel;
	ga = new GuidAddr(msg.src, msg.peer);
	if (!initList.contains(ga))
	    initList.add(ga);
	if (DEBUG)
	    Debug.printtagln(tag, "Handling confirmed MC. Starting nearest " + 
			     "neighbor. level=" + insertinglevel + 
			     ", initList=" + initList);

	//BUGFIX (bug fixed as result of another bug fix)
	//by always adding the surrogate to the initList, we make sure
	//that initList is never empty, therefore, we always wait for 
	//response
	reinitialize_nn();
    }

    /**
     * Function called by handleEvent when an application stage has sent us 
     * a TapestryNodeDetachMsg, asking us to leave the Tapestry network.
     * The sequence of events is as follows:<BR>
     * <OL>
     * <LI>A sends DynamicBeginDelete to all backpointers<BR>
     *     - Each receiver marks pointers to X as leaving<BR>
     *     - DynamicBeginDelete includes alternate node candidate for
     *       each route level<BR>
     *     - All new publishes go through nodes as if A doesn't exist,
     *        but queries still go through it<BR>
     * <LI>A sends multicast to surrogate roots in routing table, with
     * objects rooted at A<BR>
     *     - Each node forwards on objects for which it is not the root<BR>
     *     - The multicast is acked<BR>
     * <LI>X, Y, ... request republishes of all objects that go through A
     * <LI>When A gets full ack from multicast, A sends EndDelete to X, Y, ..
     * <LI>When X, Y, .. receives EndDelete, they remove A from all tables and
     * responds with EndDeleteAck back to A
     * <LI>When A receives EndDeleteAck from all backpointer nodes, it knows
     * it is safe to leave. Now send DetachEventAck to application stage.
     * </OL>
     **/
    private void handle_node_detach() {
	HashMap replacements = new HashMap();
	SecureHash guid, bestguid;
	NodeId node, bestnode;
	Double latency, bestlatency;
	GuidAddr ga;
	RouteEntry re;
	final String tag = class_tag + ".handle_node_detach";
	if(DEBUG) Debug.printtagln(tag, "called voluntary deletion algorithm");

	//start by making a list of best alternates at each level
	status = DETACH_ROUTINGOBJ;
	if (_routing_tbl != null) _routing_tbl.node_status = status;
    
	for (int i=1; i<RoutingTable.DIGITS_PER_GUID; i++) {
	    bestnode = null;
	    bestguid = null;
	    bestlatency = null;
	    for (int j=0; j<RoutingTable.DIGIT_VALUES; j++) {
		re = _routing_tbl.getEntry(i, j);
		if (re == null || re.getFirst() == null)
		    continue;
		node = re.getFirst();
		if (node.equals(my_node_id)) {
		    // our own entry, check backup node
		    node = re.getAtIndex(1);
		    if (node == null) 
			continue;
		}
		latency = _routing_tbl.node_to_latency(node);
		if (latency == null) latency = new Double(Double.MAX_VALUE);
		if (bestnode == null || bestlatency.compareTo(latency) < 0) {
		    //found a better node
		    bestnode = node;
		    bestlatency = latency;
		}
	    }
	    guid = _routing_tbl.node_id_to_guid(bestnode);
	    if (bestnode != null && guid != null) {
		// The offset of -1 is because in order to find replacements
		// for this guy in the ith level in someone else's table, we
		// need to look in our (i+1)th level
		// Example: Our closest entry on level 0 is useless to others
		replacements.put(new Integer(i-1), 
				 new GuidAddr(guid, bestnode));
		if( DEBUG ) {
		    Debug.printtagln( tag, "made new replacement, guid=" + 
				      guid + ", bestnode=" + bestnode + 
				      ", index=" + (i-1) );
		}
	    }
	}
	//now send out the alternate list to all nodes on backpointers list
	if (DEBUG)
	    Debug.printtagln(tag, "Now sending alternate list " + 
			     "to backpointers list");
	DynamicBeginDelete dbd;
	backnodes = new TreeSet();
	if (DEBUG)
	    Router.dumpRevpointers(_revpointers, 5);
	for (int i = 0; i<RoutingTable.DIGITS_PER_GUID; i++) {
	    HashMap revnodes = (HashMap) _revpointers.get(new Integer(i));
	    if (revnodes == null) continue;
	    for (Iterator it = revnodes.values().iterator(); it.hasNext(); ) {
		node = (NodeId) it.next();
		if (DEBUG)
		    Debug.printtagln(tag, "Got backnode:" + node + 
				     " for level=" + i);
		if (!backnodes.contains(node) && !node.equals(my_node_id)) {
		    backnodes.add(node);
		    dbd = new DynamicBeginDelete(node, _self_guid, 
						 replacements);
		    if (DEBUG)
			Debug.printtagln(tag, "sending begindelete to " + 
					 node);
		    dispatch(dbd);
		}
	    }
	}

	// we also need to tell everyone we point to that we're not going
	// to be pointing to them anymore (including backups!).
	// assume each node only appears once
	NodeId node_rt;
	for (int i=0; i<RoutingTable.DIGITS_PER_GUID; i++) {
	    for (int j=0; j<RoutingTable.DIGIT_VALUES; j++) {
		re = _routing_tbl.getEntry(i,j);
		if( re != null ) {
		    for( int k = 0; k < re.size(); k++ ) {
			node_rt = re.getAtIndex(k);
			if( node_rt != null && 
			    !node_rt.equals( my_node_id ) ) {
			    sendNonceMsg( new DynamicPtrMsg( node_rt, 
							     _self_guid, 
							     i, true ),
					  SHORT_TIMEOUT, 
					  new Pair( USERDATA_PTR,
						    new Integer(0)));
			}
		    }
		}
	    }
	}

	//now need to package up all our objects, request republishes for ones
	//going through us, and send as a multicast the ones rooted at 
	//this node. This requires information kept at the publishInfo level.

	requestPartialRepublish(replacements);
	HashSet objs = 
	    (HashSet) _objlevels.get(_local_ptr_index);
	if (objs != null && !objs.isEmpty())
	    rerouteRootObjs(objs);
	else if (!backnodes.isEmpty()) {
	    //if there are no objects, tell backnodes you're done
	    if (DEBUG)
		Debug.printtagln(tag, "No objects rooted at this node, " + 
				 "send EndDelete");
	    send_end_deletes();
	} else {
	    // no backnodes and no objects, end now
	    delete_done();
	}

    }

    /**
     * Send DynamicEndDelete messages to all nodes that care
     */
    private void send_end_deletes() {
	DynamicEndDelete ded;
	NodeId node;
	for (Iterator it = backnodes.iterator(); it.hasNext(); ) {
	    node = (NodeId) it.next();
	    ded = new DynamicEndDelete(node, _self_guid, 
				       DynamicEndDelete.REQUEST_TYPE);
	    dispatch(ded);
	}
    }

    /**
     * Voluntary deletion is completed
     */
    private void delete_done() {
	final String tag = class_tag + ".delete_done";
	if (DEBUG)
	    Debug.printtagln(tag, "VOLUNTARY DELETE COMPLETE");
	//VOLUNTARY DELETION COMPLETE!!
	status = NOT_ATTACHED;
	if (_routing_tbl != null) _routing_tbl.node_status = status;
	dispatch(new TapestryDetachConfirm());
	if(DEBUG) Debug.printtagln(tag, "done voluntary deletion algorithm");
    }

    /**
     * A route has been switched by Router/ChannelHandler
     * Republish through.
     */
    private void handle_frls_change( FRLSChangeMsg msg ) {

	final String tag = class_tag + ".handle_frls_change";
	if( DEBUG ) {
	    Debug.printtagln( tag, "called" );
	}

	republishOnRouteSwitch( msg.digit );

	/*
	if( _routing_tbl == null ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Leaving, no routing table yet." );
	    }
	    return;
	}

	// find the guid of the node that was switched out
	RouteEntry entry = _routing_tbl.getEntry( msg.digit, msg.value );
	if( entry == null ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Routing entry was null." );
	    }
	    return;
	}
	NodeId old_node = entry.getAtIndex( msg.old_frls );
	if( old_node == null ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Node at old_frls was null." );
	    }
	    return;
	}
	
	if( DEBUG ) {
	    Debug.printtagln( tag, "Node " + old_node + " was switched out." );
	}

	SecureHash old_guid = _routing_tbl.node_id_to_guid( old_node );
	if( old_guid == null ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Couldn't translate " + old_node
				  + " to a guid." );
	    }
	    return;
	}

	// find out how many digits are matched:
	int level = Router.guidCompare( old_guid, _self_guid );

	// now republish
	republishOnRouteSwitch( level );
	*/

    }

    /**
     * Republish all objects on the same level as a node we have 
     * detected as lossy and switched out.
     */
    private void republishOnRouteSwitch( int level ) {

	final String tag = class_tag + ".republishOnRouteSwitch";

	if( _objlevels == null ) {
	    // we haven't started yet, so don't have anything to republish
	    return;
	}

	if( level == Router.LOCALROOTOBJS ) {
	    Carp.die( tag + " called with level: " + level + "!" );
	}

	// publish all object guids that match you at that level
	HashSet objs = (HashSet) _objlevels.get( new Integer(level) );
	if( objs == null ) {
	    return;
	}
	Iterator iter = objs.iterator();
	while( iter.hasNext() ) {
	    SecureHash objguid = (SecureHash) iter.next();
	    // send these to myself, and let Router determine where the 
	    // next hop is.
	    do_publishes( my_node_id, level-1, objguid, true );
	}

    }

    /**
     * Send out PartialRebuplish messages with replacements
     */
    private void requestPartialRepublish(HashMap replacements) {
	//organize objects by level, but ignore local objs 
	//(int i = LOCALROOTOBJS)
	for (int i=0; i<RoutingTable.DIGITS_PER_GUID; i++){
	    HashSet objs =
		(HashSet) _objlevels.get(new Integer(i));
	    if (objs == null || objs.isEmpty())
		continue;
	    SecureHash objguid;
	    for (Iterator it = objs.iterator(); it.hasNext(); ) {
		objguid = (SecureHash) it.next();
		LinkedList ll = (LinkedList) _backpointers.get(objguid);
		if (ll == null || ll.isEmpty())
		    continue;
		synchronized( ll ) {
		    for (Iterator it2 = ll.iterator(); it2.hasNext(); ) {
			NodeId lasthop = ((PublishInfo) it2.next()).lasthop;
			if (lasthop != null) {
			    GuidAddr ga = 
				(GuidAddr) replacements.get(new Integer(i));
			    if (ga == null) 
				dispatch(new PartialRepublish(objguid, lasthop,
							      _self_guid, 
							      null));
			    else 
				dispatch(new PartialRepublish(objguid, lasthop,
							      _self_guid, 
							      ga.getAddr()));
			}
		    }
		}
	    }
	}
    }

    /**
     * Method called to take the objects which are rooted at this node, and
     * route them out via a multicast, so that they end up at an alternate 
     * surrogate node.
     */
    private void rerouteRootObjs(HashSet objs) {
	final String tag = class_tag+".rerouteRootObjs";
	if (DEBUG) Debug.printtagln(tag, 
				    "Now handling object reroutes on: "+objs);
	//since all the objects we are rooted at all resolve to us at 
	//this last routing level, we should just take all those objects, 
	//and pass them to the next highest router up. Question is, is that 
	//always the case that all these objects came to us on the same 
	//routing level. Assume so for now...
	SecureHash objguid;
	int count = 1;
	int size = objs.size();

	//Hypothesis: All objects rooted at a particular node via surrogate 
	//routing will move to the same new surrogate when this root is 
	//removed. Surrogate is the highest level non-self node in the routing 
	//table. I currently believe this is true, and will do so until 
	//proven otherwise.
	//OTHERWISE: now we know this is NOT true. tapestry surrogate routing
	//as it is implemented does not result in a contiguous namespace 
	//allocation
    
	NodeId peer=null;
	//begin searching for surrogate node

	peer = _routing_tbl.nextSurrogate(null, _self_guid);
	if (peer == null) {
	    //this should never happen, unless he's already the only node left
	    //in the entire network.
	    //OR if the only nodes available rejected us, and are on the 
	    //ignore list
	    //so we just 
	    //BUG("Bug in searching for surrogate to move objs to. 
	    //     Contact Ben.");
	    //HERE just return and do nothing
	    return;
	}
    
	if (_leavingobjs == null) _leavingobjs = new HashMap();
	if (DEBUG)
	    Debug.printtagln(tag, "Chosen surrogate node is: "+peer);
	for (Iterator i=objs.iterator(); i.hasNext(); ) {
	    objguid = (SecureHash) i.next();
	    LinkedList ll = (LinkedList) _backpointers.get(objguid);
	    if(ll==null)
		BUG("_backpointers.get(objguid=" + objguid + ")==" + ll +
		    ". Contact Ben ASAP.");
	    //if this object list contains ourselves, we remove ourself
	    //if we're the only entry in the replica list, then do not reroute
	    synchronized( ll ) {
		ListIterator li = ll.listIterator();
		PublishInfo pi;
		while (li.hasNext()) {
		    pi = (PublishInfo) li.next();
		    if (_self_guid.equals(pi.src))
			li.remove();
		}
	    }
	    if (ll.isEmpty()) continue;
	    DynamicObjMapMsg domm = 
		new DynamicObjMapMsg(peer, my_node_id, 
				     objguid, ll, size, count, 
				     DynamicObjMapMsg.DELETION);
	    _leavingobjs.put(objguid, ll);
	    Integer deletion = new Integer(DynamicObjMapMsg.DELETION);
	    forceRegisterNonce( LONG_TIMEOUT, 
				new Pair( objguid, deletion ) );
	    sendNonceMsg (domm, SHORT_TIMEOUT, null);
	}
    }

    /**
     * Method called to handle the BeginDelete message received from a node
     * that intends to leave, and this current node is pointing to it in 
     * some routing table.  We now remove the entry from our primary routing
     * tables, and instead put it into a query only routing table.
     */
    private void handle_begin_delete(DynamicBeginDelete msg) {
	//first find it in our routing table, so that we can move it
	final String tag = class_tag + ".handle_begin_delete";
	if (DEBUG) Debug.printtagln(tag, "called with"+ msg);
	int [] oldguid = RoutingTable.digits(msg.src);
	int maxindex = Router.guidCompare(msg.src, _self_guid);
	if (maxindex == -1) {
	    if (DEBUG)
		Debug.printtagln(tag, "Unexpected: Deleting myself!!!");
	    return;
	}
	Map ids = _routing_tbl.get_node_map();
	for (int i=0; i<=maxindex; i++) {
	    Integer index = new Integer(i);
	    //implicit assumption invariant:
	    //a node that is not the node itself should only occur in its
	    //routing table once, except as backups to the self-node in
	    //earlier routing levels
	    RouteEntry re = _routing_tbl.getEntry(i, oldguid[i]);
	    if (re == null) {
		if (DEBUG)
		    Debug.printtagln(tag, "Unexpected behavior, " + 
				     "inconsistent revpointer" +
				     " led to missing routeEntry.");
		continue;
	    }
	    if (re.markLeaving(msg.peer)) {
		if (msg.replacements.containsKey(index)) {
		    NodeId replace = 
			((GuidAddr) msg.replacements.get(index)).getAddr();
		    SecureHash guid = 
			((GuidAddr) msg.replacements.get(index)).getGuid();
		    //if we do add this replacement node, send it a ptrmsg
		    if (DEBUG)
			Debug.printtagln(tag, "Replacement at level " + i +
					 ", Node:"+replace);
		    NodeId[] removed_node = new NodeId[1];
		    Pair p [] = new Pair [1];
		    if (re.add(replace, new Double(UNKNOWN_PING_VAL), 
			       removed_node, p)) {
			_routing_tbl.addDistance(UNKNOWN_PING_VAL, 
						 replace, guid);
			ids.put(replace, 
				((GuidAddr) 
				 msg.replacements.get(index)).getGuid());
			if (DEBUG_FT)
			  Debug.printtagln(tag, _routing_tbl.toShortString()+
					   "handle_begin_delete");
			send_dynamic_ptr( replace, i, oldguid[i] );
			// if a node got removed send a remove msg 
			// to the old guy
			if( removed_node[0] != null ) {
			    sendNonceMsg( new DynamicPtrMsg( removed_node[0], 
							     _self_guid, 
							     i, true ),
					  SHORT_TIMEOUT,
					  new Pair( USERDATA_PTR,
						    new Integer(0)));
			}
		    }
		    if (p[0] != null) 
		      dispatch(new FRLSChangeMsg(i, oldguid[i],
				      ((Integer)p[0].first()).intValue(),
				      ((Integer)p[0].second()).intValue()));
		}
		if (DEBUG)
		    Debug.printtagln(tag, "just marked " + msg.peer + 
				     " as leaving.");
	    } else if (DEBUG)
		Debug.printtagln(tag, "Failed to mark " + msg.peer + 
				 " as leaving.\nLeaving guid:" + 
				 msg.src + ", RouteEntry:" + re);
	} // end for (int i...
    }

    /**
     * Method to handle the EndDelete message received from a node 
     * that intends to leave. This message tells us that it is completely
     * done, and we should remove all references to it.  By now, we should
     * have already removed it from the main routing table, and only look it
     * up for queries, but not publishes. We now remove that reference as well
     * and then respond with an EndDelete that is an ACK to end the 
     * transaction.
     */
    private void handle_end_delete(DynamicEndDelete msg) {
	final String tag = class_tag+".handle_end_delete";
	if (DEBUG)
	    Debug.printtagln(tag, "Handling "+msg+" from " +msg.peer);
	if (msg.type == DynamicEndDelete.REQUEST_TYPE) {
	    int [] oldguid = RoutingTable.digits(msg.src);
	    int maxindex = Router.guidCompare(msg.src, _self_guid);
	    if (maxindex == -1) {
		if (DEBUG)
		    Debug.printtagln(tag, "Error: Deleting myself!!!");
		return;
	    }

	    for (int i=0; i<=maxindex; i++) {
		RouteEntry re = _routing_tbl.getEntry(i, oldguid[i]);
		if (DEBUG && re!=null)
		    Debug.printtagln(tag, "Deleting routing entry:"+
				     re.getFirst()+" on level:"+i);
		if (re!=null && re.getFirst()!= null) { 
		    re.removeEntry(msg.peer);
		    SecureHash removedGuid = 
			_routing_tbl.node_id_to_guid(msg.peer);
		    send_routing_table_remove(msg.peer, removedGuid, -1);
		} else if (DEBUG)
		    Debug.printtagln(tag, "Inconsistent revpointer led " + 
				     "to missing route entry.");
	    }
	    if (DEBUG)
		Debug.printtagln(tag, "After deletion: routing table is:"+
				 _routing_tbl);
	    dispatch(new DynamicEndDelete(msg.peer, msg.src,
					  DynamicEndDelete.CONFIRM_TYPE));
	} else {
	    if (backnodes.remove(msg.peer) == false && DEBUG) {
		Debug.printtagln(tag, "Error: unexpected EndDelete " + 
				 "confirmation");
	    }
	    if (backnodes.isEmpty()) {
		delete_done();
	    }
	}
    }

    /**
     * Method called to look through the _objlevels mapping, and take all
     * objects which route through this node, and send requests for republish
     * to their originating nodes. This function is called by the handler
     * of NodeDetach, after the BeginDelete message has been sent to the 
     * backpointed nodes.
     */
    public void requestRepublish() {
	HashSet level;
	SecureHash objguid;
	PublishInfo pi;
	RepublishRequestMsg rrm;
	final String tag = class_tag + ".requestRepublish";
	if (DEBUG) Debug.printtagln(tag, "Level LOCALOBJS:"+ 
				    (HashSet)_objlevels.get(_local_ptr_index));
    
	for (int i = 0; i<RoutingTable.DIGITS_PER_GUID; i++) {
	    level = (HashSet) _objlevels.get(new Integer(i));
	    if (DEBUG) Debug.printtagln(tag, "Level "+i+" objlevel:"+level);
	    if (level == null) continue;
	    for (Iterator it = level.iterator(); it.hasNext(); ) {
		objguid = (SecureHash) it.next();
		LinkedList ll = (LinkedList) _backpointers.get(objguid);
		if (ll == null) {
		    if (DEBUG)
			Debug.printtagln(tag, "objlevel not found inside " + 
					 "_backpointers.");
		    continue;
		}
		LinkedList messages = new LinkedList();
		synchronized( ll ) {
		    for (Iterator iter = ll.iterator(); iter.hasNext(); ) {
			pi = (PublishInfo) iter.next();
			rrm = new RepublishRequestMsg(objguid, pi.src, 
						      _self_guid);
			messages.add(rrm);
			if (DEBUG)
			    Debug.printtagln(tag, 
					     "sending republish request to :"
					     + pi.src);
		    }
		}
		// dispatch these afterward, to avoid any possible deadlock 
		// with Router
		for( int j = 0; j < messages.size(); j++ ) {
		    dispatch( (QueueElementIF) messages.get(j) );
		}
	
	    }
	}
    }

    /**
     * Send a DynamicPtrMsg (and let Patchwork know about it)
     */
    private void send_dynamic_ptr( NodeId newAddr, int level, int value ) {

	final String tag = class_tag + ".send_dynamic_ptr";
	sendNonceMsg( new DynamicPtrMsg(newAddr, _self_guid, level), 
		      SHORT_TIMEOUT, new Pair( USERDATA_PTR,
					       new Integer(0)));
	// let patchwork know there's someone new
	if (DEBUG_FT) {
	  Debug.printtagln(tag, "new monitor: "+level+":"+value+":"+newAddr);
	}
	if (MONITORING)
	  dispatch(new MonitorMsg(newAddr, level, 
				  new Triple(level, value, newAddr), my_sink));
	
    }

  /***
   ** remove all references to a node from the reverse ptr list
   **/
  private void remove_reverse_ptr(NodeId nid) {
    SecureHash guid = _routing_tbl.node_id_to_guid(nid);
    if (guid == null) {
      System.err.println("BUG: missing nid to guid mapping");
      return;
    }
    Integer key;
    HashMap revs;
    for (Iterator i = _revpointers.keySet().iterator(); i.hasNext(); ) {
      key = (Integer) i.next();
      revs = (HashMap) _revpointers.get(key);
      if (revs == null) continue;
      if (revs.remove(guid) != null)
	send_reverse_ptr_remove(nid, guid, -1);
    }
  }

    /**
     * Someone is pointing to you, add them to your reverse pointers
     */
    private void handle_dynamic_ptrmsg(DynamicPtrMsg msg) {
	final String tag = class_tag + ".handle_dynamic_ptrmsg";
	if (DEBUG)
	    Debug.printtagln(tag, "Handling ptrmsg: "+msg);
	HashMap revs = (HashMap) _revpointers.get(new Integer(msg.level));
	if (revs == null) {
	    revs = new HashMap();
	    _revpointers.put(new Integer(msg.level), revs);
	}
	if (!msg.remove) {
	    if (!revs.containsKey(msg.srcguid)) {
		revs.put(msg.srcguid, msg.peer);
	        send_reverse_ptr_add( msg.peer, msg.srcguid, -1 );
	    }
	} else {
	    //remove this revpointer
	    revs.remove(msg.srcguid);
	    send_reverse_ptr_remove( msg.peer, msg.srcguid, -1 );
	}
    }
  
    /**
     * Someone is requesting that you republish certain pointers
     */
    private void handle_partialrepub(PartialRepublish msg) {
	final String tag = class_tag + ".handle_partialrepub";

	//first figure out who to send the republish to 
	//(not msg.peer obviously).
	NodeId nexthop = null;

	//now send out the publish message as if we were just forwarding
	//on a standard publish message

	//compare the selfGuid and the leavingNode's guid.
	//look for a backup to the leavingNode.
	//If one is found, use it. Otherwise, use the suggestedBackup
	int hop = Router.guidCompare(_self_guid, msg.requestor);
	if (hop == -1) return;
	int [] objdigits = RoutingTable.digits(msg.objGuid);
	RouteEntry entry = _routing_tbl.getEntry(hop, objdigits[hop]);
	if (entry == null) 
	    nexthop = msg.suggestedBackup;
	else {
	    for (int i=0; i<RouteEntry.REDUNDANT_LVL; i++) {
		nexthop = entry.getAtIndex(i);
		if (nexthop == null) {
		    nexthop = msg.suggestedBackup;
		    break;
		}
		if (!nexthop.equals(msg.peer)) 
		    break;
	    }
	}
	
	if (nexthop == null) {
	    nexthop = _routing_tbl.surrogate(objdigits, objdigits[hop], hop);
	    if (nexthop == null)
		//we are the root, just return now
		return;
	}

	// just return if the nexthop is us, because that could
	// throw us into an infinite loop
	if( nexthop.equals( my_node_id ) ) {
	    return;
	}

	do_publishes( nexthop, hop, msg.objGuid, false );
    }

    /**
     * Send out publish messages about the given guid to the next hop
     *
     * @param inbound whether to mark the message as inbound.  Used to 
     *                send network messages to yourself
     */
    private void do_publishes( NodeId next_hop, int hop, SecureHash objguid,
			       boolean inbound ) {

	final String tag = class_tag + ".do_publishes";
	if( DEBUG ) {
	    Debug.printtagln( tag, "called with next_hop=" + next_hop +
			      ", hop=" + hop + ", objguid=" + objguid
			      + ", inbound=" + inbound );
	}


	LinkedList publish_infos = (LinkedList) _backpointers.get(objguid);
	if (publish_infos == null || publish_infos.isEmpty()) {
	    if (DEBUG)
		Debug.printtagln(tag, "null or empty local publishinfo");
	    return;
	}


	PublishInfo pi;
	LinkedList messages = new LinkedList();
	synchronized( publish_infos ) {
	    for (Iterator it = publish_infos.iterator(); it.hasNext(); ) {
		pi = (PublishInfo) it.next();
		if (DEBUG) {
		    if (pi.src == null)
			System.err.println("pi.src=null, pi:"+pi);
		    if (objguid == null)
			System.err.println("msg.objGuid=null");
		}
		PublishMsg pm = new PublishMsg(next_hop, pi.src, objguid,
					       pi.tag, hop, true);
		pm.inbound = inbound;
		messages.add( pm );
		if( DEBUG ) {
		    Debug.printtagln( tag, "sending " + pm );
		}
	    }
	}
	// send these afterward to avoid deadlock
	for( int i = 0; i < messages.size(); i++ ) {
	    dispatch( (QueueElementIF) messages.get(i) );
	}

    }

    /**
     * handle a message asking you for your routing info
     */
    public void handle_debugmsg(DebugMsg msg) {
    
	if (msg.type == DebugMsg.ROUTING_TABLE_REQUEST) {
      
	    if (_routing_tbl == null) {
		dispatch(new DebugRespMsg(msg.peer, true, DebugMsg.DONE, 
					  msg.requestID, _self_guid, 
					  "No Info"));
	    } else {
		//need to send back entire routing level plus backpointers
		//we make a copy of the revpointers map, then go level by 
		//level, and send pointers while deleting duplicates from 
		//the revpointers map
		boolean sent = send_dynamic_map_msgs( msg.peer, msg.level );
	
		// If no messages were sent, send a message saying we're done
		if (!sent) {
		    dispatch(new DebugRespMsg(msg.peer, true, DebugMsg.DONE, 
					      msg.requestID, _self_guid, 
					      "No Info"));
		}
	
		if (DEBUG)
		    System.out.println("Routing Table Sent");
	    }
      
	} else if (msg.type == DebugMsg.BACK_PTR_REQUEST) {
	    if (_backpointers == null) {
		dispatch(new DebugRespMsg(msg.peer, true, DebugMsg.DONE, 
					  msg.requestID, _self_guid, 
					  "No Info"));
	    } else if (_backpointers.keySet() == null) {
		dispatch(new DebugRespMsg(msg.peer, true, DebugMsg.DONE,
					  msg.requestID, _self_guid, 
					  "No Info"));
	    } else {
		boolean sent = false;
		SecureHash objguid;
		DynamicObjMapMsg domm;
		HashSet l = new HashSet();

		for(Iterator i = _backpointers.keySet().iterator(); 
		    i.hasNext();) {
		    objguid = (SecureHash) i.next();
		    l.add(objguid);
		}
		int c=1;
		for (Iterator i = l.iterator(); i.hasNext();) {
		    objguid = (SecureHash) i.next();
		    domm = new
			DynamicObjMapMsg(msg.peer, my_node_id, objguid,
					(LinkedList)_backpointers.get(objguid),
					 l.size(), c);
		    dispatch(domm);
		    sent = true;
		    
		}
	
		// If no messages were sent, send a message saying we're done
		if (!sent) {
		    dispatch(new DebugRespMsg(msg.peer, true, DebugMsg.DONE, 
					      msg.requestID, _self_guid, 
					      "No Info"));
		}
	
		if (DEBUG) System.out.println("Back Pointers Sent.");
	    }
      
	} else if (msg.type == DebugMsg.REVERSE_PTR_REQUEST) {
	    if (_revpointers == null) {
		dispatch(new DebugRespMsg(msg.peer, true, DebugMsg.DONE, 
					  msg.requestID, _self_guid, 
					  "No Info"));
	    } else {
		boolean sent = false;
		HashMap currlevel;
		SecureHash s;
		for (int i = 0; i < msg.level; i++) {
		    currlevel = (HashMap) _revpointers.get(new Integer(i));
		    if (currlevel != null) {
			for (Iterator it = currlevel.keySet().iterator(); 
			     it.hasNext();) {
			    s = (SecureHash) it.next();
			    DebugRespMsg drm = 
				new DebugRespMsg(msg.peer, true, 
						 DebugMsg.BACK_PTR_REQUEST,
						 msg.requestID, _self_guid, 
						 "Sending Reverse Pointer",
						 s, (NodeId) currlevel.get(s), i);
			    dispatch(drm);
			    sent = true;
			}
		    }
		}
		
		// If no messages were sent, send a message saying we're done
		if (!sent) {
		    dispatch(new DebugRespMsg(msg.peer, true, DebugMsg.DONE,
					      msg.requestID, _self_guid, 
					      "No Info"));
		}
	
		if (DEBUG) System.out.println("Reverse Pointers Sent.");
	    }
	} else if (msg.type == DebugMsg.INSERTION_STATE) {
	    StringBuffer output = new StringBuffer("");;
	    if (_mcast != null) {
		HashMap nodeMap = _mcast.getNodeMap();
		SecureHash guid;
		NodeId addr;
		for (Iterator it = nodeMap.keySet().iterator(); 
		     it.hasNext(); ) {
		    guid = (SecureHash) it.next();
		    output.append("Inserting node w/ guid:" + guid + "\n");
		    NodeState ns = (NodeState) nodeMap.get(guid);
		    output.append("alpha=" + ns.alpha + ", parent=" + 
				  ns.parent+", addr=" + ns.addr + 
				  ", Waiting for acks from:");
		    if (ns.mcaddrs != null) {
			for (Iterator it2 = ns.mcaddrs.iterator(); 
			     it2.hasNext(); ) {
			    output.append(":"+it2.next());
			}
		    }
		    output.append(", waitingLocks:");
		    if (ns.locks != null) {
			for (Iterator it2 = ns.locks.iterator(); 
			     it2.hasNext(); ) {
			    output.append(":"+it2.next());
			}
		    }
		    output.append(", notifyReceived:"+ns.notified+"\n");
		}
	    }
	    output.append("InsertingLevel="+insertinglevel+", SeedCount="
			  +seedmap_nodes.size()+", initList="+initList);
	    dispatch(new DebugRespMsg(msg.peer, true, msg.type, msg.requestID,
				      _self_guid, output.toString(), null, 
				      null,-1));
	}
    }
  
    /**
     * A child is done with the multicast
     */
    private void handle_dynamic_dmack( DynamicDMAckMsg msg ) {

	final String tag = class_tag + ".handle_dynamic_dmack";

	if (DEBUG) Debug.printtagln(tag, "Handling DynamicDMAckMsg: "+msg);

	// record this acknowledgement
	try {
	    // reconstruct the 
	    DMCastUserData ud = new DMCastUserData( msg.peer, msg.newGuid,
						    -1, -1, -1 );
	    int nonce = findAssociatedNonce( ud );
	    NonceAckMsg nam = new NonceAckMsg( msg.peer );
	    nam.addNonceAck( nonce );
	    if( DEBUG ) {
		Debug.printtagln( tag, "Using " + msg + 
				  " as an acknowledgement." );
	    }
	    receivedAck( nam );
	} catch( NonceNotFoundException e ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Could not find nonce: " + e + 
				  ". Must have timed out.  Ignoring " + msg );
	    }
	    return;
	}

	boolean ack=false;
	try {
	    if (DEBUG)
		Debug.printtagln(tag, "Calling ackMCast with guid:"+msg.newGuid
				 +", after ack from node:"+msg.peer);
	    ack = _mcast.ackMCast(msg.newGuid, msg.src, msg.peer);
	} catch (MissingObjectException moe) {
	    if (DEBUG) {
		Debug.printtagln(tag, moe + ": "+msg);
	    }
	}
	NodeState ns = _mcast.getNodeState(msg.newGuid);
	if (ack && ns.notified && (ns.locks==null || ns.locks.isEmpty())) {
	    //all acks for this guid at this level are back
	    //we need to return an ack to our parent, and unlock this guy

	    ack_multicast( msg.newGuid, _mcast.getMCAddr(msg.newGuid) );

	}

    }

    /**
     * Asynchronous pings have finished, so resume the code
     */
    private void handle_quick_stat_resp( QuickStatResp msg ) {

	final String tag = class_tag + ".handle_quick_stat_resp";

	if (msg.userData instanceof DynamicRepairRespMsg) {
	  handle_repair_resp_msg_cont(msg);
	  return;
	} else if (msg.userData instanceof NodeStatusMsg) {
	  handle_nodestatus_cont(msg);
	  return;
	} else if (msg.userData instanceof GuidAddr []) {
	  measure_and_optimize_cont(msg);
	  return;
	}

	// do different things for the different types
	for( int i = 0; i < msg.peers.length; i++ ) {
	    NodeId currNode = msg.peers[i];
	    // comes back in microseconds, so convert
	    double latency = msg.stats[i].getOverallMean()/1000.0;
	    if( msg.failures[i] ) {
		// set latency to be very high
		latency = UNKNOWN_PING_VAL;
	    }
	    if( DEBUG ) {
		Debug.printtagln( tag, "Got asynch ping time for " + currNode +
				  " of " + latency + " ms." );
	    }
	    _ping_cache.put(currNode.address().getHostAddress(), 
			    new Double(latency));
	    if( msg.userData instanceof DynamicNotifyMsg ) {
		
		DynamicNotifyMsg dnm = (DynamicNotifyMsg) msg.userData;
		
		if( dnm.type == DynamicNotifyMsg.FILLHOLE_TYPE ) {
		    
		    // find the guid for this node
		    SecureHash guid = null;
		    for( int j = 0; j < dnm.list.length; j++ ) {
			if( dnm.list[j].getAddr().equals( currNode ) ) {
			    guid = dnm.list[j].getGuid();
			    break;
			}
		    }
		    if( guid == null ) {
			Carp.die( "handle_quick_stat_resp couldn't figure " +
				  "out the guid for " + currNode );
		    }
		    boolean sendNotify = currNode.equals(dnm.peer);
		    finish_do_notify( currNode, guid, latency,
				      dnm.minlevel, sendNotify );

		} else {

		    finish_do_notify( dnm.peer, dnm.src, latency,
				      dnm.minlevel, !dnm.internal );

		}

	    } else if( msg.userData instanceof Integer ) {
		if( msg.userData == APING_REINIT_NN ) {
		    _currentList_pings.put(currNode, new Double(latency));
		    if( _waiting_pings != null ) {
			_waiting_pings.remove( currNode );
		    } // else late, duplicate response.  Ignore.
		}
	    }
	}

	// per node processing is done, now finish up
	if( msg.userData instanceof Integer ) {
	    if( msg.userData == APING_REINIT_NN ) {
		if( _waiting_pings != null ) {
		    if( _waiting_pings.isEmpty() ) {
			_waiting_pings = null;
			_nearest_neighbors = new TreeMap();
			requestSeeds();
		    }// else wait for other pings.
		} // else late, duplicate response.  Ignore.
	    }
	}
    }

  void handle_repair_req_msg(DynamicRepairReqMsg msg) {
    final String tag = class_tag + ".handle_repair_req_msg";
    
    if (DEBUG_FT) 
      System.err.println(tag+":Handling DynamicRepairReqMsg: "+msg);
    
    RouteEntry re;
    NodeId next;
    TreeMap tm = new TreeMap();
    RouteEntry.NodeEntry ne;

    ArrayList al = new ArrayList(msg.number);

    if (msg.level == 1) {

      for (int i=0; i< RoutingTable.DIGIT_VALUES; i++) {
	re = _routing_tbl.getEntry(msg.digit+1, i);
	if (re != null) {
	  ne = re.getEntryAt(re.get_frls());
	  if (ne != null && !ne.address.equals(my_node_id)
	      && (ne.status==RouteEntry.STATUS_ALIVE)) {
	    tm.put(new Double(ne.latency), ne.address);
	  }
	}
      }
      
      //now take out of sorted list and send back
      int i =0;
      NodeId id;
      for (Iterator it = tm.values().iterator(); 
	   it.hasNext() && i<msg.number; i++) {
	id = (NodeId) it.next();
	al.add(new GuidAddr((SecureHash) 
			    _routing_tbl.node_id_to_guid(id), id));
      }
    } else if (msg.level == 2) {
      re = _routing_tbl.getEntry(msg.digit, msg.value);
      if (re != null) {
	for (int i=0; i< RouteEntry.REDUNDANT_LVL; i++) {
	  ne = re.getEntryAt(i);
	  if (ne != null && (ne.status == RouteEntry.STATUS_ALIVE)) {
	    al.add(new GuidAddr((SecureHash)
				_routing_tbl.node_id_to_guid(ne.address),
				ne.address));
	  }
	}
      } //end if (re != null)
    }
    if (al.size() > 0) 
      dispatch(new DynamicRepairRespMsg(msg.peer, _self_guid, al, 
					msg.digit, msg.value, msg.level));
    
    return;
  }

  void handle_repair_resp_msg(DynamicRepairRespMsg msg) {
    final String tag = class_tag + ".handle_repair_resp_msg";
    if (DEBUG_FT)
      System.err.println(tag+":Handling DynamicRepairRespMsg:"+msg);
    if (msg.al != null && !msg.al.isEmpty()) {
      NodeId [] listtoping;
      double mindist, currdist;
      mindist = Double.MAX_VALUE;

      //need to do first filter out nodes we already have
      RouteEntry entry = _routing_tbl.getEntry(msg.digit, msg.value);
      HashSet oldaddrs = new HashSet();
      for (int i = 0; entry != null && i<RouteEntry.REDUNDANT_LVL; i++) {
	RouteEntry.NodeEntry ne = entry.getEntryAt(i);
	if (ne != null) {
	  oldaddrs.add(ne.address);
	}
      }

      NodeId node;
      ArrayList al = new ArrayList();
      for (int i = 0; i<msg.al.size(); i++) {
	node = ((GuidAddr) msg.al.get(i)).getAddr();
	if (!oldaddrs.contains(node))
	  al.add(node);
      }
      listtoping = new NodeId [al.size()];
      al.toArray(listtoping);
      QuickStatReq qsr =
	new QuickStatReq(listtoping, Patch.TYPE_LATENCY,
			 PING_TIMEOUT, msg, my_sink);
      qsr.just_once = true;
      dispatch(qsr);
    } else {
      //got empty msg back, there are no good replacements
      if (DEBUG_FT)
	System.err.println(tag+":Repair message reports no replacements.");
    }
  }

  /**
   * Function to handle responses from patchwork.
   * Take all the latencies that succeeded, and choose the closest one
   * and put into the routing table
   */
  void handle_repair_resp_msg_cont(QuickStatResp msg) {
    final String tag = class_tag + ".handle_repair_req_msg_cont";
    
    double currlat;
    NodeId currnode;
    TreeMap sortedbylat = new TreeMap();
    for (int i=0; i<msg.peers.length; i++) {
      //find the closest node out of all the ones that connected
      if (!msg.failures[i]) {
	currnode = msg.peers[i];
	currlat = msg.stats[i].getOverallMean()/1000.0;
	sortedbylat.put(new Double(currlat), currnode);
      }
    }
    DynamicRepairRespMsg drrm = (DynamicRepairRespMsg) msg.userData;
    RouteEntry entry = _routing_tbl.getEntry(drrm.digit, drrm.value);
    if (entry != null) {
      //now we use the best out of the available results to 
      //fill in dead holes in our routes
      RouteEntry.NodeEntry ne;
      Double time;
      Iterator it = sortedbylat.keySet().iterator();
      GuidAddr ga;

      // old frls info
      int oldfrlspos = entry.get_frls();
      NodeId oldfrls = entry.getAtIndex( oldfrlspos );

      for (int i=0; i<entry.REDUNDANT_LVL; i++) {
	if (!it.hasNext()) break;
	ne = entry.getEntryAt(i);
	if (ne == null || (ne.status != RouteEntry.STATUS_ALIVE)) {
	  time = (Double) it.next();
	  currnode = (NodeId) sortedbylat.get(time);
	  Debug.printtagln(tag, "Replacing entry:"+ne+" with nodeid:"+
			   currnode+", lat:"+time);
	  //first find matching GUID of nodeid
	  SecureHash guid = null;
	  for (int j=0; j<drrm.al.size(); j++) {
	    ga = (GuidAddr) drrm.al.get(j);
	    if (currnode.equals(ga.getAddr())) {
	      guid = ga.getGuid();
	      _routing_tbl.put_node_id_to_guid(currnode, guid);
	      break;
	    }
	  }
	  entry.setEntryAt(i, new RouteEntry.NodeEntry(currnode, time, 
						       false, false));
	  // set the time in the routing table as well
	  _routing_tbl.addDistance( time.doubleValue(), currnode, guid );
	}
      }  

      // recalculate the frls
      Pair p = entry.recalcFRLS();
      if( DEBUG ) {
	  Debug.printtagln( tag, "RouteEntry at (" + drrm.digit + "," +
			    drrm.value + ") is now " + entry );
      }

      int newfrlspos = entry.get_frls();
      NodeId newfrls = entry.getAtIndex( newfrlspos );
      // if the frls node changed, send out this message to trigger 
      // republishes
      if( (oldfrls == null && newfrls != null) ||
	  (oldfrls != null && newfrls == null) ||
	  (oldfrls != null && newfrls != null && !oldfrls.equals(newfrls)) 
	  ) {
	  dispatch(new FRLSChangeMsg(drrm.digit, drrm.value,
				     oldfrlspos, newfrlspos));
      }
    
    } else {
      if (DEBUG_FT)
	Debug.printtagln(tag, "Unexpected missing route on repair request:"+
			 drrm.digit+":"+drrm.value);
    }
  }
  
  /**
   * Function to choose a neighbor to request new neighbors from.
   * @param int Int that indicates how desperate we are for new neighbors
   * @param Triple The index of the entry to be repaired
   * @return nodeId [], an array of nodes to contact, null if none found
   */
  NodeId [] chooseRepairNeighbors (int urgency, Triple index, NodeId dead) {
    RouteEntry.NodeEntry nentry = null;
    RouteEntry re;
    NodeId [] requestSet;
    if (urgency == 1) {
      re = _routing_tbl.getEntry(index.level, index.value);
      //now figure out who to ask for more neighbors
      for (int i=0; i<RouteEntry.REDUNDANT_LVL; i++) {
	nentry = re.getEntryAt(i);
	if (nentry == null) continue;
	if (!nentry.address.equals(dead) && 
	    nentry.status == RouteEntry.STATUS_ALIVE)
	  break;
      }
      requestSet = new NodeId[1];
      if (nentry == null) {
	requestSet = null;
      } else requestSet[0] = nentry.address;
    } else if (urgency == 2) {
      HashSet hs = new HashSet();
      for (int i = RoutingTable.DIGITS_PER_GUID - 1; i >= index.level; i--) {
	for (int j = 0; j < RoutingTable.DIGIT_VALUES; j++) {
	  re = _routing_tbl.getEntry(i, j);
	  if (re == null) continue;
	  nentry = re.getEntryAt(re.get_frls());
	  if (nentry != null && !nentry.address.equals(my_node_id)
	      && nentry.status == RouteEntry.STATUS_ALIVE) {
	      if( DEBUG ) {
		  Debug.printtagln( class_tag + ".chooseRepairNeighbors",
				    "Adding " + nentry.address + 
				    " to neighbors");
	      }
	      hs.add(nentry.address);
	  }
	}
	/**
	// try to avoid going to the same level as the null entry if possible,
	// because that one could have a lot of neighbors.
	if( hs.size() > 0 && i == index.level+1 ) {
	    break;
	}
	*/
      }
      if (hs.size() == 0) requestSet = null;
      else {
	requestSet = new NodeId [hs.size()];
	hs.toArray (requestSet);
      }
    } else {
      requestSet = null;
    }

    return requestSet;
  }


  /**
   * Function to handle node failure notifications from Patchwork
   * The algorithm is as follows:  When we know an entry is dead, we
   * immediately request a neighbor replacement from someone with the
   * same prefix requirement.  If the dead node is in our routing table
   * in level A, value B, then we ask all other entries at (A,B) to send us
   * replacements from their routing level A+1.  If all entries at (A,B)
   * are dead, then we need to talk to nodes at level A, and request
   * all their entries at value B.
   * @param msg The message we got from patchwork indicating node status
   */
  void handle_nodestatus(NodeStatusMsg msg) {

    //if listening to status messages from patchwork is disabled, return;
    if (!LISTEN_STATUS)
      return; 

    final String tag = class_tag + ".handle_nodestatus";
    if( DEBUG_FT ) {
	Debug.printtagln(tag, "*****Received PatchWork msg: "+msg);
    }

    if (msg.status == msg.STATUS_ALIVE) {
      Triple index = (Triple) msg.userData;
      RouteEntry.NodeEntry nentry =
	(RouteEntry.NodeEntry) deadEntries.get(index);
      if (nentry == null) {
	if (DEBUG_FT)
	  //this is ok, happens when you start monitoring a node that has died before
	  //and you start monitoring it before patchwork has reported that it's alive again
	  Debug.printtagln(tag, "Unexpected missing dead entry w/ key:"+index);
	return;
      }
      QuickStatReq qsr =
	new QuickStatReq(nentry.address, Patch.TYPE_LATENCY, msg, my_sink);
      qsr.just_once = true;
      dispatch(qsr);

    } else if (msg.status == msg.STATUS_DEAD) {
      // status is dead
      // do not dispatch stop msg, just add to deadlist and wait for wakeup
      Triple index = (Triple) msg.userData;
      RouteEntry re = _routing_tbl.getEntry(index.level, index.value);
      RouteEntry.NodeEntry nentry = null;

      // make sure to remove the ping time from the cache
      _ping_cache.remove( msg.peer );
      
      boolean found = false;
      //scan through entry for the dead node
      for (int i=0; i<RouteEntry.REDUNDANT_LVL; i++) {
	nentry = re.getEntryAt(i);
	if (nentry == null) continue;
	if( DEBUG_FT ) {
	    Debug.printtagln(tag, "NodeCompare " + nentry.address + 
			     " to "+msg.peer);
	}
	if (nentry.address.equals(msg.peer)) {
	  found = true;
	  nentry.status = RouteEntry.STATUS_REPAIR_ONE;

	  //now put the dead entry into a dead list, and clear it out
	  //of the RT.
	  deadEntries.put(index, nentry);
	  if (DEBUG_FT)
	    Debug.printtagln(tag, "Adding "+nentry+
			     " to deadlist with key "+index);
	  int oldfrlspos = re.get_frls();
	  NodeId oldfrls = re.getAtIndex( oldfrlspos );
	  re.setEntryAt(i, null);
	  //if clearing out of routing table, then remove from RevPtr list
	  remove_reverse_ptr(msg.peer);

	  //now move entries up, and recalcFRLS
	  for (int j=i; j<RouteEntry.REDUNDANT_LVL-1; j++) {
	    re.setEntryAt(j, re.getEntryAt(j+1));
	    re.setEntryAt(j+1, null);
	  }
	  Pair p = re.recalcFRLS();
	  if( DEBUG ) {
	      Debug.printtagln( tag, "RouteEntry at (" + index.level + "," +
				index.value + ") is now " + re );
	  }
	  int newfrlspos = re.get_frls();
	  NodeId newfrls = re.getAtIndex( newfrlspos );
	  // if the frls node changed, send out this message to trigger 
	  // republishes
	  if( (oldfrls == null && newfrls != null) ||
	      (oldfrls != null && newfrls == null) ||
	      (oldfrls != null && newfrls != null && !oldfrls.equals(newfrls)) 
	      ) {
	      dispatch(new FRLSChangeMsg(index.level, index.value,
					 oldfrlspos, newfrlspos));
	  }
	  //TODO: deadEntries does not support replacement via LRU yet
	  //when node is kicked out, stop monitoring it
	  break;
	}
      }
      
//      System.out.println("Finished handle_node_status=DEADA");
      if (!found) {
	if (!deadEntries.containsKey(index)) {
	  if (DEBUG_FT) {
	    Debug.printtagln(tag, "WEIRD: *****Unexpected missing nentry at: "+
			     index);
	    Debug.printtagln(tag, "deadEntries:");
	    for (Iterator it=deadEntries.keySet().iterator(); it.hasNext();) {
	      Debug.printtagln(tag, ""+(Triple) it.next());
	    }
	    Debug.printtagln(tag, "Routing table:"+
			     _routing_tbl.toShortString());
	  }
	  //if we can't find it int our table, we tell patchwork to stop it
	  dispatch(new MonitorStopMsg(index.addr, index, my_sink));
	}
	return;
      }

//      System.out.println("Finished handle_node_status=DEADB");
      NodeId [] neighbors = chooseRepairNeighbors(1, index, msg.peer);
      int level = 1;
      if (neighbors != null) {
	_waiting_repair_nonces.put(index, 
				   new Pair(new Integer(neighbors.length),
					    msg.peer));
      } else {
	// (neighbors == null), go to level 2
	neighbors = chooseRepairNeighbors(2, index, msg.peer);
	level = 2;
	if (neighbors == null)
	  //should go to level 3 (whatever that is)
	  return;
      }
//      System.out.println("Finished handle_node_status=DEADC");
      for (int i = 0; i<neighbors.length; i++) {
	DynamicRepairReqMsg drrm =
	  new DynamicRepairReqMsg(neighbors[i], _self_guid, index.level,
				  index.value, level, 5); //start with 5
	if (level == 1) {
	  sendNonceMsg(drrm, REPAIRREQ_TIMEOUT, index);
	} else if (level == 2) {
	  dispatch(drrm);
	}
	if( DEBUG_FT ) {
	  Debug.printtagln(tag, "*****Sending DynamicRepairRequest: "+drrm);
	}
      }
//      System.out.println("Finished handle_node_status=DEADD");
    }
  }

  private void handle_nodestatus_cont (QuickStatResp msg) {
    NodeStatusMsg nsm = (NodeStatusMsg) msg.userData;
    Triple index = (Triple) nsm.userData;

    double latency = msg.stats[0].getOverallMean()/1000.0;
    RouteEntry.NodeEntry nentry = 
      new RouteEntry.NodeEntry(index.addr, new Double(latency), false, false);

    // remove from list of dead entries
    deadEntries.remove( index );

    if (nentry == null) {
      if (DEBUG_FT)
	Debug.printtagln(tag, "*****Missing expected dead entry*****");
      return;
    }
    
    RouteEntry entry = _routing_tbl.getEntry(index.level, index.value);
    if (entry != null) {
      //fill in vacancy
      for (int i=0; i<entry.REDUNDANT_LVL; i++) {
	if (entry.getEntryAt(i) == null ||
	    entry.getEntryAt(i).status != RouteEntry.STATUS_ALIVE) {
	  if (DEBUG_FT)
	    Debug.printtagln(tag, "Reinserting entry:"+nentry+
			     " at level "+index.level+", value "+index.value);
	  entry.setEntryAt(i, nentry);
	  // for the love of god, only insert it once!
	  break;
	}
      }
    } else {
      entry = new RouteEntry();
      entry.setEntryAt(0, nentry);
      _routing_tbl.putEntry(index.level, index.value, entry);
    }
  }

  private void handle_periodic_reminder(PeriodicReminder pr) {
    //send out for next round
    classifier.dispatch_later(pr, 100000);

    if (status != ATTACHED || my_node_id == null) return;
    if (pr.requestId == PeriodicReminder.ONE) {
      //want to generate a watchlist of sorts to pass to neighbors
      HashSet watchlist = new HashSet();
      HashSet dests = new HashSet();
      int maxlevel = 0;
      RouteEntry re;
      for (int i=0; i<RoutingTable.DIGITS_PER_GUID; i++) {
	for (int j=0; j<RoutingTable.DIGIT_VALUES; j++) {
	  re = _routing_tbl.getEntry(i,j);
	  if (re != null) {
	    if (re.getFRLS() != null &&
		!my_node_id.equals(re.getFRLS()))
	      maxlevel = i;
	  }
	}
      }

      //now that we know the maxlevel, we collect holes all the way up to it
      for (int i=0; i<= maxlevel; i++) {
	for (int j=0; j<RoutingTable.DIGIT_VALUES; j++) {
	  if ((re = _routing_tbl.getEntry(i,j)) == null ||
	      re.getFRLS() == null)
	    watchlist.add(new Pair(new Integer(i), new Integer(j)));
	  else if (i == maxlevel) 
	    dests.add(re.getFRLS());
	}
      }

      for (Iterator i = dests.iterator(); i.hasNext(); ) {
	PeriodicRoutePatch prp = 
	  new PeriodicRoutePatch(PeriodicRoutePatch.REQUEST,
				 (NodeId) i.next(), _self_guid, watchlist);
	dispatch(prp);
      }
    }
    //ignore other types for now
  }

  private void handle_periodic_routepatch(PeriodicRoutePatch prp) {
    if (prp.type == prp.REQUEST) {

      HashSet filler = new HashSet();
      Pair p;
      RouteEntry re;
      NodeId nid;
      for (Iterator it = prp.set.iterator(); it.hasNext(); ) {
	p = (Pair) it.next();
	re = _routing_tbl.getEntry(((Integer) p.first()).intValue(),
				   ((Integer) p.second()).intValue());
	if (re != null && re.getFRLS() != null) {
	  nid = re.getFRLS();
	  filler.add(new GuidAddr(_routing_tbl.node_id_to_guid(nid),nid));
	}
      }
      //send back response
      dispatch(new PeriodicRoutePatch(PeriodicRoutePatch.RESPONSE,
				      prp.peer, _self_guid, filler));

      GuidAddr [] set = new GuidAddr[1];
      set[0] = new GuidAddr(prp.srcguid, prp.peer);
      measure_and_optimize(set);

    } else if (prp.type == prp.RESPONSE) {
      GuidAddr [] set = new GuidAddr[prp.set.size()];
      prp.set.toArray(set);
      measure_and_optimize(set);

    } else 
      System.err.println("ERROR: unrecognized type of PeriodicRoutePatch");
  }

  private void measure_and_optimize(GuidAddr [] ga) {
    NodeId [] peers = new NodeId[ga.length];
    
    for (int i = 0; i<ga.length; i++) {
      peers[i] = ga[i].getAddr();
    }
    QuickStatReq qsr = new QuickStatReq(peers, Patch.TYPE_LATENCY, 
					PING_TIMEOUT, ga, my_sink);
    qsr.just_once = true;
    dispatch(qsr);
  }

  private void measure_and_optimize_cont(QuickStatResp msg) {
    LinkedList ll;
    double latency;
    GuidAddr [] guids = (GuidAddr []) msg.userData;
    int changed_level;
    for (int i=0; i<msg.failures.length; i++) {
      if (msg.failures[i] == true)
	continue;
      latency = msg.stats[i].getOverallMean()/1000.0;
      add_node_to_rt(msg.peers[i], guids[i].getGuid(), 0, //minlevel =0
		     latency);
      //we just call add_node_to_rt, which does all the dirty work for us
    }
  }

    /**
     * Send an incoming event to the correct handler
     */
    public void handle_event(QueueElementIF item) 
	throws EventHandlerException {

	super.handle_event( item );

	try {
	    final String tag = class_tag + ".handle_event";
	    
	    if (DEBUG) {
		if (item instanceof RoutingTable.Available)
		    Debug.printtagln(tag, "got "+item.getClass());
		else 
		    Debug.printtagln (tag, "got " + item.getClass () + 
				      " event: " + item + ".");
	    }

	    if (item instanceof StagesInitializedSignal) {
		handle_stages_init_signal ();
		
	    } else if (item instanceof NodeGuid) {
		handle_node_guid ((NodeGuid)item);
		
	    } else if (item instanceof RoutingTable.Available) {
		RoutingTable.Available rta = (RoutingTable.Available) item;
		handle_routing_table_available (rta);
		
	    } else if (item instanceof Router.ObjectMap) {
		handle_object_map((Router.ObjectMap) item);

	    } else if (item instanceof ModifyObjLevelsMsg) {
		ModifyObjLevelsMsg molm = (ModifyObjLevelsMsg) item;
		modify_objlevels( molm.obj_guid, molm.level, molm.publish );

	    } else if (item instanceof QuickStatResp) {
		handle_quick_stat_resp((QuickStatResp) item);
		
	    } else if (item instanceof DynamicGatewayMsg) {
		handle_gateway( (DynamicGatewayMsg) item );
		
	    } else if (item instanceof DynamicSurrogateMsg) {
		handle_dynamic_surrogate_msg((DynamicSurrogateMsg) item);
		
	    } else if (item instanceof DynamicMapMsg) {
		handle_dynamic_map( (DynamicMapMsg) item );
		
	    } else if (item instanceof DynamicObjReqMsg) {
		handle_dynamic_obj_req( (DynamicObjReqMsg) item );
		
	    } else if (item instanceof DynamicObjMapMsg) {
		handle_dynamic_obj_map( (DynamicObjMapMsg) item );
		
	    } else if (item instanceof DynamicNotifyMsg) {
		DynamicNotifyMsg msg = (DynamicNotifyMsg) item;
		//msg from outside, not internal
		handle_dynamic_notify(msg, false);
	    } else if (item instanceof DynamicDMCastMsg) {
		DynamicDMCastMsg msg = (DynamicDMCastMsg) item;
		handle_dynamic_dmcast(msg);
	    } else if (item instanceof DynamicDMAckMsg) {
		handle_dynamic_dmack( (DynamicDMAckMsg) item );
		
	    } else if (item instanceof DynamicMCReqMsg) {
		DynamicMCReqMsg msg = (DynamicMCReqMsg) item;
		handle_dynamic_mcreq(msg);
		
	    } else if (item instanceof DynamicSeedMsg) {
		DynamicSeedMsg msg = (DynamicSeedMsg) item;
		handle_dynamic_seed(msg);
		
	    } else if (item instanceof DynamicSeedMapMsg) {
		DynamicSeedMapMsg msg = (DynamicSeedMapMsg) item;
		handle_dynamic_seedmap(msg);
		
	    } else if (item instanceof TapestryNodeDetachMsg) {
		handle_node_detach();
		
	    } else if (item instanceof DynamicBeginDelete) {
		DynamicBeginDelete msg = (DynamicBeginDelete) item;
		handle_begin_delete(msg);
		
	    } else if (item instanceof DynamicEndDelete) {
		DynamicEndDelete msg = (DynamicEndDelete) item;
		handle_end_delete(msg);
		
	    } else if (item instanceof DynamicPtrMsg) {
		handle_dynamic_ptrmsg((DynamicPtrMsg) item);
		
	    } else if (item instanceof TapestryIntegrateMsg) {
		TapestryIntegrateMsg msg = (TapestryIntegrateMsg) item;
		handle_integrate(msg);
		
	    } else if (item instanceof PartialRepublish) {
		PartialRepublish msg = (PartialRepublish) item;
		handle_partialrepub(msg);

	    } else if (item instanceof NodeStatusMsg) {
	      NodeStatusMsg msg = (NodeStatusMsg) item;
	      handle_nodestatus(msg);
	    } else if (item instanceof FRLSChangeMsg) {
		handle_frls_change((FRLSChangeMsg) item);
		
	    } else if (item instanceof DebugMsg) {
		DebugMsg msg = (DebugMsg) item;
		handle_debugmsg(msg);
	    } else if (item instanceof DynamicRepairReqMsg) {
		//Added by Ben Y. Zhao and Ling Huang for fault 
		//tolerance routing.
		handle_repair_req_msg((DynamicRepairReqMsg) item);
	    } else if (item instanceof DynamicRepairRespMsg) {
		handle_repair_resp_msg((DynamicRepairRespMsg) item);
	    } else if (item instanceof PeriodicReminder) {
	      handle_periodic_reminder((PeriodicReminder) item);

	    } else if (item instanceof NonceMsgDelay) {
		NonceMsgDelay nmd = (NonceMsgDelay) item;
		sendNonceMsg( nmd.nm, nmd.timeout, nmd.ud );
	    } else if (item instanceof PeriodicRoutePatch) {
	      handle_periodic_routepatch((PeriodicRoutePatch) item);

	    } else if( item instanceof NonceAckMsg ) {
		// ignore, for AckStage
	    } else if( item instanceof ostore.dispatch.Signal.AlarmSignal ) {
		// ignore, for AckStage

	    } else if (DEBUG)
		Carp.die( "ERROR in DynamicTClient: unrecognized " + 
			  "incoming message:" + item);
	} catch(Exception e) {
	    if(status != DETACH_ROUTINGOBJ || status != DETACH_FINALACK ||
	       status != NOT_ATTACHED) {
		Carp.dieGracefully(e, my_node_id);
	    } else {
		Carp.setThrowException(false);
		Carp.die(e);
	    }
	}
    }

    protected void timeoutCallback( NonceMsg nm, Object userData ) {
	final String tag = class_tag + ".timeoutCallback";

	if( DEBUG ) {
	    Debug.printtagln( tag, "Got timeout callback for " + nm + 
			      ", userdata: " + userData );
	}

	if( userData instanceof Pair ) {

	    Pair pair = (Pair) userData;
	    // which one was it?

	    if( pair.first instanceof Integer && 
		pair.first.equals( USERDATA_GATEWAY ) ) {

		////////////////////////////////////////////
		//  DynamicGatewayMsg + DynamicSurrogateMsg

		int position = ((Integer)pair.second).intValue();
		handle_gateway_timeout( nm, position );

	    } else if( pair.first instanceof Integer && 
		       pair.first.equals( USERDATA_PTR ) ) {

		////////////////////////////////////////////
		//  DynamicPtrMsg
		handle_ptr_timeout( nm, ((Integer)pair.second).intValue() );

	    } else if( pair.first instanceof Integer && 
		       pair.first.equals( USERDATA_NOTIFY ) ) {

		////////////////////////////////////////////
		//  DynamicNotifyMsg
		handle_notify_timeout( (DynamicNotifyMsg) nm, 
					((Integer)pair.second).intValue() );

	    } else if( pair.first instanceof Integer && 
		       pair.first.equals( USERDATA_SEED ) ) {

		////////////////////////////////////////////
		//  DynamicSeedMsg
		handle_seed_timeout( (DynamicSeedMsg) nm, 
				     ((Integer)pair.second).intValue() );


	    } else {
		if( DEBUG ) {
		    Debug.printtagln( tag, 
				      "Didn't recognize pair, ignoring." );
		}
	    }

	} else if( userData instanceof DMCastUserData ) {

	    ////////////////////////////////////////////
	    //  DynamicDMCast + DynamicDMAck

	    DMCastUserData ud = (DMCastUserData) userData;
	    handle_dmcast_timeout( nm, ud );

	} else if (userData instanceof Triple) {
	  //this is from a dynamicrepairmsg
	  Triple index = (Triple) userData;
	  Pair p = (Pair) _waiting_repair_nonces.remove(userData);
	  // for some reason it's not in there
	  if( p == null ) {
	      return;
	  }
	  Integer count = (Integer) p.first();
	  NodeId peer = (NodeId) p.second();
	  int c = count.intValue();
	  c--;
	  if (c == 0) {
	    //means all requests have failed
	    NodeId [] neighbors = chooseRepairNeighbors(2, index, peer);
	    if (neighbors == null)
	      return;
	    for (int i = 0; i<neighbors.length; i++) {
	      DynamicRepairReqMsg drrm = 
		new DynamicRepairReqMsg(neighbors[i], _self_guid, 
					index.level, index.value, 
					2, /*level 2*/  5 /*5 entries pls*/);
	      if( DEBUG_FT ) {
		Debug.printtagln(tag, "*****Sending DynamicRepairRequest: "
				 +drrm);
	      }
	      dispatch(drrm);
	    }
	  }
	} else if( userData instanceof Integer && 
		   nm instanceof DynamicMCReqMsg  ) {
	    
	    handle_mcreq_timeout( (DynamicMCReqMsg) nm, 
				  ((Integer) userData).intValue() );

	} else {
	    if( DEBUG ) {
		Debug.printtagln( tag, 
				  "Didn't recognize userData, ignoring." );
	    }
	} 

    }

    /**
     * Called when a DynamicMCReqMsg was not acked
     */
    private void handle_mcreq_timeout( DynamicMCReqMsg msg, int retries ) {

	final String tag = class_tag + ".handle_mcreq_timeout";

	// try up to five times, then forget about it
	if( retries >= 4 ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Giving up on " + msg );
	    }
	    return;
	} else {
	    retries++;
	    sendNonceMsg( msg, SHORT_TIMEOUT, new Integer(retries) );
	}

    }

    /**
     * Called when a Gateway (or Surrogate) mesage times out
     */
    private void handle_gateway_timeout( NonceMsg nm, int position ) {

	// if this is a timeout for the DynamicGatewayMsg
	// (and not the surrogate message), then cancel the timeout
	// for the surrogate message
	if( nm instanceof DynamicGatewayMsg ) {
	    DynamicGatewayMsg dgm = (DynamicGatewayMsg) nm;
	    NonceAckMsg nam = new NonceAckMsg( my_node_id );
	    nam.addNonceAck( dgm.surr_nonce );
	    receivedAck( nam );
	}
	
	// if we've tried contacting all the gateways twice,
	// give up
	position++;
	if( position >= GATEWAY_TIMEOUT_RETRIES * _gate_node_id.length ) {
	    Carp.die( "Tried all gateways "+GATEWAY_TIMEOUT_RETRIES
		      +" times, dying." );
	} else {
	    // does it matter if we send more than one out at a time?
	    send_gateway_msg(position);
	}
    }

    /**
     * Called when a DynamicPtr messages times out
     */
    private void handle_ptr_timeout( NonceMsg nm, int retries ) {

	final String tag = class_tag + ".handle_ptr_timeout";

	// if we've already retried twice, just give up
	if( retries >= 2 ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Giving up on " + nm );
	    }
	    // TODO: should I remove the node from my routing table here?
	    // but what about if it's a remove message?  I shouldn't 
	    // put it back in . . .
	    return;
	} else {
	    retries++;
	    sendNonceMsg( nm, SHORT_TIMEOUT, new Pair( USERDATA_PTR, 
						       new Integer(retries)));
	}

    }

    /**
     * Called when a DynamicNotify messages times out
     */
    private void handle_notify_timeout( DynamicNotifyMsg msg, int retries ) {

	final String tag = class_tag + ".handle_notify_timeout";

	// if we've already retried twice, just give up
	if( retries >= 2 ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Giving up on " + msg );
	    }
	    // if it was a message that expected a response,
	    // pretend the response came back in order to clear state
	    if (msg.type != DynamicNotifyMsg.RESPONSE_TYPE && 
		_mcast.setNotify(msg.src)) {
		ack_multicast( msg.src, msg.peer );	
	    }
	    return;
	} else {
	    retries++;
	    sendNonceMsg( msg, SHORT_TIMEOUT, new Pair( USERDATA_NOTIFY, 
							new Integer(retries)));
	}

    }

    /**
     * Called when a DynamicSeed messages times out
     */
    private void handle_seed_timeout( DynamicSeedMsg msg, int retries ) {

	final String tag = class_tag + ".handle_seed_timeout";

	// if we've already retried twice, just give up
	if( retries >= 2 ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Giving up on " + msg );
	    }
	    // pretend we got it and go on with our lives
	    seedmap_nodes.remove(msg.peer);
	    if( seedmap_nodes.isEmpty() ) {
		reinitialize_nn();
	    }
	    return;
	} else {
	    retries++;
	    sendNonceMsg( msg, SHORT_TIMEOUT, new Pair( USERDATA_SEED, 
							new Integer(retries)));
	}

    }

    /**
     * Called when a DMCast message times out
     */
    private void handle_dmcast_timeout( NonceMsg nm, DMCastUserData ud ) {

	final String tag = class_tag + ".handle_dmcast_timeout";

	// if it's a timeout on the DynamicDMCast, cancel the timeout
	// for the DynamicDMAck
	if( nm instanceof DynamicDMCastMsg ) {
	    NonceAckMsg nam = new NonceAckMsg( my_node_id );
	    nam.addNonceAck( ud.ack_nonce );
	    receivedAck( nam );
	}
	
	boolean stillInTable = false;
	RouteEntry entry = null;
	if( ud.digit != -1 ) {
	    // routing table dmcast
	    entry = _routing_tbl.getEntry( ud.level, ud.digit );
	    if( entry != null && entry.getAtIndex(ud.entry/2) != null &&
		entry.getAtIndex(ud.entry/2).equals( ud.peer ) ) {
		stillInTable = true;
	    }
	} else {
	    // lock dmcast
	    Object lockindex = getLockIndex( ud.newId );
	    TreeSet templocks = 
		(TreeSet) _mcast.indexToLocks.get(lockindex);
	    for (Iterator it=templocks.iterator(); it.hasNext(); ) {
		SecureHash itguid = (SecureHash) it.next();
		if (itguid.equals(ud.newId))
		    continue;
		if( _mcast.getNodeState(itguid) != null && 
		    _mcast.getNodeState(itguid).addr != null && 
		    _mcast.getNodeState(itguid).addr.equals( ud.peer ) ) {
		    // TODO: what does it mean if the above null checks fail?
		    stillInTable = true;
		    break;
		}
	    }
	}
	// the retry policy is that is to give nodes two chances, if
	// they are in the routing table still.  Otherwise, move
	// to a backup.  If there is no backup, die.
	ud.entry++;
	NodeState ns = _mcast.getNodeState(ud.newId);
	if( ns == null ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "ns is null, so cleanSlate must " +
				  "have already happened. Ignoring." );
	    }
	    return;
	}
	if( ud.entry % 2 != 0 && stillInTable ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Giving " + ud.peer + 
				  " another chance for a dmcast" );
	    }
	    // it's the first try, so give it one more shot
	    // don't need to set up any new multicast state
	    ud.ack_nonce = forceRegisterNonce( LONG_TIMEOUT, ud );
	    sendNonceMsg( ud.msg, SHORT_TIMEOUT, ud );
	} else if( ud.digit != -1 && stillInTable && 
		   ud.entry/2 < entry.size() && 
		   entry.getAtIndex(ud.entry/2) != null ) {
	    // we've given up on this guy, move to the next
	    NodeId old_peer = ud.msg.peer;
	    ud.msg.peer = entry.getAtIndex(ud.entry/2);
	    ud.peer = ud.msg.peer;
	    // add multicast state
	    _mcast.addMCast(ud.newId, ud.msg.peer);
	    ns.addMCast(ud.msg.peer);
	    // erase the multicast state used by the last person
	    if (DEBUG) {
		Debug.printtagln(tag, "Faking ackMCast with guid:"
				 + ud.newId + ", after timeout on node:" + 
				 old_peer);
	    }
	    try {
		_mcast.ackMCast(ud.newId, ud.msg.src, old_peer);
	    } catch (MissingObjectException moe) {
		if (DEBUG) {
		    Debug.printtagln(tag, moe.toString());
		}
	    }
	    ud.ack_nonce = forceRegisterNonce( LONG_TIMEOUT, ud );
	    sendNonceMsg( ud.msg, SHORT_TIMEOUT, ud );
	} else {
	    // otherwise you've reached the end, or the routing entry
	    // changed, or the node has been removed from the locks
	    // in this case, just give up entirely
	    // TODO: is this the right behavior?
	    boolean ack = false;
	    if (DEBUG) {
		Debug.printtagln(tag, "Faking ackMCast with guid:"
				 + ud.newId + ", after timeout on node:" + 
				 ud.peer + "; giving up!");
	    }
	    try {
		ack = _mcast.ackMCast(ud.newId, ud.msg.src, ud.peer);
	    } catch (MissingObjectException moe) {
		if (DEBUG) {
		    Debug.printtagln(tag, moe.toString());
		}
	    }
	    if (ack && ns.notified && (ns.locks==null || 
				       ns.locks.isEmpty())) {
		//all acks for this guid at this level are back
		//we need to return an ack to our parent, 
		// and unlock this guy
		ack_multicast( ud.newId, _mcast.getMCAddr(ud.newId) );
	    }
	}

    }

    /**
     * Extract ping information from the APP file, and cache them
     */
    private void parse_all_pings(String allpairsfile) {
	final String tag = class_tag + ".parse_all_pairs_pings";
	if (DEBUG)
	    Debug.printtagln (tag, "parsing the all_pairs_pings file" );
    
	BufferedReader br = null;
	try {
	    br = new BufferedReader( new FileReader(allpairsfile) );
	} catch( FileNotFoundException fnfe ) {
	    Carp.die( "Could not open all_pairs_pings_file: " + fnfe );
	}
    
	// first, we read in a line containing the names of all servers
	StringTokenizer serversST = null;
	try {
	    // throw away time
	    br.readLine();
	    // throw away ping count
	    br.readLine();
	    serversST = new StringTokenizer( br.readLine() );
	} catch( IOException ioe ) {
	    Carp.die( "Could not read line from all_pairs_pings file: " + 
		      ioe);
	}
	String[] servers = new String[ serversST.countTokens() ];
	// need a map to index servers to number
	HashMap server_indices = new HashMap();
	for( int i = 0; i < servers.length; i++ ) {
	    servers[i] = serversST.nextToken();
	    NodeId currNode = null;
	    try {
		currNode = new NodeId( servers[i] + ":3630" );
	    } catch( Exception e ) {
		Carp.die( "Exception creating NodeId for " + servers[i] +
			  ": " + e );
	    }
	    server_indices.put( currNode.address(), new Integer(i) );
	    if( DEBUG )
		Debug.printtagln( tag, "creating entry for " + currNode );
	}
    
	// now we need to find the line corresponding to our pings
	Integer bi = (Integer) server_indices.get( my_node_id.address() );
	if( bi == null ) {
	    Carp.die( "Could not find ourselves in the all_pairs_ping file" );
	}
    
	int self_index = bi.intValue();
    
	StringTokenizer pingsST = null;
	try {
	    for( int j = 0; j < self_index; j++ ) {
		br.readLine();
	    }
	    pingsST = new StringTokenizer( br.readLine() );
	} catch( IOException ioe ) {
	    Carp.die( "Could not read line from all_pairs_pings file: " + 
		      ioe );
	}
    
	// parse out our ping times
	double pings[] = new double[ pingsST.countTokens() ];
	for( int j = 0; j < pings.length; j++ ) {
	    // read out av ping time (the one in the middle)
	    StringTokenizer slashST = new StringTokenizer( pingsST.nextToken(),
							   "/" );
	    // throw away min time
	    slashST.nextToken();
	    // now get get av time
	    pings[j] = Double.parseDouble( slashST.nextToken() );
	}
    
	// they gotta be the same length
	if( pings.length != servers.length ) {
	    Carp.die( "The length of the pings row (" + pings.length +
		      ") does not equal the length of the servers row (" +
		      servers.length + ")." );
	}

	for (int i = 0; i<pings.length; i++) {
	    _ping_cache.put(servers[i], new Double(pings[i]));
	}
    }

    /**
     * Get the ping time to a node.
     * TODO: make asynchronous somehow
     */
    public double getPingTime(NodeId remoteHost) {
	final String tag = class_tag + ".getPingTime";
	double pingtime;
	if (simulator) {
	    if (_ping_cache.containsKey(remoteHost)) {
		pingtime = ((Double) 
			    _ping_cache.get(remoteHost)).doubleValue();
		if (DEBUG)
		    Debug.printtagln(tag, "Retrieved ping time from cache " + 
				     "for node: " + remoteHost + 
				     ", value in ms is:"+pingtime);
		return pingtime;
	    }

	    pingtime = (double) net.route_time_ms(my_node_id, 
						  remoteHost, 1);
	    _ping_cache.put(remoteHost, new Double(pingtime));
	    return pingtime;
	}

	String addr = remoteHost.address().getHostAddress();
	if (_ping_cache.containsKey(addr)) {
	    pingtime = ((Double) _ping_cache.get(addr)).doubleValue();
	    if (DEBUG)
		Debug.printtagln(tag, "Retrieved ping time from cache " + 
				 "for node: " + addr + ", value in ms is:" + 
				 pingtime);
	    return pingtime;
	}
	return -1;
    }

    /**
     * Have Patchwork ping some nodes asychronously
     */
    private void startAPings( NodeId[] remoteHosts, Object whereToFinish ) {
	if( (whereToFinish instanceof Integer) &&
	    (whereToFinish == APING_REINIT_NN) ) {
	    // We had a bug caused by poor handling of QuickStatReqs
	    // with multiple peers.  Now we send one req per
	    // remoteHost, but must wait for all QuickStatResps before
	    // continuing the nearest-neighbors initialization.
	    _waiting_pings = new HashMap();
	}
	for( int i=0; i<remoteHosts.length; i++ ) {
	    NodeId remoteHost = remoteHosts[i];
	    if( (whereToFinish instanceof Integer) &&
		(whereToFinish == APING_REINIT_NN) ) {
		_waiting_pings.put( remoteHost,
				    Clock.current_date(my_node_id) );
	    }
	    QuickStatReq qsr =
		new QuickStatReq( remoteHost, Patch.TYPE_LATENCY,
				  PING_TIMEOUT, whereToFinish, my_sink );
	    // by default, we just want one ping to happen
	    qsr.just_once = true;
	    dispatch( qsr );
	}
    }

    /**
     * Print an id in hex format
     */
    public String print_id(int[] id) {
	String out = "";
	if (id ==null) return out;
	for (int i=0; i<id.length; i++) {
	    out += Integer.toHexString(id[i]);
	}
	return out;
    }

  /**
   * Tell Patchwork to start monitoring all the nodes in the routing table
   */
  private void send_monitor_msgs() {

    //monitoring is on by default    
    if (!MONITORING)   
      return;
    
    final String tag = class_tag + ".send_monitor_msgs";

    // tell patchwork to start monitoring these node
    // we start from the lowest level so that nodes are defined
    // by their highest level in the table, even if it's a backup
    // on a lower level
    for( int i = 0; i < _routing_tbl.digits_per_guid(); i++ ) {	    
      for( int j = 0; j < _routing_tbl.digit_values(); j++ ) {
	RouteEntry re = _routing_tbl.getEntry(i, j);
	if( re == null ) {
	  continue;
	}
	// go through each node, and send a message
	for( int k = 0; k < re.size(); k++ ) {
	  NodeId peer = re.getAtIndex( k );
	  Double latencyd = _routing_tbl.node_to_latency(peer);
	  long latency = latencyd.longValue();
	  if( DEBUG ) {
	      Debug.printtagln(tag, "Sending monitor start:"+peer);
	  }
	  dispatch( new MonitorMsg( peer, i, latency*1000, Patch.TYPE_LATENCY, 
				    new Triple(i,j,peer), my_sink ) );
	}
      }
    }
  }

    /**
     * Should this object be moved to the new node?
     */
    public boolean moveObj(int [] olddigits, int [] newdigits,
			   SecureHash objGuid) {
	final String tag = class_tag + ".moveObj";
	//do we move the object to the new guy
	int i,j;
	int [] objdigits = RoutingTable.digits(objGuid);
	if (DEBUG)
	    Debug.printtagln(tag, "Comparing object:" + 
			     RoutingTable.printGuid(objGuid)+
			     ", oldG="+print_id(olddigits)+
			     ", newG="+print_id(newdigits));

	for(i=0; i<RoutingTable.DIGITS_PER_GUID; i++)
	    if (olddigits[i] != objdigits[i])
		break;
	for(j=0; j<RoutingTable.DIGITS_PER_GUID; j++)
	    if (newdigits[j] != objdigits[j])
		break;
	if (j > i) return true;
	else if (j < i) return false;

	// i==j
	while (olddigits[i] == newdigits[i]) i++;

	j = objdigits[i];
	if (DEBUG)
	    Debug.printtagln(tag, "j="+j+", i="+i);
	while (true) {
	    if (j == olddigits[i]) return false;
	    if (j == newdigits[i]) return true;
	    j = (j == RoutingTable.DIGIT_VALUES - 1) ? 0 : j+1;
	}      

    }

    /**
     * Send DynamicPtrMsg.remove to all nodes that have been removed 
     * from the routing table.
     * This method should be called after every call to RoutingTable.write or
     * RoutingTable.optimize, in order to the revpointers of your neighbors 
     * consistent.
     */
    private void send_remove_pointers() {

	// get the list of recently removed node
	LinkedList removed_nodes = _routing_tbl.get_removed_nodes();

	while( removed_nodes.size() > 0 ) {
	    Pair currPair = (Pair) removed_nodes.remove(0);
	    Integer level = (Integer) currPair.second;
	    sendNonceMsg( new DynamicPtrMsg( (NodeId) currPair.first, 
					     _self_guid, 
					     level.intValue(), true ),
			  SHORT_TIMEOUT, new Pair( USERDATA_PTR,
						   new Integer(0)));
	}

    }

    
    /**
     ** The flag that determines whether the DynamicTClient turns on
     ** node monitoring via Patchwork.  The default is set here to true,
     ** but can be overridden by config argument.
     ** Config name = disable_monitor
     **/
    private boolean MONITORING = true;
    
    /**
     ** The flag that determines whether DynamicTClient listens to node dead
     ** status messages from Patchwork. The default set here to true, but can
     ** be overridden by config argument.  Config name = ignore_status_msgs
     **/
    private boolean LISTEN_STATUS = true;
    
    private long timer;
    private boolean sync;
    private boolean _fulltest;
    private boolean simulator;  //are we running in the simulator
    private soss.network.Network net;  
    //instance of the soss network which we use for getting ping times
    private String allpairspingsfile;

    private HashMap _revpointers;
    private HashMap _objlevels;  
    //mapping from Int-->HashSet of (Obj Guids), Int = -1 for root objects
    private HashMap _leavingobjs;
    //mapping from Guid of leaving node --> DeleteMCState
    private TreeSet backnodes;
    //mapping of backpointer nodes used in deletion

    private int status;   //are we inserting, deleting, what...
  
    private String _dynamic_route;
    private Map _backpointers;  //mapping of objGuids->sets of publishInfo
    private HashMap _movingObjs;    //temp storage for objects in transit
    private BitSet _route_bitmap; //whether a level is filled in or not
    private NodeId[] _gate_node_id;
    private NodeId _surrogate;
    private SecureHash _self_guid;
    private int [] _self_digits;
    private Map _out_times;
    private RoutingTable _routing_tbl;

    //data fields for the new node to be inserted
    private TreeSet initList; 
    //initial list of nearest neighbor candidates, each entry is a GuidAddr
    private int insertinglevel; //route level the pings are getting entries for
    private HashMap _waiting_pings;
    // list of outstanding pings, nodeid-->date sent
    private HashMap _ping_cache;
    // cache of all ping latencies in insertion process, nodeid-->latency
    private HashMap _currentList_pings;
    // ping latencies for all nodes on current closest neighbor list
    // entries can be copied over from _ping_cache
    private TreeMap _nearest_neighbors;
    // sorted Treemap of nodes sorted by latency, used to generate "seed"
    // nodes for the next round
    private HashMap _waiting_notifies;

    // need to keep around lists of DMAck expected nonces
    private HashMap _waiting_dmack_nonces = new HashMap();

    // list of expected nonces for repair requests (to know when req failed)
    private HashMap _waiting_repair_nonces = new HashMap();

    //data fields for other nodes helping the insertion
    private MulticastState _mcast;
    // cached objects storing info on multicasts which have been sent out
    // nodeid-->MulticastState

    // the local obj pointers index
    private static final Integer _local_ptr_index = 
	new Integer(Router.LOCALROOTOBJS);
    private HashMap deadEntries;
    // keep track of the nodes to which seedmsgs are sent,
    // because you could get duplicates
    private TreeSet seedmap_nodes = new TreeSet();
}
