package ostore.tapestry.impl;
import ostore.network.NetworkMessage;
import ostore.util.NodeId;
import ostore.util.SecureHash;
import ostore.util.ByteUtils;
import ostore.util.QSException;
import ostore.util.OutputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;

/**
 * FullTestMsg.java
 *
 * The message that carries a request or response for a test action or
 * sequence. Sent between the FullTestDriver and any number of
 * FullTestMembers.
 *
 * @author Ben Y. Zhao
 * @version $Id: FullTestMsg.java,v 1.19 2003/04/28 21:58:57 hling Exp $
 */
public class FullTestMsg 
  extends NetworkMessage implements QuickSerializable {

  /**
   * Test is finished, exit.
   */
  public static final int EXIT_REQUEST = -1;

  /**
   * Message requesting basic node info like Guid and # of local objs
   * Data fields: none
   */
  public static final int INFO_REQUEST = 0;

  /**
   * Message requesting node "create" some objects with GUID and publish
   * Data fields: SecureHash [] objguid
   */
  public static final int PUBLISH_REQUEST = 1;

  /**
   * Message requesting node "destroy" local objects with GUID and unpublish
   * Data fields: SecureHash [] objguid
   */
  public static final int UNPUBLISH_REQUEST = 2;

  /**
   * Message requesting node send TapestryLocate to GUID
   * Data fields: SecureHash [] objguid
   */
  public static final int LOOKUP_REQUEST = 3;

  /**
   * Message requesting node send TapestryRouteMsg to node GUID
   * Data fields: SecureHash [] nodeguid
   */
  public static final int ROUTE_REQUEST = 4;

  /**
   * Message requesting node to do dynamic node insertion with node GUID
   * Data fields: int millsToWait, SecureHash nodeguid
   */
  public static final int INTEGRATE_REQUEST = 5;

  /**
   * Message requesting node to detach from Tapestry network
   * Data fields: int millsToWait
   */
  public static final int DETACH_REQUEST = 6;

  /**
   * Message requesting node to detach, wait for some time, then reinsert
   * itself with node GUID 
   * Data fields: int millsToWait, SecureHash nodeguid
   */
  public static final int DETACH_INTEGRATE_REQUEST = 7;

  /**
   * Message requesting node to detach, wait for some time, then reinsert
   * itself with node GUID
   * Data fields: none
   */
  public static final int DEBUG_TEST_REQUEST = 8;

  /**
   * Message requesting routing table and revpointers from node
   */
  public static final int VERIFY_REQUEST = 9;

  /**
   * Message requesting bandwidth counts
   */
  public static final int CLEAR_BANDWIDTH_COUNTS = 10;

  /**
   * Message sending trace data to test member
   */
  public static final int TRACE_DATA = 11;

  /**
   * Message telling test member trace experiment start
   */
  public static final int TRACE_START = 12;

  /**
   * Message telling test member to inject fault to itself
   */
  public static final int INJECT_FAULT = 13;

  /**
   * Message telling test member to find out overlay routing
   * path between guids[0] and guids[1];
   */
  public static final int ROUTING_PATH_REQUEST = 14;

  /**
   * Message from test member to test drive:
   * the routing path with some hop.
   * Data field: OverlayPath.
   */
  public static final int ROUTING_PATH_RESULTS = 15;


  public int type;
  public int value;
  public SecureHash requestID;
  public int msToWait;     //optional
  public double failure;  //optional
  public SecureHash [] guids;  //required
  public NodeId [] nodeids;  //optional
  public NodeId gateway; //optional
  public long [] time;   // time in millisecond  
  public boolean repeat_lookups = false;
  public String  comment = " ";
  public int hop=0;
  public int test = 3;

  public FullTestMsg (NodeId peer, int type, SecureHash reqID,
		      int msToWait, int hop) {
    super (peer, false /*outbound*/);
    this.type = type;
    this.requestID = reqID;
    this.msToWait = msToWait;
    this.guids = null;
    this.nodeids = null;
    this.gateway = null;
    this.hop = hop;
  }


  public FullTestMsg (NodeId peer, int type, SecureHash reqID,
		      int msToWait, SecureHash [] guids) {
    super (peer, false /*outbound*/);
    this.type = type;
    this.requestID = reqID;
    this.msToWait = msToWait;
    this.guids = guids;
    this.nodeids = null;
    this.gateway = null;
  }

  public FullTestMsg (NodeId peer, int type, SecureHash reqID,
		      int msToWait, SecureHash [] guids,
		      NodeId [] nodeids, String comment) {
    super (peer, false /*outbound*/);
    this.type = type;
    this.requestID = reqID;
    this.msToWait = msToWait;
    this.guids = guids;
    this.nodeids = nodeids;
    this.gateway = null;
    this.comment = comment;
  }


  public FullTestMsg (NodeId peer, int type, SecureHash reqID,
		      int msToWait, SecureHash [] guids, NodeId [] nodeids) {
    super (peer, false /*outbound*/);
    this.type = type;
    this.requestID = reqID;
    this.msToWait = msToWait;
    this.guids = guids;
    this.nodeids = nodeids;
    this.gateway = null;
  }

  public FullTestMsg (NodeId peer, int type, SecureHash reqID,
		      int msToWait, SecureHash [] guids, NodeId gateway) {
    super (peer, false /*outbound*/);
    this.type = type;
    this.requestID = reqID;
    this.msToWait = msToWait;
    this.guids = guids;
    this.nodeids = null;
    this.gateway = gateway;
  }

  public FullTestMsg (NodeId peer, int type, int val, SecureHash reqID, NodeId gateway,
		      long [] time ) {
    super (peer, false /*outbound*/);
    this.type = type;
    this.value = val;
    this.time = time;
    this.requestID = reqID;
    this.msToWait = 0;
    this.guids = null;
    this.nodeids = null;
    this.gateway = gateway;

  }

  public FullTestMsg (NodeId peer, int type, SecureHash reqID, NodeId gateway,
		      long [] time ) {

    super (peer, false /*outbound*/);
    this.type = type;
    this.time = time;
    this.requestID = reqID;
    this.msToWait = 0;
    this.guids = null;
    this.nodeids = null;
    this.gateway = gateway;

  }

  public FullTestMsg (InputBuffer buffer) throws QSException {
    super(buffer);
    type  = buffer.nextInt();
    value = buffer.nextInt();
    requestID = (SecureHash) buffer.nextObject();
    msToWait = buffer.nextInt();
    failure  = buffer.nextDouble();
    hop      = buffer.nextInt();
    test     = buffer.nextInt();
    boolean  have_guids = buffer.nextBoolean();
    if ( have_guids )
    {
	int size = buffer.nextInt();
	guids = new SecureHash [size];
	for (int i = 0 ; i<size; i++) {
	    guids[i] = (SecureHash) buffer.nextObject();
	}
    }
    boolean  have_nodeids = buffer.nextBoolean();
    if ( have_nodeids )
    {
	int size = buffer.nextInt();
	nodeids = new NodeId [size];
	for (int i = 0 ; i<size; i++) {
	    nodeids[i] = (NodeId) buffer.nextObject();
	}
    }
    boolean have_gateway = buffer.nextBoolean();
    if (have_gateway)
      gateway = (NodeId) buffer.nextObject();
    else gateway = null;

    boolean have_comment = buffer.nextBoolean();
    if (have_comment)
      comment = buffer.nextString();
    else comment = " ";

    //For trace_driven experiment.
    repeat_lookups = buffer.nextBoolean();
    if (type == TRACE_DATA)
    {
	boolean  have_times = buffer.nextBoolean();
	if ( have_times )
	{
	    int size = buffer.nextInt();
	    time = new long [size];
	    for (int i = 0 ; i<size; i++) {
		time[i] = buffer.nextLong();
	    }
	}
    }
  }

  public void serialize (OutputBuffer buffer) {
    super.serialize (buffer);
    buffer.add(type);
    buffer.add(value);
    buffer.add(requestID);
    buffer.add(msToWait);
    buffer.add(failure);
    buffer.add(hop);
    buffer.add(test);
    if (guids == null)
	buffer.add(false);
    else
    {
	buffer.add(true);
	buffer.add(guids.length);
	for (int i=0; i<guids.length; i++) {
	    buffer.add(guids[i]);
	}
    }
    if (nodeids == null)
	buffer.add(false);
    else
    {
	buffer.add(true);
	buffer.add(nodeids.length);
	for (int i=0; i<nodeids.length; i++) {
	    buffer.add(nodeids[i]);
	}
    }
    if (gateway == null)
      buffer.add(false);
    else {
      buffer.add(true);
      buffer.add(gateway);
    }

    if (comment == null)
      buffer.add(false);
    else {
      buffer.add(true);
      buffer.add(comment);
    }


    //For trace_driven experiment.
    buffer.add( repeat_lookups );
    if (type == TRACE_DATA)
    {
	if (time == null)
	    buffer.add(false);
	else
	{
	    buffer.add(true);
	    buffer.add(time.length);
	    for (int i=0; i<time.length; i++) {
		buffer.add(time[i]);
	    }
	}	
    }
  }

  public Object clone () throws CloneNotSupportedException {
    FullTestMsg result = (FullTestMsg) super.clone();
    result.type = type;
    result.value = value;
    result.requestID = requestID;
    result.msToWait = msToWait;
    result.guids = guids;
    result.nodeids = nodeids;
    result.gateway = gateway;
    return result;
  }

  public String toString() {
    String out = "(FullTestMsg: type="+type+", reqID="+
	requestID+", to:"+peer+", guids=";
    if (guids != null)
    {
	for (int i=0; i<guids.length; i++) 
	    out += guids[i]+" ";
    }
    return out;
  }

}





