/**
 * FaultInjectionStage.java
 *
 * @author	Jeremy Stribling
 * @version	$Id: FaultInjectionStage.java,v 1.4 2002/08/02 19:52:42 geels Exp $
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ostore.tapestry.patchwork;

import ostore.util.InvalidConfigFileException;

import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.tapestry.impl.RoutingTable;
import ostore.util.NodeId;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.UnknownEventException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkFullException;   
import seda.sandStorm.api.StagesInitializedSignal; 
import java.util.Random;
import java.util.LinkedList;

/**
 * @author 	Jeremy Stribling
 * @version	$Id: FaultInjectionStage.java,v 1.4 2002/08/02 19:52:42 geels Exp $
 */  

public class FaultInjectionStage implements EventHandlerIF
{
    ////////////////////       CONSTANTS	////////////////////
    private boolean DEBUG = DebugFlags.TAPESTRY_PATCHWORK_FAULTINJECTIONSTAGE;
    private static final String class_tag = "FaultInjectionStage"; 

    private Classifier classifier;

    private static final String FAULT_PERIOD_KEY = "fault_period";
    private static final int FAULT_PERIOD_DEFAULT = 20000;
    private static final String OUTSTANDING_FAULTS_KEY = "outstanding_faults";
    private static final int OUTSTANDING_FAULTS_DEFAULT = 4;

    private static final int MAX_NUM_PROBES_DROPPED = 500;

    private int fault_period;
    private int outstanding_faults;

    // A list of all the faults currently injected 
    // (as <code>PatchworkFaultMsgs</code>)
    // At most <code>outstanding_faults</code> faults in this list. 
    private LinkedList faults;

    // from routing table
    private int digits_per_guid;
    private int digit_values;

    private Random my_rng;

    private SinkIF my_sink;
    private RoutingTable routing_table;

    private NodeId self_node_id;

    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {

	Debug.turnOn( class_tag + ".init" );
	Debug.turnOn( class_tag + ".handleAlarmMsg" );

    }

    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void handleEvents(QueueElementIF elemarr[]) 
	throws EventHandlerException
    {

	final String tag = class_tag+".handleEvents";
	if( DEBUG ) Debug.printtagln( tag, "called: "
				      + elemarr.length + " elements" );
	for( int i=0; i<elemarr.length; i++ ) {
	    handleEvent( elemarr[i] );
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }


    /** 
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void handleEvent(QueueElementIF element) 
	throws EventHandlerException
    {

	final String tag = class_tag+".handleEvent";
	if( DEBUG ) Debug.printtagln( tag, "called" );

        if (DEBUG) Debug.printtagln (tag, "got " + element.getClass () +
				     " " + element + ".");

	if( element instanceof StagesInitializedSignal ) {

	  // set first alarms only if they're to be probed
	  if( fault_period != 0 ) {
	    setAlarm( fault_period );
	  }
 
	} else if( element instanceof FaultAlarmMsg ) {

	    handleAlarmMsg( (FaultAlarmMsg) element );

        } else if( element instanceof RoutingTable.Available ) {
	
	  handleRoutingTableAvailable( (RoutingTable.Available) element );
	
	} else {

	  Carp.error( "Dropping unexpected message: " + element );

	}

    }

    /**
     * It's time to take some measurements!
     */
    private void handleAlarmMsg( FaultAlarmMsg msg ) {

	final String tag = class_tag+".handleAlarmMsg";
	if (DEBUG) Debug.printtagln( tag, " called" );

	// no routing table yet
	if( routing_table == null ) {
	  if( DEBUG ) Debug.printtagln( tag, "No routing table yet" );
	  setAlarm( fault_period );
	  return;
	}

	// take away a fault if the list is full
	if( faults.size() > 0 && faults.size() == outstanding_faults ) {
	    PatchworkFaultMsg oldFault = 
		(PatchworkFaultMsg) faults.removeFirst();
	    if (DEBUG) Debug.printtagln( tag, " removing old fault: " + 
					 oldFault );
	    oldFault.value = 0;
	    dispatch( oldFault );
	}

	// pick a random position in the routing table
	// and send a random fault to it

	NodeId faultee = null;
	while( faultee == null || faultee.equals( self_node_id ) ) {

	    int i = my_rng.nextInt( digits_per_guid );
	    int j = my_rng.nextInt( digit_values );
	    faultee = routing_table.read( i, j );

	}

	// percent or value?
	/*
	boolean isPercent = false;
	if( my_rng.nextDouble() > .5 ) {
	    isPercent = true;
	}

	// figure out value
	int value;
	if( isPercent ) {
	    value = my_rng.nextInt( 100 );
	} else {
	    value = my_rng.nextInt( MAX_NUM_PROBES_DROPPED );
	}

	PatchworkFaultMsg newFault = new PatchworkFaultMsg( faultee, value,
							    isPercent );
	*/
	PatchworkFaultMsg newFault = new PatchworkFaultMsg( faultee, 100,
							    true );
	if (DEBUG) Debug.printtagln( tag, " about to inject fault: " 
				     + newFault );
	dispatch( newFault );
	faults.add( newFault );

	// reset alarm
	setAlarm( fault_period );

    }

    /**
     * Save a reference to the routing table
     */
    private void handleRoutingTableAvailable( RoutingTable.Available msg ) {

	final String tag = class_tag+".handleRoutingTableAvailable";

	if( DEBUG ) { 
	  Debug.printtagln( tag, "called with: \n" + msg.table );
	  if( routing_table != null ) { 
	      Carp.die( "got two versions of the routing table." );
	  }
	}

	routing_table = msg.table;
	faults = new LinkedList();

	digits_per_guid = routing_table.digits_per_guid();
	digit_values = routing_table.digit_values();

    }

    /** 
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void init(ConfigDataIF config) throws Exception
    {
	final String tag = class_tag+".init";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	Random starter_rng = new Random();
	long my_rng_starter = starter_rng.nextLong();
	if( DEBUG ) Debug.printtagln( tag, " the starter random number is " 
				      + my_rng_starter );
	my_rng = new Random( my_rng_starter );

	///////////////////
	// get config info

	fault_period = config.getInt( FAULT_PERIOD_KEY );
	if( fault_period == -1 ) {
	  fault_period = FAULT_PERIOD_DEFAULT;
	} else if( fault_period < 0 ) {
	  throw new InvalidConfigFileException( "The " + FAULT_PERIOD_KEY +
						"tag must be >= 0");
 	}

	outstanding_faults = config.getInt( OUTSTANDING_FAULTS_KEY );
	if( outstanding_faults == -1 ) {
	  outstanding_faults = OUTSTANDING_FAULTS_DEFAULT;
	} else if( outstanding_faults < 0 ) {
	  throw new InvalidConfigFileException( "The " + 
						OUTSTANDING_FAULTS_KEY +
						"tag must be >= 0");
 	}

	self_node_id = new NodeId (config.getString ("node_id"));

	//////////////////
	// initialize filters

	classifier = Classifier.getClassifier( 
			new NodeId( config.getString( "node_id" ) ) );
	my_sink = config.getStage().getSink();
	Filter filter;

	// Subscribe to all my favorite events

	String[] eventNames = {
	    "seda.sandStorm.api.StagesInitializedSignal",	
	    "ostore.tapestry.impl.RoutingTable$Available",
	    "ostore.tapestry.patchwork.FaultInjectionStage$FaultAlarmMsg"
	};

	for( int i = 0; i < eventNames.length; i++ ) {
	  filter = new Filter();
	  Class eventClass;
          if( DEBUG ) Debug.printtagln( tag, "Finding class: " +
                                               eventNames[i] );
          try {
            eventClass = Class.forName( eventNames[i] );
	    if( DEBUG ) Debug.printtagln( tag, "Requiring class: " +
	  				       eventNames[i] );
	    if( !filter.requireType( eventClass ) || 
	        !classifier.subscribe( filter, my_sink ) ) {

	      Carp.die( "Could not require class " + eventNames[i] ); 

	    } 

          } catch ( ClassNotFoundException cnfe ) {
            Carp.die( cnfe );
          }

        }

    }

    /**
     * An internal method used for dispatching events
     */
    private void dispatch( QueueElementIF event ) {
                                                
	final String tag = class_tag+".dispatch";
	if( DEBUG ) Debug.printtagln( tag, "called with event " + event );

      try {
        classifier.dispatch( event );
      } catch ( SinkFullException sfe ) {
        //TODO: Better way to deal with SinkFull
        try {
          Thread.currentThread().sleep( 5000 );
        } catch( InterruptedException ie ) {
          //ignore
        }
      }                                         
      
    }

    /**
     * Sets this stage's alarm clock for <code>millis</code> milliseconds
     * from now.
     */	
    private void setAlarm( long millis ) {

	final String tag = class_tag+".setAlarm";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	classifier.dispatch_later( new FaultAlarmMsg(), millis );

    }

    /** 
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void destroy()
    {
        final String tag = class_tag+".destroy";
        if( DEBUG ) Debug.printtagln( tag, "destroyed" );
    }

    ////////////////////     INNER CLASSES	////////////////////

    private class FaultAlarmMsg implements QueueElementIF {

    }

}
