/**
 * PatchworkFaultMsg.java
 *
 * @author Jeremy Stribling
 * @version    $Id: PatchworkFaultMsg.java,v 1.6 2003/01/29 22:40:50 strib Exp $
 */

package ostore.tapestry.patchwork;
import ostore.network.NetworkMessage;
import ostore.util.ByteUtils;
import ostore.util.NodeId;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.QSException;

/**
 * This is a message sent out by the <code>FaultInjectionStage</code>
 * to inform a peer to start dropping probe messages.
 *
 * @author Jeremy Stribling
 * @version    $Id: PatchworkFaultMsg.java,v 1.6 2003/01/29 22:40:50 strib Exp $
 */
public class PatchworkFaultMsg extends NetworkMessage 
    implements QuickSerializable {

    public boolean isPercent;
    public boolean allNodes;
    public int value;

    public NodeId nodeToDrop;

    /**
     * Constucts a fault injection message, which will let a peer
     * know that it should begin dropping messages.  Either an
     * absolute number of messages can be specified, or a percentage 
     * of messages for an indefinite amount of time.  Note that if
     * a percentage is given, it can be removed by sending another
     * fault injection message with the value at 0.  Any arriving
     * fault message overrides the previous one.
     *
     * @param peer  the node to contact
     * @param value the amount of messages to drop, either as a 
     *              number of messages or a percent of messages
     * @param isPercent  true if <code>value</code> specifies a percentage
     */
    public PatchworkFaultMsg (NodeId peer, int value, boolean isPercent) {
	this( peer, value, isPercent, false );
    }

    /**
     * Constucts a fault injection message, which will let a peer
     * know that it should begin dropping messages.  Either an
     * absolute number of messages can be specified, or a percentage 
     * of messages for an indefinite amount of time.  Note that if
     * a percentage is given, it can be removed by sending another
     * fault injection message with the value at 0.  Any arriving
     * fault message overrides the previous one.
     *
     * @param peer  the node to contact
     * @param value the amount of messages to drop, either as a 
     *              number of messages or a percent of messages
     * @param isPercent  true if <code>value</code> specifies a percentage
     * @param boolean   true iff this fault is to apply to all nodes probing it
     */
    public PatchworkFaultMsg (NodeId peer, int value, boolean isPercent, 
			      boolean allNodes) {

	super (peer, false /* outbound */);

	this.value = value;
	this.isPercent = isPercent;
	this.allNodes = allNodes;

    }

    /**
     * Constucts a fault injection message, which will let a peer
     * know that it should begin dropping messages.  Either an
     * absolute number of messages can be specified, or a percentage 
     * of messages for an indefinite amount of time.  Note that if
     * a percentage is given, it can be removed by sending another
     * fault injection message with the value at 0.  Any arriving
     * fault message overrides the previous one.
     *
     * @param peer  the node to contact
     * @param value the amount of messages to drop, either as a 
     *              number of messages or a percent of messages
     * @param isPercent  true if <code>value</code> specifies a percentage
     * @param nodeToDrop Drop messages from a specific node, instead of sender 
     */
    public PatchworkFaultMsg (NodeId peer, int value, boolean isPercent, 
			      NodeId nodeToDrop) {

	super (peer, false /* outbound */);

	this.value = value;
	this.isPercent = isPercent;
	allNodes = false;
	this.nodeToDrop = nodeToDrop;

    }

    public PatchworkFaultMsg (InputBuffer buffer) throws QSException {
	super( buffer );
	isPercent = buffer.nextBoolean();
	value = buffer.nextInt();
	allNodes = buffer.nextBoolean();
	if( buffer.nextBoolean() ) {
	    nodeToDrop = (NodeId) buffer.nextObject();
	}
    }

    public void serialize (OutputBuffer buffer) {
	super.serialize (buffer);
	buffer.add( isPercent );
	buffer.add( value );
	buffer.add( allNodes );
	if( nodeToDrop == null ) {
	    buffer.add( false );
	} else {
	    buffer.add( true );
	    buffer.add( nodeToDrop );
	}
    }

    public Object clone () throws CloneNotSupportedException {
	PatchworkFaultMsg newone = (PatchworkFaultMsg) super.clone ();
	newone.isPercent = isPercent;
	newone.value = value;
	newone.allNodes = allNodes;
	return newone;
    }

    public String toString () {
	return "(PatchworkFaultMsg, value: " + 
		value + ", isPercent: " + isPercent + ", allNodes:"
		+ allNodes + " " + super.toString () + ")";
    }
}

