/**
 * PatchworkModel.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.tapestry.patchwork;

import ostore.util.QuickSerializable;
import ostore.util.QuickSerializable;
import ostore.util.QSException;
import ostore.util.InputBuffer;
import ostore.util.QSVector;
import ostore.util.QSWindow;
import ostore.util.TypeTable;
import ostore.util.Types;
import ostore.util.ByteUtils;

import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.SecureHash;
import ostore.util.QSInt;
import ostore.util.OutputBuffer;
import ostore.oil.Matrix;
import ostore.oil.Array;
import ostore.oil.Model;
import java.util.HashMap;
import java.util.Iterator;

/**
 * <p>
 * The <code>PatchworkModel</code> is built by the <code>PatchworkStage</code>
 * code, and is used for collecting network statistics.  It implements
 * the <code>ostore.oil.Model</code> interface, although not all methods
 * are currently implemented.  It remembers observations sent to it
 * (only up to <code>MAX_OBSERVATIONS</code> of them, which is currently 
 * set at 1000), and when <code>recalculate</code> is called it produces a 
 * <code>PatchworkModel.Delta</code> that contains the difference between
 * the current model and the last one produced from a call to
 * <code>recalculate</code>. </p><p>
 *
 * The values contained in the model represent averages of the statistics
 * over the number of observations that existed for that node when 
 * <code>recalculate</code> was called.  It only considers a change worthy
 * of noting if it is greater than <code>VALID_DELTA_CHANGE</code> 
 * percent, currently set at 5%.</p><p>
 *
 * Currently, the integer values and their corresponding types are as
 * follows:</p><p>
 *
 * <b>latency</b>: <code>PatchworkModel.LATENCY</code></p><p>
 *
 * <b>bandwidth</b>: <code>PatchworkModel.BANDWIDTH</code></p><p>
 *
 * <b>loss</b>: <code>PatchworkModel.LOSS</code></p><p>
 *
 * For a more complete description of these statistics, see
 * <code>PatchworkStage</code>.</p>
 *
 * @see ostore.oil.Model
 *
 * @author	Jeremy Stribling
 * @version	$Id: PatchworkModel.java,v 1.11 2003/02/05 22:49:38 hling Exp $
 */

public class PatchworkModel implements Model
{
    ////////////////////       CONSTANTS	////////////////////

    private final static boolean DEBUG = 
	ostore.util.DebugFlags.TAPESTRY_PATCHWORK_PATCHWORKMODEL;

    private final static String class_tag = "PatchworkModel";

    // Types of possible measurements
    public final static int LATENCY = 0;
    public final static int BANDWIDTH = 1;
    public final static int LOSS = 2;

    private static QSInt QS_LATENCY;
    private static QSInt QS_BANDWIDTH;
    private static QSInt QS_LOSS;

    // the maximum number of observations to store at one time
    private final static int MAX_OBSERVATIONS = 1000;

    // the least percentage change a number must be from the current model
    // to count as a delta worthy of saving.
    private final static int VALID_DELTA_CHANGE = 1;

    // how much to emphasize the newest average in the calculations
    private final static double ALPHA = .8;

    ////////////////////       VARIABLES	////////////////////

    // a window of the most recent MAX_OBSERVATIONS observations
    private QSWindow window;
    
    // a measurement matrix
    private Matrix measureMatrix;

    // a vector of nodes to be removed at the next recalculation
    private QSVector toBeRemoved;

    ////////////////////      CONSTRUCTORS	////////////////////

    static {

	Debug.turnOn( class_tag + ".record" );
	Debug.turnOn( class_tag + ".clear" );

	QS_LATENCY = new QSInt( LATENCY );
	QS_BANDWIDTH = new QSInt( BANDWIDTH );
	QS_LOSS = new QSInt( LOSS );

    }
    
    /**
     * Create a new <code>PatchworkModel</code>.
     */
    public PatchworkModel() {

	window = new QSWindow( MAX_OBSERVATIONS );
	measureMatrix = new Matrix();
	toBeRemoved = new QSVector();; 

    }

    public PatchworkModel( InputBuffer buffer ) throws QSException {
	window = (QSWindow) buffer.nextObject();
	measureMatrix = (Matrix) buffer.nextObject();
	toBeRemoved = (QSVector) buffer.nextObject();
    }

    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Specified by ostore.oil.Model interface 
     */
    public void record( QuickSerializable[] observations )
    {
	for( int i=0; i<observations.length; i++ ) {
	    record( observations[i] );
	}
    }

    /**
     * Specified by ostore.oil.Model interface 
     */
    public void record( QuickSerializable observation )
    {

        final String tag = class_tag+".record";
        if( DEBUG ) Debug.printtagln( tag, "called with observation "
                                      + observation );

	window.add( observation );
    }

    /**
     * Specified by ostore.oil.Model interface 
     */
    public synchronized void clear( int num ) 
    {

        final String tag = class_tag+".clear";
        if( DEBUG ) Debug.printtagln( tag, "called with num = " + num );

	window.clear( num );

    }

    /**
     * Specified by ostore.oil.Model interface 
     */
    public synchronized void clear() 
    {
        final String tag = class_tag+".clear";
        if( DEBUG ) Debug.printtagln( tag, "all" );

	window.clear();

    }

    /** 
     * Returns the most recently calculated double value of type 
     * <code>type</code> for the specified node id <code>peer</code>.  
     * Returns Double.NaN if a value is not recorded.
     *
     * @param peer   the node for which to lookup the statistic
     * @param type   the type of statistic to collect (i.e. 
     *               LATENCY, BANDWIDTH, etc)
     *
     * @return the statistic if it exists, Double.NaN otherwise
     * @throws InvalidTypeException if an invalid type is passed in
     */
    public double lookupDouble( SecureHash peer, int type ) {

	Array nodeArray = measureMatrix.get( peer );

	if( nodeArray == null ) {
	    return Double.NaN;
	}

	QSInt qsType = getQSType( type );

	if( qsType == null ) {
	    throw new InvalidTypeException( type + " is not a valid type." );
	} else {
	    return nodeArray.get( qsType );
	}

    }

    /**
     * Tells the model to remove all statistics for this peer at the
     * time of the next recalculation.  This is best called when it is known
     * for certain that a node wil not be observed ever again, or at least
     * for a long time.
     *
     * @param peer      the node for which to remove the statistics
     */
    public void remove( SecureHash peer ) {

	// make sure it's not already in it
	if( !toBeRemoved.contains( peer ) ) {
	    toBeRemoved.add( peer );
	}

    }

    /** 
     * Not yet implemented.  Specified by ostore.oil.Model interface 
     */
    public synchronized QuickSerializable[] outliers()
    {
	throw new RuntimeException( "Not implemented yet, sorry." );
    }
    

    /**
     * Specified by ostore.oil.Model interface 
     */
    public synchronized void add_delta( Model.Delta d )
    {

        final String tag = class_tag+".add_delta";
        if( DEBUG ) Debug.printtagln( tag, "called" );

	if( !( d instanceof Delta) ) {
	    throw new InvalidTypeException( "Delta is not of a valid type.  " 
					    + "Must be of type " 
					    + "PatchworkModel.Delta" );
	}

	Delta delta = (Delta) d;

	measureMatrix.add( delta.getMatrix() );

	// removed those who need it
	QSVector removeThem = delta.getToBeRemoved();
	for( int i = 0; i < removeThem.size(); i++ ) {
	    measureMatrix.remove( removeThem.get( i ) );
	}

    }

    /**
     * Not yet implemented.  Specified by ostore.oil.Model interface 
     */
    public synchronized double loglikelihood()
    {
	throw new RuntimeException( "Not implemented yet, sorry." );
    }
 
    /**
     * Specified by ostore.oil.Model interface 
     */
    public synchronized Model.Delta recalculate()
    {

        final String tag = class_tag+".recalculate";
        if( DEBUG ) Debug.printtagln( tag, "called" );

	HashMap nodeValues = new HashMap();
	Delta newDelta = new Delta( toBeRemoved );
	toBeRemoved = new QSVector();

	QuickSerializable[] obs = window.toArray();

	// collect observations for each node
	for( int i = 0; i < obs.length; i++ ) {

	    Observation currOb = (Observation) obs[i];

	    Averagizer currAv = (Averagizer) nodeValues.get( currOb.peer );
	    
	    if( currAv == null ) {
		currAv = new Averagizer();
		nodeValues.put( currOb.peer, currAv );
	    }

	    currAv.collectDouble( currOb.value, currOb.type );

	}

	clear();

	// for each node, figure out a DeltaNode
	Iterator iter = nodeValues.keySet().iterator();
	while( iter.hasNext() ) {

	    Array newArray = new Array();

	    boolean hasLatency = false;
	    boolean hasBandwidth = false;
	    boolean hasLoss = false;

	    SecureHash nodeId = (SecureHash) iter.next();
	    Averagizer currAv = (Averagizer) nodeValues.get( nodeId );
	    Array currArray = measureMatrix.get( nodeId );

	    if( currArray == null ) {
		currArray = new Array( );
		measureMatrix.put( nodeId, currArray );
	    }

	    // determine latency:
	    double delta;
	    double currValue = currArray.get( QS_LATENCY );
	    double latency = currAv.calculateAverage( LATENCY );
	    if( Double.isNaN( currValue ) ) {
		delta = latency;
	    } else {
	        delta = ((1-ALPHA)*currValue + ALPHA*latency) - currValue;
	    }
	    // consider delta valid if modelNode value was zero
	    // or if the percentage change was great enough.
	    if( (currValue == 0) ||
		( ( ( Double.isNaN( currValue ) ) ||
		    ( Math.abs(delta)*100/currValue 
		      >= VALID_DELTA_CHANGE ) ) &&  
		  !Double.isNaN( latency ) ) ) {

		hasLatency = true;
		newArray.put( QS_LATENCY, delta );

	    }

	    // determine bandwidth:
	    currValue = currArray.get( QS_BANDWIDTH );
	    double bandwidth = currAv.calculateAverage( BANDWIDTH );
	    if( Double.isNaN( currValue ) ) {
		delta = bandwidth;
	    } else {
	        delta = ((1-ALPHA)*currValue + ALPHA*bandwidth) - currValue;
	    }
	    // consider delta valid if modelNode value was zero
	    // or if the percentage change was great enough.
	    if( (currValue == 0) ||
		( ( ( Double.isNaN( currValue ) ) ||
		    ( Math.abs(delta)*100/currValue 
		      >= VALID_DELTA_CHANGE ) ) &&  
		  !Double.isNaN( bandwidth ) ) ) {

		hasBandwidth = true;
		newArray.put( QS_BANDWIDTH, delta );

	    }

	    // determine loss:
	    currValue = currArray.get( QS_LOSS );
	    double loss = currAv.calculateAverage( LOSS );
	    if( Double.isNaN( currValue ) ) {
		delta = loss;
	    } else {
	        delta = ((1-ALPHA)*currValue + ALPHA*loss) - currValue;
	    }
	    // consider delta valid if modelNode value was zero
	    // or if the percentage change was great enough.
	    if( (currValue == 0) ||
		( ( ( Double.isNaN( currValue ) ) ||
		    ( Math.abs(delta)*100/currValue 
		      >= VALID_DELTA_CHANGE ) ) &&  
		  !Double.isNaN( loss ) ) ) {

		hasLoss = true;
		newArray.put( QS_LOSS, delta );

	    }

	    // register this dNode:
	    if( hasLatency || hasBandwidth || hasLoss ) {
	      newDelta.storeDeltas( nodeId, newArray );
	    }

	}
	
	add_delta( newDelta );

	return newDelta;

    }

    /**
     * Not yet implemented.  Specified by ostore.oil.Model interface 
     */
    public synchronized Model.Prediction predict( int horizon )
    {
	throw new RuntimeException( "Not implemented yet, sorry." );
    }    

    /**
     * Specified by ostore.util.QuickSerializable 
     */
    public synchronized void serialize (OutputBuffer buffer) 
    {
	buffer.add(window);
	buffer.add(measureMatrix);
	buffer.add(toBeRemoved);
    }

    /**
     * Return the QSInt corresponding to the type.  Return null if type
     * is unrecognized.
     */
    private QSInt getQSType( int type ) {
	if( type == LATENCY ) {
	  return QS_LATENCY;
	} else if( type == BANDWIDTH ) {
	  return QS_BANDWIDTH;
	} else if( type == LOSS ) {
	  return QS_LOSS;
	} else {
	  return null;
	}
    }

    /**
     * Returns a string representation of the object 
     */
    public String toString() {

	String s = "PatchworkModel:\t\tLATENCY\t\tBANDWIDTH\tLOSS\n";
	
	Iterator iter = measureMatrix.ids();
	while( iter.hasNext() ) {
	
	    SecureHash peer = (SecureHash) iter.next();
	    Array nodeArray = measureMatrix.get( peer );

	    double value;

	    if( nodeArray == null ) {
		continue;
	    }

	    s += "\t" + peer + "\t";
	    if( Double.isNaN( (value = nodeArray.get( QS_LATENCY ) ) ) ) {
		s += "---\t\t";
	    } else {
		s += Math.floor(value*100.0)/100.0 + "\t\t";
	    }

	    if( Double.isNaN( (value = nodeArray.get( QS_BANDWIDTH ) ) ) ) {
		s += "---\t\t";
	    } else {
		s += Math.floor(value*100.0)/100.0 + "\t\t";
	    }

	    if( Double.isNaN( (value = nodeArray.get( QS_LOSS ) ) ) ) {
		s += "---\n";
	    } else {
		s += Math.floor(value*100.0)/100.0 + "\n";
	    }	

	}

	return s;

    }

    ////////////////////////////////////////////////////////////////
    ////////////////////     INNER CLASSES	////////////////////

    /**
     * A <code>PatchworkModel.Delta</code> object contains a diff 
     * of the sufficient statistics for a <code>PatchworkModel</code>.
     *
     * @author	Jeremy Stribling
     * @version	$Id: PatchworkModel.java,v 1.11 2003/02/05 22:49:38 hling Exp $
     */

    public static class Delta implements Model.Delta
    {
	////////////////////     CONSTANTS	////////////////////
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////
	
	private Matrix deltaMatrix;
	private QSVector toBeRemoved;
	
	////////////////////    CONSTRUCTORS	////////////////////

	/**
	 * Construct a <code>Delta</code> from its
	 * <code>QuickSerializable</code> form.
	 */
	public Delta( InputBuffer buffer ) throws QSException {
	    deltaMatrix = (Matrix)buffer.nextObject();
	    toBeRemoved = (QSVector)buffer.nextObject();
	}

	/**
 	 * Construct a new <code>Delta</code>. 
	 */
	public Delta() {
	    this( (QSVector) null );
	}

	/** 
	 * Construct a new <code>Delta</code>.
	 *
	 * @param toBeRemoved  a <code>QSVector</code> of <code>SecureHash</code>s
	 *                     that should be removed from the 
	 *                     <code>Model</code>.
	 */
	public Delta( QSVector toBeRemoved ) {
	    deltaMatrix = new Matrix();
	    this.toBeRemoved = toBeRemoved;
	}

	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////
	
	/**
	 * Store an <code>Array</code> representing the deltas
	 * for the node <code>id</code>.  The deltas are doubles
	 * indexed by <code>QSInt</code> types in <code>PatchworkModel</code>.
	 * Overwrites any existing deltas for that id.
	 *
	 * @param id   the node to record deltas for
	 * @param deltas   the deltas themselves
	 */
	public void storeDeltas( SecureHash id, Array deltas ) {
	    deltaMatrix.put( id, deltas );
	}

	/** 
	 * Returns the most recently probed value from delta update with type
	 * <code>type</code> for the specified node id <code>peer</code>.  
	 * Returns Double.NaN if a value is not recorded.
	 *
	 * @param peer   the node for which to lookup the statistic
	 * @param type   the type of statistic to collect (i.e. 
	 *               LATENCY, BANDWIDTH, etc)
	 *
	 * @return the statistic if it exists, Double.NaN otherwise
	 * @throws InvalidTypeException if an invalid type is passed in
	 */
	public double lookupDouble( SecureHash peer, int type ) 
	{
	    /*
	    Array nodeArray = deltaMatrix.get( peer );
	    if( nodeArray == null ) {
		return Double.NaN;
	    }
	    
	    QSInt qsType = getQSType( type );
	    if( qsType == null ) {
		throw new InvalidTypeException( type + " is not a valid type." );
	    } else {
		return nodeArray.get( qsType );
	    }
	    */

	    return 1.0;
	}

	/**
 	 * Returns a <code>Matrix</code> containing all the delta information.
	 * <code>Arrays</code> of deltas are indexed by <code>SecureHash</code>
	 * while deltas within those <code>Arrays</code> are indexed by
	 * <code>QSInts</code>.
	 *
	 * @return the <code>Matrix</code> of deltas
	 */
	public Matrix getMatrix() {
	  return deltaMatrix;
	}

	/**
	 * Get the list of <code>SecureHash</code>s whose statistics should be 
	 * removed from the <code>Model>/code>
	 */
	public QSVector getToBeRemoved() {
	    return toBeRemoved;
	}

	/**
	 * Specified by ostore.oil.Model.Delta interface 
	 */
	public Model.Delta sum( Model.Delta other )
	{

	    if( !( other instanceof Delta) ) {
	        throw new InvalidTypeException( "Delta is not of a valid" + 
						" type.  " + "Must be of type " 
					        + "PatchworkModel.Delta" );
	    }

	    Delta otherCast = (Delta) other;
	    Delta newDelta = new Delta();

	    Matrix newMatrix = newDelta.getMatrix();
	    newMatrix.add( deltaMatrix );
	    newMatrix.add( otherCast.getMatrix() );

	    return newDelta;
	}

	/**
	 * Specified by ostore.util.QuickSerializable 
	 */
	public synchronized void serialize (OutputBuffer buffer)
	{
	    buffer.add (deltaMatrix);
	    buffer.add (toBeRemoved);
	}
	

	/**
	 * Clone this delta 
	 */
	public Object clone() throws CloneNotSupportedException {

	    Delta newDelta = new Delta();

	    newDelta.deltaMatrix = (Matrix) deltaMatrix.clone();

	    return newDelta;

	}

	/**
	 * Returns a String representation of this <code>Delta</code> 
	 */
	public String toString() {

	    String s = "Delta:\t\t\t\tLATENCY\t\tBANDWIDTH\tLOSS\n";

	    Iterator iter = deltaMatrix.ids();
	    while( iter.hasNext() ) {
	
	        SecureHash peer = (SecureHash) iter.next();
	        Array nodeArray = deltaMatrix.get( peer );

	        double value;

	        s += "\t" + peer + "\t";
	        if( Double.isNaN( (value = nodeArray.get( QS_LATENCY ) ) ) ) {
		    s += "---\t\t";
	        } else {
		    s += Math.floor(value*100.0)/100.0 + "\t\t";
	        }

	        if( Double.isNaN( (value = nodeArray.get( QS_BANDWIDTH ) ) ) ) {
		    s += "---\t\t";
	        } else {
		    s += Math.floor(value*100.0)/100.0 + "\t\t";
	        }

	        if( Double.isNaN( (value = nodeArray.get( QS_LOSS ) ) ) ) {
		    s += "---\n";
	        } else {
		    s += Math.floor(value*100.0)/100.0 + "\n";
	        }	

	    }

	    return s;

	}

    }	// end class Delta
    

    /**
     * A <code>Patchwork.Observation</code> object contains an
     * observation recorded by the patchwork stage.
     *
     * @author	Jeremy Stribling
     * @version	$Id: PatchworkModel.java,v 1.11 2003/02/05 22:49:38 hling Exp $
     */

    public static class Observation implements QuickSerializable
    {
	////////////////////     CONSTANTS	////////////////////
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////
	
	private SecureHash peer;
	private double value;
	private int type;
	
	////////////////////    CONSTRUCTORS	////////////////////

	/**
	 * Construct a <code>Patchwork.Observation</code> from its
	 * <code>QuickSerializable</code> form.
	 */
	
	public Observation( InputBuffer buffer ) throws QSException {
	    peer = (SecureHash) buffer.nextObject();
	    value = buffer.nextDouble();
	    type = buffer.nextInt();
	}

	/**
	 * Record the observed double value of type <code>type</code>
	 * for the specified node id <code>peer</code>.
 	 * @param observed_peer  the node to which the statistic belongs
  	 * @param value   the value of the observed statistic
  	 * @param type    the type of statistic (i.e. LATENCY, etc)
	 */
	public Observation( SecureHash observed_peer, double value, int type ) {

	    if( (type != LATENCY) && (type != BANDWIDTH) &&
		(type != LOSS ) ) {
		throw new InvalidTypeException( type + 
						" is not a valid type." );
	    }

	    this.peer = observed_peer;
	    this.value = value;
	    this.type = type;

	}

	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////
	

	/**
	 * Specified by ostore.util.QuickSerializable 
	 */
	public synchronized void serialize (OutputBuffer buffer)
	{
	    buffer.add (peer);
	    buffer.add (value);
	    buffer.add (type);
	}

	/** 
	 * Returns a string representation of this 
         * <code>Observation</code> 
	 */
	public String toString() {

	    String s = "";
	    s += "{peer: " + peer + ", ";
	    s += "value: " + value + ", ";
	    s += "type: " + type + "}";

	    return s;

	}

    }	// end class Observation


    /**
     * A <code>Patchwork.InvalidTypeException</code> notes that
     * an invalid type has been given to the PatchworkModel system
     *
     * @author	Jeremy Stribling
     * @version	$Id: PatchworkModel.java,v 1.11 2003/02/05 22:49:38 hling Exp $
     */

    public static class InvalidTypeException extends RuntimeException
    {
	////////////////////     CONSTANTS	////////////////////
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////
	
	////////////////////    CONSTRUCTORS	////////////////////

	public InvalidTypeException( String s ) {
	    super( s );
	}

	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////
	
    }	// end class InvalidTypeException

    /**
     * A <code>Patchwork.Averagizer</code> collects statistics and
     * counts the number of statistics collected.
     * @author	Jeremy Stribling
     * @version	$Id: PatchworkModel.java,v 1.11 2003/02/05 22:49:38 hling Exp $
     */

    private static class Averagizer
    {
	////////////////////     CONSTANTS	////////////////////
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////
	
	private double currLatencySum;
	public int currLatencyCount;

	private double currBandwidthSum;
	public int currBandwidthCount;
	
	private double currLossSum;
	public int currLossCount;

	////////////////////    CONSTRUCTORS	////////////////////

	public Averagizer( ) {
	    currLatencySum = 0;
	    currLatencyCount = 0;
	    currBandwidthSum = 0;
	    currBandwidthCount = 0;
	    currLossSum = 0;
	    currLossCount = 0;
	}

	////////////////////   STATIC METHODS 	////////////////////

	/**
	 * Collect an integer observation.
	 */
	public void collectDouble( double value, int type ) {

	    if( type == LATENCY ) {
		currLatencySum += value;
		currLatencyCount++;
	    } else if( type == BANDWIDTH ) {
		currBandwidthSum += value;
		currBandwidthCount++;
	    } else if( type == LOSS ) {
		currLossSum += value;
		currLossCount++;
	    } else {
		throw new InvalidTypeException( type + 
						" is not a valid type." );
	    }	    

	}
	
	/**
	 * Calculate the average so far for a certain type.
	 */
	public double calculateAverage( int type ) {

	    double currCount;
	    double  currSum;

	    if( type == LATENCY ) {
		currCount = currLatencyCount;
		currSum = currLatencySum;
	    } else if( type == BANDWIDTH ) {
		currCount = currBandwidthCount;
		currSum = currBandwidthSum;
	    } else if( type == LOSS ) {
		currCount = currLossCount;
		currSum = currLossSum;
	    } else {
		throw new InvalidTypeException( type + 
						" is not a valid type." );
	    }	

	    if( currCount == 0 ) {
		return Double.NaN;
	    } else {
		return currSum / currCount;
	    }

	}

	//////////////////// 	  METHODS	////////////////////
	
    }	// end class Averagizer

}
