#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <linux/ioctl.h>
#include <linux/if.h>
#include <linux/ip.h>
#include <linux/if_tun.h>

#include "ipd.h"

// Hard-coded here to avoid an extra JNI lookup every packet
#define BUFSIZE 4096
#define MAX_IPD_ERRNO 5

enum IPD_ERROR {
  IPD_ERROR_NONE           = 0,
  IPD_ERROR_DEV_OPEN       = -1,
  IPD_ERROR_DEV_CONFIG     = -2,
  IPD_ERROR_PACKET_TOO_BIG = -3,
  IPD_ERROR_WRITE_PACKET   = -4,
};

static const char *IPD_ERROR_STR[MAX_IPD_ERRNO] =
  { 
    "No error",
    "Could not open device /dev/net/tun",
    "Could not configure tun interface" ,
    "Packet too big for buffer"         ,
    "Could not write to tun interface"  ,
  };

JNIEXPORT jstring JNICALL 
  Java_iptap_IPDaemon_ipd_1get_1errstr(JNIEnv *env, jobject obj, 
				       jint errno)
{
  if (errno > 0 || errno < -MAX_IPD_ERRNO) {
    return (*env)->NewStringUTF(env, "Unknown errno");
  }
  return (*env)->NewStringUTF(env, IPD_ERROR_STR[-errno]);
}

JNIEXPORT jint JNICALL 
  Java_iptap_IPDaemon_ipd_1init(JNIEnv *env, jobject obj)
{
  struct ifreq ifr;      // tun driver interface request
  int fd;                // tun interface file descriptor

  // try to open a new tun device
  if ((fd = open("/dev/net/tun", O_RDWR)) < 0) {
    return IPD_ERROR_DEV_OPEN;
  }

  // setup tun interface configuration string
  memset(&ifr, 0, sizeof(ifr));
  ifr.ifr_flags = IFF_TUN | IFF_NO_PI;
  strncpy(ifr.ifr_name, "tun%d", IFNAMSIZ);

  // configure the opened tun interface
  if (ioctl(fd, TUNSETIFF, &ifr) < 0) {
    close(fd);
    return IPD_ERROR_DEV_CONFIG;
  }

  printf("Allocated device %s\n", ifr.ifr_name);

  // all is well, return the file descriptor
  return fd;
}

JNIEXPORT jint JNICALL 
  Java_iptap_IPDaemon_ipd_1get_1packet(JNIEnv *env, jobject obj, 
				       jobject packet_obj, jint fd)
{
  int bytesRead = 0;
  struct iphdr *header;

  int tot_len;
  char src_addr_str[16];
  char dst_addr_str[16];

  jclass packet_class;
  jfieldID packet_fid;  
  jobject array;
  jbyte *buf;

  // Here the JNI mess begins, get the class indentifier
  packet_class = (*env)->GetObjectClass(env, packet_obj);

  // Obtain the byte array buffer from the Java packet object
  packet_fid = (*env)->GetFieldID(env, packet_class, "payload", "[B");

  array = (*env)->GetObjectField(env, packet_obj, packet_fid);  

  buf = (*env)->GetByteArrayElements(env, array, 0);

  // Here the real IP packet begins, read the header
  while (bytesRead < 20) {
    bytesRead += read(fd, buf + bytesRead, 20 - bytesRead);
  }

  // Map the kernel IP header struct to our buffer
  header = (struct iphdr *) buf;

  // Check to see if our buffer is big enough for the whole packet  
  tot_len = ntohs(header->tot_len);

  if (tot_len > BUFSIZE) {
    return IPD_ERROR_PACKET_TOO_BIG;
  }

  // Set the total length field
  packet_fid = (*env)->GetFieldID(env, packet_class, "total_len", "S");
  (*env)->SetShortField(env, packet_obj, packet_fid, 
			(jshort) tot_len);

  // Set the source address field  
  packet_fid = (*env)->GetFieldID(env, packet_class, "saddr", "J");
  (*env)->SetLongField(env, packet_obj, packet_fid, 
		       (jint) ntohs(header->saddr));

  // Set the destination address field  
  packet_fid = (*env)->GetFieldID(env, packet_class, "daddr", "J");
  (*env)->SetLongField(env, packet_obj, packet_fid, 
		       (jint) ntohs(header->daddr));

  // Set the source address string field  
  packet_fid = (*env)->GetFieldID(env, packet_class, "saddr_str", 
				  "Ljava/lang/String;");
  inet_ntop(AF_INET, &header->saddr, src_addr_str, sizeof(src_addr_str));
  (*env)->SetObjectField(env, packet_obj, packet_fid, 
			 (*env)->NewStringUTF(env, src_addr_str));

  // Set the destination address string field  
  packet_fid = (*env)->GetFieldID(env, packet_class, "daddr_str", 
				  "Ljava/lang/String;");
  inet_ntop(AF_INET, &header->daddr, dst_addr_str, sizeof(dst_addr_str));
  (*env)->SetObjectField(env, packet_obj, packet_fid, 
			 (*env)->NewStringUTF(env, dst_addr_str));

  // Read the payload data into the buffer
  while (bytesRead < tot_len) {
    bytesRead += read(fd, buf + bytesRead, tot_len - bytesRead);
  }

  // Release the Java byte array
  (*env)->ReleaseByteArrayElements(env, array, buf, 0);

  // All is well, return
  return IPD_ERROR_NONE;
}

JNIEXPORT jint JNICALL 
  Java_iptap_IPDaemon_ipd_1put_1packet(JNIEnv *env, jobject obj, 
				       jobject array, 
				       jshort array_len, jint fd)
{
  jbyte *buf = (*env)->GetByteArrayElements(env, array, 0);

  if (write(fd, buf, array_len) != 0) {
    return IPD_ERROR_WRITE_PACKET;
  }

  (*env)->ReleaseByteArrayElements(env, array, buf, 0);
  return IPD_ERROR_NONE;
}
