/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

/**
***  The <code>ByteArrayInputBuffer</code> is a utility class that
***  provides an <code>InputBuffer</code> interface to a byte array.
***  It can also provide provide the interface over only a portion of
***  the array.
***
***  @author  Patrick R. Eaton
***  @author  Jeff Pang
***  @version $Id: ByteArrayInputBuffer.java,v 1.5 2002/12/07 01:47:12 geels Exp $
 */
public class ByteArrayInputBuffer
    extends ostore.util.InputBufferImpl
{
    private byte[] _data;
    private int    _first_byte;
    private int    _last_byte;
    private int    _current_byte;

    public ByteArrayInputBuffer()
    {
        this((byte[]) null, 0, 0);
    }

    public ByteArrayInputBuffer(byte[] data) 
    {
        this(data, 0, data.length);
    }

    public ByteArrayInputBuffer(byte[] data, int offset, int length)
    {
        _data         = data;
        _first_byte   = offset;
        _last_byte    = offset + length - 1;

        _current_byte = _first_byte;
    }

    public byte nextByte() 
    {
        if (_current_byte > _last_byte)
        {
            String msg = "ByteArrayInputBuffer: No more bytes.";
            throw new IndexOutOfBoundsException(msg);
        }

        return _data[_current_byte++];
    }

    public void nextBytes(byte[] output, int offset, int length)
    {
        int bytes_left = _last_byte - _current_byte + 1;
        if (length > bytes_left)
        {
            String msg =
                "ByteArrayInputBuffer: Requested too many bytes.  " +
                "requested=" + length + " available=" + bytes_left;
            throw new IndexOutOfBoundsException(msg);
        }

        System.arraycopy(_data, _current_byte, output, offset, length);
        _current_byte += length;
    }

    public void nextBytes(OutputBuffer buffer) 
    {
        int bytes_left = _last_byte - _current_byte + 1;
        buffer.add(_data, _current_byte, bytes_left);
        _current_byte += bytes_left;
    }

    public void skipBytes(int length)
    {
        int bytes_left = _last_byte - _current_byte + 1;
        if (length > bytes_left)
        {
            String msg =
                "ByteArrayInputBuffer: Cannot skip that many bytes.  " +
                "requested=" + length + " available=" + bytes_left;
            throw new IndexOutOfBoundsException(msg);
        }

        _current_byte += length;
    }

    public InputBuffer subBuffer(int length)
    {
        if (_current_byte + length > _last_byte)
        {
            String msg = "ByteArrayInputBuffer: requested too many bytes";
            throw new IndexOutOfBoundsException(msg);
        }

        ByteArrayInputBuffer sub_buffer =
            new ByteArrayInputBuffer(_data, _current_byte, length);
        _current_byte += length;

        return sub_buffer;
    }

    /**
     * Returns the current offset into the original array
     */
    public int getOffset()
    {
	return _current_byte;
    }
    
    /**
    ***  Return the number of bytes remaining in the buffer.
    **/
    public int bytesRemaining()
    {
        return (_last_byte - _current_byte + 1);
    }

    /**
    ***  Reuse the buffer with a new array of bytes.
    **/
    public void reset(byte[] data)
    {
        reset(data, 0, data.length);
    }

    /**
    ***  Reuse the buffer with a new array of bytes.
    **/
    public void reset(byte[] data, int offset, int length)
    {
        _data         = data;
        _first_byte   = offset;
        _last_byte    = offset + data.length - 1;
        _current_byte = offset;
    }

    public Object clone()
        throws CloneNotSupportedException
    {
	ByteArrayInputBuffer clone = (ByteArrayInputBuffer) super.clone();

        clone._data         = this._data;
        clone._first_byte   = this._first_byte;
        clone._last_byte    = this._last_byte;
        clone._current_byte = this._current_byte;

	return clone;
    }
}
