/**
 * Logger.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.util;

import ostore.util.Carp;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Random;
import java.io.FileWriter;
import java.io.BufferedWriter;
import java.io.IOException;


/**
 * A logging utility.
 * <p>
 * Each stage should open a <code>Logger</code> with a good filename.  Try
 * the ${InternalLog} variable in run-experiment.
 * <p>
 * If two stages open the same <code>Logger</code>, their output will be
 * interleaved, but individual entries will remain atomic. 
 * <p>
 * The backing file is wrapped in a <code>BufferedWriter</code> for
 * efficient writes.  Therefore new entries may not appear in the file instantly.
 *
 * @author	Dennis Geels
 * @version	$Id: Logger.java,v 1.3 2002/05/03 19:20:33 geels Exp $
 */

public class Logger
{
    ////////////////////       CONSTANTS	////////////////////
    ////////////////////    STATIC VARIABLES 	////////////////////

    private static HashMap loggers;
    
    ////////////////////       VARIABLES	////////////////////

    private BufferedWriter out;
    
    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {
	loggers = new HashMap();
	// Move shutdown hook to stages, so last-minute data can be written
	// before flushing and closing the logs.
	//	Runtime.getRuntime().addShutdownHook( new LogFlusher() );
    }
    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Creates a <code>Logger</code> which outputs to the named file.
     */
    private Logger( String filename ) throws IOException
    {
	out = new BufferedWriter( new FileWriter( filename ) );
    }
    
    ////////////////////     STATIC METHODS 	////////////////////

    /**
     * Returns the <code>Logger</code> for the specified file.  This method
     * will create a new <code>Logger</code> if necessary.
     */
    public static Logger open( String filename ) throws IOException
    {
	Logger l;
	synchronized( loggers ) {
	    l = (Logger)loggers.get( filename );
	    if( l == null ) {
		l = new Logger( filename );
		loggers.put( filename, l );
	    }
	}
	return l;
    }

    public static void main( String args[] ) throws IOException
    {
	int num = 10;
	Logger[] ls = new Logger[num];
	for( int i=0; i<num; i++ ) {
	    ls[i] = open( "/tmp/foo" + i );
	}
	Random r = new Random();
	while( true ) {
	    int i = r.nextInt( num );
	    ls[i].write( "Hello, World" );
	}
    }
    
    //////////////////// 	METHODS		////////////////////

    /**
     * Appends an entry to the log file, adding a timestamp and newline.
     * <p>
     * <code>IOException</code>s are ignored, for laziness.
     */
    public void write( String entry )
    {
	try {
	    // Serialize writes
	    synchronized( out ) {
		// TODO: use artificial time in simulator?
		long time = System.currentTimeMillis();
		out.write( time + " :" + entry + "\n" );
	    }
	} catch( IOException ioe ) {
	    Carp.carp( ioe );
	    Carp.carp( "Ignoring exception" );
	}
    }

    /**
     * Flushes the log to disk.
     */
    public void flush()
    {
	try {
	    //	    write( "Flushing to disk" );
	    out.flush();
	} catch( IOException ioe ) {
	    Carp.carp( ioe );
	    Carp.carp( "Ignoring exception" );
	}
    }

    ////////////////////     INNER CLASSES	////////////////////

    /**
     * A shutdown hook that flushes all the logs
     *
     * @see Runtime#addShutdownHook
     */
    private static class LogFlusher extends Thread
    {
	public void run()
	{
	    //	    Carp.out( "Shutdown hook running" );
	    synchronized( loggers ) {
		Iterator iter = loggers.values().iterator();
		while( iter.hasNext() ) {
		    Logger l = (Logger)iter.next();
		    l.flush();
		}
	    }
	    return;
	}
    }
}
