/**
 * QSClone.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.util;
import ostore.network.NetworkInputBuffer;
import java.math.BigInteger;
import java.util.LinkedList;
import java.util.Iterator;

public class QSClone {

    /**
     * Clones a quick serializable object using the serialize function,
     * rather than the Java clone function.  For use in testing
     * serialization code, and for cloning QS objects that don't implement
     * the clone function (as done, for example, in the Router stage).
     */
    public static QuickSerializable qs_clone (QuickSerializable value) {
	QSCloneBuffer buf = new QSCloneBuffer ();
	buf.add (value);
	try {
	    return buf.nextObject ();
	}
	catch (QSException e) {
	    Carp.die (e);
	}

	return null;
    }

    /**
     * This class is private for a reason; using it outside of QSClone will
     * result in your being sacked.  :)
     */
    public static class QSCloneBuffer implements InputBuffer, OutputBuffer {

	private static class ByteArray {
	    ByteArray (byte [] d, int o, int l) {
		data = d; offset = o; length = l;
	    }
	    byte [] data;
	    int offset, length;
	}

	private LinkedList objects = new LinkedList ();

	public Object clone () throws CloneNotSupportedException {
	    QSCloneBuffer result = (QSCloneBuffer) super.clone ();
            result.objects = (LinkedList) objects.clone();
	    return result;
	}

	public byte nextByte () {
	    byte rv = (byte)-1;
	    Object o = objects.getFirst();
	    if(o instanceof Byte)
		rv = ((Byte)objects.removeFirst()).byteValue ();
	    else if(o instanceof ByteArray)
	    {
		byte[] data = {0};
		nextBytes(data, /*offset-*/0, /*length=*/1);
		rv = data[0];
	    }
	    else
 		Carp.die("QSClone.QSCloneBuffer.nextByte:" +
			 " Called with object.getFirst()==" + o + 
			 " class==" + (o==null?null:o.getClass().getName()) +
                         ", But do not understand that type.  Dying...");
	    return rv;
	}

	public void nextBytes (byte [] output, int offset, int length) {
	    do
            {
		ByteArray ba = (ByteArray) objects.removeFirst ();
                int bytes_available = ba.length;
		if (bytes_available > length) 
		{
                    System.arraycopy(ba.data, ba.offset,
                                     output, offset, length);

                    objects.addFirst(new ByteArray(ba.data,
                                                   ba.offset + length,
                                                   ba.length - length));

                    offset += length;
                    length = 0;
		}
                else
		{
                    System.arraycopy(ba.data, ba.offset,
                                     output, offset, bytes_available);

                    offset += bytes_available;
                    length -= bytes_available;
                }
	    } while (length > 0);
            return;
	}

	public void nextBytes (OutputBuffer output)
        {
	    Iterator i = objects.iterator();
	    while (i.hasNext())
            {
                //ByteArray ba = (ByteArray) i.next();
                //output.add(ba.data);

		Object obj = i.next();
		if (obj instanceof Byte)
		    output.add (((Byte) obj).byteValue ());
		else if (obj instanceof Short)
		    output.add (((Short) obj).shortValue ());
		else if (obj instanceof Integer)
		    output.add (((Integer) obj).intValue ());
		else if (obj instanceof Long)
		    output.add (((Long) obj).longValue ());
		else if (obj instanceof Double)
		    output.add (((Double) obj).doubleValue ());
		else if (obj instanceof Boolean)
		    output.add (((Boolean) obj).booleanValue ());
		else if (obj instanceof String)
		    output.add ((String) obj);
		else if (obj instanceof ByteArray) {
		    ByteArray ba = (ByteArray) obj;
		    output.add (ba.data, ba.offset, ba.length);
		}
		else {
		    throw new AssertionViolatedException ("unknown");
		}
            }

	    return;
	}

	public short nextShort () {
	    short rv = -1;
	    Object o = objects.getFirst();
	    if(o instanceof Short)
		rv = ((Short)objects.removeFirst()).shortValue ();
	    else if(o instanceof ByteArray)
		rv = InputBufferImpl.nextShort(this);
	    else
 		Carp.die("QSClone.QSCloneBuffer.nextShort:" +
			 " Called with object.getFirst()==" + o + 
			 " class==" + (o==null?null:o.getClass().getName()) +
                         ", but do not understand that type.  Dying...");
	    return rv;
	}

	public int nextInt () {
	    int rv = -1;
	    Object o = objects.getFirst();
	    if(o instanceof Integer)
		rv = ((Integer)objects.removeFirst()).intValue ();
	    else if(o instanceof ByteArray)
		rv = InputBufferImpl.nextInt(this);
	    else
 		Carp.die("QSClone.QSCloneBuffer.nextInt:" +
			 " Called with object.getFirst()==" + o + 
			 " class==" + (o==null?null:o.getClass().getName()) +
                         ", but do not understand that type.  Dying...");
	    return rv;
	}

	public long nextLong () {
	    long rv = -1;
	    Object o = objects.getFirst();
	    if(o instanceof Long)
		rv = ((Long)objects.removeFirst()).longValue ();
	    else if(o instanceof ByteArray)
		rv = InputBufferImpl.nextLong(this);
	    else
 		Carp.die("QSClone.QSCloneBuffer.nextLong:" +
			 " Called with object.getFirst()==" + o + 
			 " class==" + (o==null?null:o.getClass().getName()) +
                         ", but do not understand that type.  Dying...");
	    return rv;
	}

	public double nextDouble () {
	    double rv = -1.0;
	    Object o = objects.getFirst();
	    if(o instanceof Double)
		rv = ((Double)objects.removeFirst()).doubleValue ();
	    else if(o instanceof ByteArray)
		rv = InputBufferImpl.nextDouble(this);
	    else
 		Carp.die("QSClone.QSCloneBuffer.nextDouble:" +
			 " Called with object.getFirst()==" + o + 
			 " class==" + (o==null?null:o.getClass().getName()) +
                         ", but do not understand that type.  Dying...");
	    return rv;
	}

	public boolean nextBoolean () {
	    boolean rv = false;
	    Object o = objects.getFirst();
	    if(o instanceof Boolean)
		rv = ((Boolean)objects.removeFirst()).booleanValue ();
	    else if(o instanceof ByteArray)
		rv = InputBufferImpl.nextBoolean(this);
	    else
 		Carp.die("QSClone.QSCloneBuffer.nextBoolean:" +
			 " Called with object.getFirst()==" + o + 
			 " class==" + (o==null?null:o.getClass().getName()) +
                         ", but do not understand that type.  Dying...");
	    return rv;
	}

	public String nextString () {
	    String rv = null;
	    Object o = objects.getFirst();
	    if(o instanceof String)
		rv = (String)objects.removeFirst();
	    else if(o instanceof ByteArray)
		rv = InputBufferImpl.nextString(this);
	    else
 		Carp.die("QSClone.QSCloneBuffer.nextString:" +
			 " Called with object.getFirst()==" + o + 
			 " class==" + (o==null?null:o.getClass().getName()) +
                         ", but do not understand that type.  Dying...");
	    return rv;
	}

	public BigInteger nextBigInteger () {
	    return InputBufferImpl.nextBigInteger (this);
	}

	public QuickSerializable nextObject () throws QSException {
	    return InputBufferImpl.nextObject (this);
	}

	public InputBuffer subBuffer (int length) {
	    CountBuffer cb = new CountBuffer ();
	    QSCloneBuffer result = new QSCloneBuffer ();
	    while (cb.size () < length) {
		Object obj = objects.removeFirst ();
		if (obj instanceof Byte)
		    cb.add (((Byte) obj).byteValue ());
		else if (obj instanceof Short)
		    cb.add (((Short) obj).shortValue ());
		else if (obj instanceof Integer)
		    cb.add (((Integer) obj).intValue ());
		else if (obj instanceof Long)
		    cb.add (((Long) obj).longValue ());
		else if (obj instanceof Double)
		    cb.add (((Double) obj).doubleValue ());
		else if (obj instanceof Boolean)
		    cb.add (((Boolean) obj).booleanValue ());
		else if (obj instanceof String)
		    cb.add ((String) obj);
		else if (obj instanceof ByteArray) {
		    ByteArray ba = (ByteArray) obj;
		    int needed = length - cb.size();
		    if( needed >= ba.length ) {
			// Add the whole ByteArray
			cb.add (ba.data, ba.offset, ba.length);
		    } else {
			// Split the array in two; save the second part.
			ByteArray first =
			    new ByteArray( ba.data, ba.offset, needed );
			ByteArray second =
			    new ByteArray( ba.data, ba.offset + needed,
					   ba.length - needed );
			objects.addFirst( second );
			// Add the first part, not the whole:
			cb.add( first.data, first.offset, first.length );
			obj = first;
		    }
		}
		else {
		    throw new AssertionViolatedException ("unknown");
		}

		result.objects.addLast (obj);
	    }

	    if (cb.size () > length)
		throw new AssertionViolatedException ("BUG: length=" +
			length + ", cb.size()=" + cb.size () + 
			", result=(" + result + ")");

	    return result;
	}

	public void add (byte input) {
	    objects.addLast (new Byte (input));
	}

	public void add (byte[] value) {
	    add (value, 0, value.length);
	}

	public void add (byte[] value, int offset, int length) {
	    // objects.addLast (new ByteArray (value, offset, length));
	    // TODO: remove this copy if it still works afterwards
	    byte [] copy = new byte [length];
	    System.arraycopy (value, offset, copy, 0, length);
	    objects.addLast (new ByteArray (copy, 0, length));
	}

	public void add (QuickSerializable value) {
	    OutputBufferImpl.add (this, value);
	}

	public void add (InputBuffer buffer) {
	    if (buffer instanceof QSCloneBuffer) {
		QSCloneBuffer buf = (QSCloneBuffer) buffer;
		while (! buf.objects.isEmpty ())
		    objects.addLast (buf.objects.removeFirst ());
	    }
            else if (buffer instanceof ByteArrayInputBuffer)
            {
                ((ByteArrayInputBuffer) buffer).nextBytes((OutputBuffer) this);
            }
            else if (buffer instanceof NetworkInputBuffer)
            {
                ((NetworkInputBuffer) buffer).nextBytes((OutputBuffer) this);
            }
	    else
            {
		Carp.die("Called to add data from buffer of type " + 
                         buffer.getClass ().getName () + ": " + buffer +
                         ", but do not understand that type.  Dying...");
	    }
	}

	public void add (boolean b) {
	    objects.addLast (new Boolean (b));
	}

	public void add (short i) {
	    objects.addLast (new Short (i));
	}

	public void add (int i) {
	    objects.addLast (new Integer (i));
	}

	public void add (long i) {
	    objects.addLast (new Long (i));
	}

	public void add (double x) {
	    objects.addLast (new Double (x));
	}

	public void add (String s) {
	    objects.addLast (s);
	}

	public void add (BigInteger value) {
	    OutputBufferImpl.add (this, value);
	}

	public String toString () {

	    if (objects.isEmpty ())
		return "QSCloneBuffer empty";

	    String result = "QSCloneBuffer ";
	    Iterator i = objects.iterator ();

	    while (true) {
		
		Object obj = i.next ();

		if (obj instanceof Byte)
		    result += "Byte 0x" + Integer.toHexString (
			    0xff & ((Byte) obj).byteValue ());
		else if (obj instanceof Short)
		    result += "Short 0x" + Integer.toHexString (
			    0xffff & ((Short) obj).shortValue ());
		else if (obj instanceof Integer)
		    result += "Integer 0x" + Integer.toHexString (
			    ((Integer) obj).intValue ());
		else if (obj instanceof Long)
		    result += "Long 0x" + Long.toHexString (
			    ((Long) obj).longValue ());
		else if (obj instanceof Double)
		    result += "Double " + obj;
		else if (obj instanceof Boolean)
		    result += "Boolean " + obj;
		else if (obj instanceof String)
		    result += "String \"" + obj + "\"";
		else if (obj instanceof ByteArray) {
		    ByteArray ba = (ByteArray) obj;
		    result += "ByteArray data.length=" + ba.data.length + 
			", offset=" + ba.offset + ", length=" + ba.length;
		}
		else {
		    throw new AssertionViolatedException ("unknown");
		}

		if (i.hasNext ())
		    result += ", ";
		else
		    break;
	    }

	    return result;
	}
    }
}
