/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;
import java.security.*;

/**
 * SHA1Hash is the hash class used by all entities which are SHA1 hashes.
 * (In the prototype, this means Guids and VHashes).
 *
 * @author     Chris WElls
 * @version    $Id: SHA1Hash.java,v 1.43 2003/06/18 21:09:45 srhea Exp $
 */
public class SHA1Hash implements SecureHash {

	/**
	 * The number of bytes all <code>SHA1Hash</code>s are composed of
	 */
	public static final int NUM_BYTES = 20;

        /**
	 * constants to determine what type of SHA1 to test
	 */
        private static final int TEST_CRYPTIX = 0;
        private static final int TEST_SUN = 1;
        private static final int TEST_C_AND_JNI = 2;
        private static final int TEST_JUST_C = 3;

	/**
	 * This is the value of the null hash for SHA1
	 */
	public static final byte[] NULL_HASH_VALUE =
		{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

	/**
	 * This is the null hash for SHA1
	 */
	public static final SHA1Hash NULL_HASH = new SHA1Hash();

        static
        {
	    //  Add to the list of security providers.
	    //Security.addProvider(new sun.security.provider.Sun());
	    Security.addProvider(new cryptix.jce.provider.CryptixCrypto());
	    Security.addProvider(new ostore.util.SHA1NativeProvider());
        }

	/**
	 * These are the bytes in this Hash.
	 */
	protected byte _bytes[];

	/**
	 * default constructor (creates null hash)
	 */
	public SHA1Hash() {
		_bytes = new byte[NUM_BYTES];
	}

	/**
	 * copy constructor (creates new hash with value of old hash)
	 * @param shash == old hash whose value we are to copy
	 */
	public SHA1Hash(SHA1Hash shash) {
		_bytes = new byte[NUM_BYTES];
		for (int i = 0; i < NUM_BYTES; ++i) {
			_bytes[i] = shash._bytes[i];
		}
	}

	/**
	 * hashes the given buffer to create a new <code>SHA1Hash</code>
	 * @param the array of bytes over which to hash
	 */
	public SHA1Hash(byte[] bytes) {
		hash(bytes);
	}

        /**
	 * takes the byte array and just uses it. way to control what your
	 * guids look like
	 */
        public SHA1Hash(byte[] bytes, boolean fake) {
	  if (fake) 
	    this._bytes = bytes;
	  else hash(bytes);
	}

	/**
	 * hashes the given buffer to create a new <code>SHA1Hash</code>
	 * @param the array of bytes over which to hash
	 * @param offset in bytes at which to start hashing
	 * @param number of bytes over which to hash
	 */
	public SHA1Hash(byte[] bytes, int offset, int len) {
		hash(bytes, offset, len);
	}

	/**
	 * Construct out of an input buffer.
	 */
	public SHA1Hash (InputBuffer buffer) {
	    _bytes = new byte [NUM_BYTES];
	    buffer.nextBytes (_bytes, 0, NUM_BYTES);
	}

	/**
	 * hashes over the bytes of the input String
	 * @param string over which to hash
	 */
	public SHA1Hash(String s) {
		hash(s);
	}

	/**
	 * hashes over the input object
	 * @param object over which to hash
	 */
	public SHA1Hash (QuickSerializable d) {
            hash(d);
	}

	/**
	 * Computes and returns the digest of the specified buffer.
	 * @param buffer	A <code>SecureHashBuffer</code>.
	 */
	public SHA1Hash (SHA1HashBuffer buffer) {
	    _bytes = buffer.compute_bytes();	    
	}

	/**
	 * Create an exact copy of this hash
	 * @return new SHA1Hash identical in value to this one
	 * @exception CloneNotSupportedException if clone() is not supported
	 */
        public Object clone() throws CloneNotSupportedException {
		return super.clone();
	}

	/**
	 * This routine compares this <code>SHA1Hash</code> to another
	 * <code>Object</code>
	 * @param o 	SHA1Hash
	 * @return -1 if <code>this < o</code>, 0 if <code>this == o</code>,
	 * 	or 1 if <code>this > o</code>. 
	 * @exception ClassCastException if <code>o</code> is not a SHA1Hash
	 */
	public int compareTo( Object o ) throws ClassCastException {
	    if( o == null ) {
		throw new NullPointerException();
	    } else if( !(o instanceof SHA1Hash) ) {
		throw new ClassCastException( "cannot compareTo non-SHA1Hash" );
	    } else if( o == this ) {
		return 0;
	    } else {
		SHA1Hash other = (SHA1Hash)o;
		// Compare each successive byte of the two hashes
		for (int i = 0; i < NUM_BYTES; ++i) {
			if( _bytes[i] < other._bytes[i] )
				return -1;
			else if( _bytes[i] > other._bytes[i] )
				return 1;
		}
		return 0;
	    }
	}

	/**
	 * Return the number of bytes used in the representation of this hash
	 * (for example, SHA1 uses 20 bytes)
	 * @return number of bytes used in the implementation
	 */
        public int size() {
		return NUM_BYTES;
	}

	/**
	 * Determines if this SecureHash is a NULL hash or not.
	 * @return true iff this Hash has a NULL value
	 */
	public boolean isNull() {
		for (int i = 0; i < NUM_BYTES; ++i) {
			if (_bytes[i] != NULL_HASH_VALUE[i])
				return false;
		}
		return true;
	}

	/**
	 * Returns the implementation's NULL hash
	 * (typically all zeroes).
	 * @return the implementation's null hash
	 */
	public SecureHash nullHash() {
		return NULL_HASH;
	}

	/**
	 * Returns the bytes contained in this hash
	 * @return the bytes of this hash
	 */
	public byte[] bytes() {
		return _bytes;
	}

	/**
	 * Compares this SecureHash with another hash, and returns
	 * true iff they are the same type of hash and have the
	 * same byte representation
	 * @param the object with which to compare this SecureHash
	 * @return true iff this and o contain the same byte values
	 */
	public boolean equals(Object other) {
	    boolean equal;
	    if (other == null) {
		equal = false;
	    } else if (other == this) {
		equal = true;
	    } else if (!(other instanceof SHA1Hash)) {
		equal = false;
	    } else {
		SHA1Hash rhs = (SHA1Hash)other;
		equal = true;
		
		for (int i = 0; i < NUM_BYTES; ++i) {
		    if (_bytes[i] != rhs._bytes[i]) {
			equal = false;
			break;
		    }
		}
	    }
	    return equal;
	}

	/**
	 * Calculates a Java hash code for this SecureHash object
	 * (by taking its lower order bytes)
	 * @return the Java hash code for this SecureHash
	 */
	public int hashCode(){
		int result = 0; 
		for (int i = 0; i < 4; ++i) {
			result <<= 8;
			result |= 0xff & ((int) _bytes[i]);
		}
		return result;
	}

	/**
	 * Calculates a Java hash code for this SecureHash object
	 * (by taking its lower order bytes)
	 * @return the Java hash code for this SecureHash
	 */
	public long lower64bits(){
		long result = 0; 
		for (int i = 0; i < 8; ++i) {
			result <<= 8;
			result |= 0xff & ((int) _bytes[i]);
		}
		return result;
	}

	/**
	 * Hashes over the input bytes, starting at offset, going for len
	 * @param byte array over which to hash
	 * @param offset in bytes[] at which to begin hashing
	 * @param number of bytes over which to hash
	 */
	public void hash(byte bytes[], int offset, int len) {
		try {
		    MessageDigest md = null;
		    if(SHA1NativeIF.available)
			md = MessageDigest.getInstance("SHA", "SHA1Native");
		    else
			md = MessageDigest.getInstance("SHA", "SUN");

			md.update(bytes, offset, len);
			_bytes = new byte[NUM_BYTES];
			md.digest(_bytes, 0, NUM_BYTES);

		} catch (java.security.NoSuchAlgorithmException e) {
			System.err.println(
				"SHA1Hash: cannot locate SHA-1 algorithm");
		} catch (java.security.NoSuchProviderException e) {
			System.err.println(
				"SHA1Hash: cannot locate " + 
				(SHA1NativeIF.available?"SHA1Native":"SUN")+
				 " provider");
		} catch (java.security.DigestException e) {
			System.err.println(
				"SHA1Hash: DigestException " + e);
		}
	}

	/**
	 * Hashes over the input bytes
	 * @param bytes the byte array over which to hash
	 */
	public void hash(byte bytes[]) {
		hash(bytes, 0, bytes.length);
	}

	/**
	 * Hashes over a QuickSerializable object
	 * @param o the object over which to hash
	 */
	public void hash(QuickSerializable o) {
	    SHA1HashBuffer digest = new SHA1HashBuffer();
	    // TODO: kludge!  We don't want the type code in here.
	    o.serialize (digest);
	    _bytes = digest.compute_bytes();
	}

	/**
	 * Hashes over the bytes of the input String
	 * @param str the String over which to hash
	 */
	public void hash(String str) {
		hash(str.getBytes());
	}

	/**
	 * This routine concatenates two hashes and stores the result
	 * in the input byte array.
	 * @param hash1 the first hash to concatenate
	 * @param hash2 the second hash to concatenate
	 * @param buf the byte array in which to storet the concatenation
	 */
	public static void concatenateHashes(SHA1Hash hash1, SHA1Hash hash2,
						byte buf[])
    	{
		ByteArrayOutputBuffer ob = new ByteArrayOutputBuffer (buf);
		hash1.serialize (ob);
		hash2.serialize (ob);
	}

	/**
	 * Verifies a QuickSerializable object
	 * @param the object over which to hash
	 * @return true iff the input object hashes to
	 *                   a hash value equal to this hash value
	 */
	public boolean verify(QuickSerializable o) {
            SecureHashBuffer buffer = new SHA1HashBuffer();
	    // FIXME: don't want type code, so we do it this way for now
	    o.serialize (buffer);
            SecureHash hash = buffer.toHash();
            return equals(hash);
	}
	
	/**
	 * Converts a few bytes of this SecureHash to a human-readable String
	 * @return human-readable String representation of this Hash
	 */
	public java.lang.String toString() {
	    int num_to_print = 6;
	    String s = "(SHA1Hash:" +
		ByteUtils.print_bytes( _bytes, 0, num_to_print ) + ')';
	    return s;
	}

	/**
	 * Converts all bytes of this SecureHash to a human-readable String
	 * @return human-readable String representation of this Hash
	 * containing every nibble of this Hash's value
	 */
	public java.lang.String fullString() {
		String s = "(SHA1Hash:" +
		    ByteUtils.print_bytes( _bytes, 0, _bytes.length ) + ')';
		return s;
	}    


	public void serialize (OutputBuffer buffer) {
	    buffer.add (_bytes);
	}

        /**
	 * Tests the throughput of a given type of SHA1, with blocks of
	 * size <code>arrSz</code>.  Prints throughput and iterations/ms
	 * to <code>System.err</code>.
	 */
        private static void testTput( int arrSz, int type ) {
	    java.util.Random r = new java.util.Random();
	    testTput( arrSz, type, r.nextLong() );
	} 

        /**
	 * Tests the throughput of a given type of SHA1, with blocks of
	 * size <code>arrSz</code>.  Prints throughput and iterations/ms
	 * to <code>System.err</code>.
	 */
        private static void testTput( int arrSz, int type, long seed ) {

	    java.util.Random random = new java.util.Random( seed );

	    SecureHash hash = new ostore.util.SHA1Hash();
	    MessageDigest md = null;

	    try {
		switch( type ) {
		case TEST_CRYPTIX:
		    md = MessageDigest.getInstance( "SHA", "CryptixCrypto" );
		    break;
		case TEST_SUN:
		    md = MessageDigest.getInstance( "SHA", "SUN" );
		    break;
		case TEST_C_AND_JNI:
		    md = MessageDigest.getInstance( "SHA", "SHA1Native" );
		    break;
		default:
		}
	    } catch( NoSuchProviderException nspe ) {
		Carp.die( nspe );
	    }
	    catch( NoSuchAlgorithmException nspe ) {
		Carp.die( nspe );
	    }
	    
	    int length = 4096;
	    int iterations = 536870912/arrSz;
	    long time = 0;

	    if( type != TEST_JUST_C ) {
	      byte[][] data = new byte[2*length-1][];
	      
	      // populate the data set
	      for(int i = 0; i < data.length; i++) {
		data[i] = new byte[arrSz];
		random.nextBytes(data[i]);
	      }
	      byte[] buffer = new byte[data[0].length];
	      
	      time = System.currentTimeMillis();
	      
	      // Loop over the data set, updating and digesting 512 MB worth
	      // of data
	      for(int i = 0; i < iterations; i++) {
		md.update(data[i%data.length],0,data[0].length);
		try {
		  md.digest(hash.bytes(), 0, 
			    NUM_BYTES);
		} catch( java.security.DigestException de ) {
		  Carp.die( de );
		}
		md.reset();
	      }
	    
	      time = System.currentTimeMillis() - time;
	    } else {

	      // measure in C
	      time = (new SHA1NativeIF()).
		testTput( arrSz, iterations );

	    }

	    double tput = ((((double)(iterations*arrSz))/1048576.0)/
			   (double)((double)time/1000.0));
	    double itms = (double)((double)iterations/(double)time);
	    System.err.println( arrSz + "\t" + tput + "\t" + "\t" + itms );
	    System.err.flush();
	    System.gc();
	}

        private static void testAll() {

	    System.err.println( "Cryptix:\narrSz\ttput (Mb/s)" + 
				"\t\t\titerations/ms");
	    for(int j = 5; j < 30000; j*=2) {
		testTput( j, TEST_CRYPTIX );
	    }
	    System.err.println( "\n\nSun:\narrSz\ttput (Mb/s)" + 
				"\t\t\titerations/ms");
	    for(int j = 5; j < 30000; j*=2) {
		testTput( j, TEST_SUN );
	    }
	    System.err.println( "\n\nSHA1Native:\narrSz\ttput (Mb/s)" + 
				"\t\t\titerations/ms");
	    for(int j = 5; j < 30000; j*=2) {
		testTput( j, TEST_C_AND_JNI );
	    }
	    System.err.println( "\n\nSHA1Native, Just C:\narrSz\ttput (Mb/s)"+ 
				"\t\t\titerations/ms");
	    for(int j = 5; j < 30000; j*=2) {
		testTput( j, TEST_JUST_C );
	    }
	}
    
    /**
    ***  Measures the speed of various hash implementations.
    **/
    public static void main(String args[])
    {
        testAll();
    }
}
