#!/usr/bin/perl -w

use strict;
use Getopt::Std;
use File::Copy;
use Tools;
use vars qw($opt_a $opt_c $opt_j $opt_u $opt_k $opt_l $opt_t $opt_x $opt_n $opt_h $opt_s $opt_r);

sub usage
{
    select(STDERR);
    print <<'EOUsage';

run-stream-test.pl [options]
  Options:
    -h           - Display this help message.
    -c MACHINES  - Use MACHINES machines to apply load.  If the number
                   of machines requested is greater than the number of
                   test stages requested, the extra machines will remain
                   idle through the test.
    -n NODES     - Create a network of NODES nodes.  If this number is
                   greater than the number of tester nodes, the excess
                   nodes are assigned to be replica nodes.
    -t TESTERS   - Sweep the number of test stages used of the range
                   described by TESTERS.  The format for notating the
                   range is described below.
    -u UPDT_SIZE - Sweep the size of the update submitted by test stages
                   across the range describe by UPDT_SIZE.  The format
                   for notating the range is described below.  Note, the
                   size of the data object is set to twice the length of
                   the update size.
    -r UPDT_RATE - Perform UPDT_RATE updates per second.
    -j DEPTHS    - Create replicas a various depths in the network.  The
                   argument should be a comma-separated list of depths.
    -l LENGTH    - Run each test for LENGTH seconds.
    -x ROC       - Relocate the test in the X-cluster to start using
                   machine xROC.millennium.berkeley.edu.
		   NOTE: if no -x option is specified, the test will use
		   the machines listed in the OSTORE_SERVERS environment
		   variable.
    -k KEY_SIZE  - Use keys of of length KEY_SIZE bits.  Valid values
                   are 512 and 1024.  The default value is 512.

  Ranges can be specified in a very flexible manner.  Most generally,
  each range has a 'start' value, an 'end' value, a 'skip operator',
  and a 'skip value'.  The skip operator is one of the four basic
  arithmetic operators (+, -, *, and /).  During execution, when the
  range operator is '+' or '*', the range leads to execution as
    for (x = $start; x <= $end; x = x $skip_op $skip_value) { ; }
  When the operator is '-' or '/', the range leads to execution as
    for (x = $start; x >= $end; x = x $skip_op $skip_value) { ; }

  A range is notated as '$start-$end/$skip_op$skip_value'.  Other
  simpler forms are also accepted.  These forms include '$start' which
  executes only one point, '$start-$end' which uses the '+' operator
  with a skip value of 1, and '$start-$end/$skip_value' which defaults
  to use the '+' operator.  Examples of valid ranges include

    o  28
    o  28-56
    o  2-10/2
    o  128-1024/*2
    o  2048-128//2
EOUsage

    exit(1);
}

# Set a nice signal handler.
$SIG{INT} = \&slow_death;
$SIG{QUIT} = \&slow_death;
$SIG{HUP} = \&slow_death;
$SIG{PIPE} = \&slow_death;
$SIG{TERM} = \&slow_death;

#  Constants.
my $STREAM_RESULT_LOG = "$Tools::tmp_dir/stream_result_log.$$.log";
my @SERVERS = split /\s+/, $ENV{"OSTORE_SERVERS"} if ($ENV{"OSTORE_SERVERS"});

#  Test parameters.
my $SKIP_ARCHIVE = 'true';
my $TEST_MACHINES = 1;
my $TOTAL_NODES;
my $TEST_LENGTH = 10;
my $CLUSTER_MACHINE = 1;
my $KEY_SIZE = 512;
my $JOIN_DEPTHS = "0";

my @TEST_STAGES_SPEC;
my $TEST_STAGES_START = 1;
my $TEST_STAGES_END   = 1;
my $TEST_STAGES_OP    = '+';
my $TEST_STAGES_SKIP  = 1;

my @UPDATE_SIZE_SPEC;
my $UPDATE_SIZE_START = 1024;
my $UPDATE_SIZE_END   = 1024;
my $UPDATE_SIZE_OP    = '+';
my $UPDATE_SIZE_SKIP  = 1;
my $UPDATE_RATE       = 4;

my $exp_pid;	

#  Get the user-defined parameters.
my $opts_ok = getopts('han:c:j:u:r:k:l:t:x:s:');

if ($opt_h) { &usage; }
if ($opt_a) { $SKIP_ARCHIVE = 'false'; }
if ($opt_c) { $TEST_MACHINES = $opt_c; }
if ($opt_l) { $TEST_LENGTH = $opt_l; }
if ($opt_x) { @SERVERS = map "x$_.millennium.berkeley.edu", ($opt_x .. 42); }
if ($opt_k) { $KEY_SIZE = $opt_k; }
if ($opt_j) { $JOIN_DEPTHS = $opt_j; }
if ($opt_n) { $TOTAL_NODES = $opt_n; }
if ($opt_r) { $UPDATE_RATE = $opt_r; }
if ($opt_t)
{
    @TEST_STAGES_SPEC = &parse_range_spec($opt_t);
    my @spec = @TEST_STAGES_SPEC;
    $TEST_STAGES_START = shift(@spec);
    $TEST_STAGES_END   = shift(@spec);
    $TEST_STAGES_OP    = shift(@spec);
    $TEST_STAGES_SKIP  = shift(@spec);
}
if ($opt_u)
{
    @UPDATE_SIZE_SPEC = &parse_range_spec($opt_u);
    my @spec = @UPDATE_SIZE_SPEC;
    $UPDATE_SIZE_START = shift(@spec);
    $UPDATE_SIZE_END   = shift(@spec);
    $UPDATE_SIZE_OP    = shift(@spec);
    $UPDATE_SIZE_SKIP  = shift(@spec);
}
if( $opt_s ) {
    print "seed: $opt_s\n";
    srand $opt_s;
}

#  The command line is valid if getopts returned without error and
#  there are no arguments left over.  $#ARGV returns the number of
#  remaining arguments minus one.
if ((! $opts_ok) || ($#ARGV != -1)) { &usage; }

#  Check that the set of input parameters is valid.
if ($KEY_SIZE != 512 && $KEY_SIZE != 1024)
{
    my $old_fd = select(STDERR);
    print<<EOErrMsg;
ERROR:  Invalid key size.
EOErrMsg
    select($old_fd);
    &usage();
    exit(1);
}

if (! @SERVERS)
{
    my $old_fd = select(STDERR);
    print<<EOErrMsg;
ERROR:  Undefined set of servers.
EOErrMsg
    select($old_fd);
    &usage();
    exit(1);
}

#  Run experiments to collect data.
my $update_size = $UPDATE_SIZE_START;
my $update_size_comparator = '<=';
if ($UPDATE_SIZE_OP eq '-' || $UPDATE_SIZE_OP eq '/')
{
    $update_size_comparator = '>=';
}
while (eval("$update_size $update_size_comparator $UPDATE_SIZE_END"))
{
    my $dobj_size = $update_size * 2;

    my $test_stages = $TEST_STAGES_START;
    my $test_stages_comparator = '<=';
    if ($TEST_STAGES_OP eq '-' || $TEST_STAGES_OP eq '/')
    {
        $test_stages_comparator = '>=';
    }
    while (eval("$test_stages $test_stages_comparator $TEST_STAGES_END"))
    {
        print("Running stream test:  " .
              "TestMachines=$TEST_MACHINES, TestStages=$test_stages, " .
              "KeySize=$KEY_SIZE, DobjSize=$dobj_size, " .
              "UpdateSize=$update_size, TestLength=$TEST_LENGTH\n");

        #  Create configuration files.
        my $master_cfg  =
          &write_master_config_file($test_stages, $TEST_LENGTH, $JOIN_DEPTHS);
        my $replica_cfg =
          &write_replica_config_file();
        my $tester_cfg  =
          &write_tester_config_file($dobj_size, $update_size);
        my $exp_cfg     =
          &write_experiment_file($TEST_MACHINES, $TOTAL_NODES, $test_stages,
                                 $replica_cfg, $tester_cfg);

	my $run_exp_cmd = "./run-experiment";
	my $run_exp_log = "$Tools::tmp_dir/stream-$$.log";

        #  Run the experiment.
        if ($KEY_SIZE == 512)
        {
	    $exp_pid =
		&Tools::fork_process( $run_exp_log, $run_exp_cmd, $exp_cfg );
	    &returned_zero( $exp_pid );
        }
        elsif ($KEY_SIZE == 1024)
        {
	    $ENV{"KEYSHARE_PREFIX"} = "ir-1024-4-2";
	    $exp_pid =
		&Tools::fork_process( $run_exp_log, $run_exp_cmd, $exp_cfg );
	    &returned_zero( $exp_pid );
        }
        else
        {
            print(STDERR "Script error.");
        }

        #  Read the result log.
#         my $total_ops = &analyze_result_log();
#         &write_result_to_log($SKIP_ARCHIVE, $INLINE_ARCHIVE,
#                              $TEST_MACHINES, $test_stages, $TEST_LENGTH,
#                              $KEY_SIZE, $DOBJ_SIZE, $update_size, $total_ops);

        #  Remove the files we created.
        unlink($master_cfg, $tester_cfg, $replica_cfg, $exp_cfg);


        my @internal_logs =
	    glob($Tools::tmp_dir .
		 "/experiment-$exp_pid-{Rep,Ring,Test}*-internal.log");
        print( "Running ./extract_latencies.pl -f <logs>\n");
        system( "./extract_latencies.pl", '-f', @internal_logs );

        print( "Running ./extract_stress.pl <logs>\n");
        system( "./extract_stress.pl", @internal_logs );

        $test_stages = eval("$test_stages $TEST_STAGES_OP $TEST_STAGES_SKIP");
    }

    $update_size = eval("$update_size $UPDATE_SIZE_OP $UPDATE_SIZE_SKIP");
}


exit(0);


sub parse_range_spec
{
    my $spec = shift(@_);

    my $start = 0;
    my $end   = 0;
    my $op    = '+';
    my $skip  = 1;

    #  (\d+)                  The start of the range...
    #  (-(\d+))?              ...optionally a '-' and the end of the range...
    #  (\/([-\+\*\/])?(\d+))? ...optionally a '/', an optional 
    #                         operation (-, +, *, or /) and the skip value.
    if ($spec =~ /(\d+)(-(\d+))?(\/([-\+\*\/])?(\d+))?/ )
    {
        $start = $1;
        $end   = $start;
        if (defined($3)) { $end = $3; }
        if (defined($5)) { $op = $5; }
        if (defined($6)) { $skip = $6; }
    }

    #print("start=$start \n");
    #print("end=$end   \n");
    #print("op=$op    \n");
    #print("skip=$skip  \n");

    if (($op eq '+') || ($op eq '*'))
    {
        if ($start > $end)
        {
          die("Invalid range specification.\n" .
              "  start=$start, end=$end, operation=$op\n");
        }
    }
    else
    {
        if ($end > $start)
        {
          die("Invalid range specification.\n" .
              "  start=$start, end=$end, operation=$op\n");
        }
    }

    my @range;
    push(@range, $start);
    push(@range, $end);
    push(@range, $op);
    push(@range, $skip);

    return @range;
}

sub write_result_to_log
{
    my $skip          = shift(@_);
    my $inline        = shift(@_);
    my $test_machines = shift(@_);
    my $test_stages   = shift(@_);
    my $test_length   = shift(@_);
    my $key_size      = shift(@_);
    my $dobj_size     = shift(@_);
    my $update_size   = shift(@_);
    my $total_ops     = shift(@_);

    my $ops_tput  = $total_ops / $test_length;
    my $data_tput = ($total_ops * $update_size) / $test_length;

    my $log =
      "SkipArchive=${skip} " .
      "InlineArchive=${inline} " .
      "TestMachines=${test_machines} " .
      "TestStages=${test_stages} " .
      "TestLength=${test_length}s " .
      "KeySize=${key_size} " .
      "DobjSize=${dobj_size} " .
      "UpdateSize=${update_size} " .
      "TotalOps=${total_ops} " .
      "UpdateTput=${ops_tput}ops/s " .
      "DataTput=${data_tput}bytes/s\n";

    open(RESULT_LOG, ">> $STREAM_RESULT_LOG") ||
      die("Could not open result log ($STREAM_RESULT_LOG).");
    print(RESULT_LOG "$log");
    close(RESULT_LOG);

    return;
}

sub analyze_result_log
{
    my @internal_logs = glob("$Tools::tmp_dir/*-Master-internal.log");
    (scalar(@internal_logs) == 1) ||
      die("Cannot determine result log name.\n");

    my $log = shift(@internal_logs);
    open(RESULT_LOG, "$log") || die("Could not open result log.\n");
    my $total_ops = 0;
    while(<RESULT_LOG>)
    {
        if (/NumOps=(.*)/)
        {
            $total_ops += $1;
        }
    }

    close(RESULT_LOG);
    unlink($log) || die("Could not remove result log.\n");

    return $total_ops;
}

sub write_experiment_file
{
    my $test_machines = shift(@_);
    my $total_nodes   = shift(@_);
    my $total_testers = shift(@_);
    my $replica_cfg   = shift(@_);
    my $tester_cfg    = shift(@_);

    my $exp_cfg = "$Tools::tmp_dir/stream.$$.exp";
    open(EXP, ">$exp_cfg") ||
      die("Could not create experiment configuration file ($exp_cfg).\n");
    my $old_fd = select(EXP);
    my $fed = 0;
    my $ir0 = 1;
    my $ir1 = 2;
    my $ir2 = 3;
    my $ir3 = 4;
    my $master = 5;
    print <<EOExp;
hosts      @SERVERS
topology   NONE
Federation \${run_dir}/federation.cfg  fed     daemon $fed
RP         \${run_dir}/rp_replica.cfg  static  daemon $fed
Ring0      \${run_dir}/inner_replica.cfg       static  daemon $ir0
Ring1      \${run_dir}/inner_replica.cfg       static  daemon $ir1
Ring2      \${run_dir}/inner_replica.cfg       static  daemon $ir2
Ring3      \${run_dir}/inner_replica.cfg       static  daemon $ir3
Master     $Tools::tmp_dir/stream-master.$$.cfg   static  mortal $master
EOExp

    for (my $i = 0; $i < $test_machines; $i++)
    {
        my $machine_num = $master + 1 + $i;

        my $num_total_stages = int($total_nodes / $test_machines);
        if ($i < ($total_nodes % $test_machines))
        {
            $num_total_stages++;
        }

        my $num_test_stages = int($total_testers / $test_machines);
        if ($i < ($total_testers % $test_machines))
        {
            $num_test_stages++;
        }

        my $num_replica_stages = $num_total_stages - $num_test_stages;

        for (my $j = 0; $j < $num_replica_stages; $j++)
        {
            print("Replica$i-$j $replica_cfg static daemon $machine_num\n");
        }

        for (my $j = 0; $j < $num_test_stages; $j++)
        {
            print("Tester$i-$j  $tester_cfg  static daemon $machine_num\n");
        }
    }
    # Finally, override a few runtime parameters:
    print <<EOOverrides;    
set	Ring*.Inner.skip_archive 		$SKIP_ARCHIVE
set	Ring*.DisseminatorStage.AlarmTimeout 	3600000
set	*.RecoalesceBlkStage.AlarmTimeout 	3600000
set	*.DTree.logger 				\${InternalLog}
set	*.DTree.flood_publish			true
set	*.ReplicaStage.apply_updates 		true
set	*.Router.local_opt_nearest		30
set	*.TClient.all_pairs_pings_file	\${allPairsPingsFile}
set	*.TClient.remember_all_static_nodes	true
EOOverrides
    
    close(EXP);
    select($old_fd);

    return $exp_cfg;
}

sub write_master_config_file
{
    my $test_stages = shift(@_);
    my $test_length = shift(@_);
    my $join_depths = shift(@_);

    my $master_cfg = "$Tools::tmp_dir/stream-master.$$.cfg";
    open(MASTER, ">$master_cfg") ||
      die("Could not create master configuration file ($master_cfg).\n");
    my $old_fd = select(MASTER);
    &write_master_header();
    print("      NumTesters  $test_stages\n");
    print("      TestLength  $test_length\n");
    print("      JoinDepths  $join_depths\n");
    print("      Writer      \${Tester0-0ID}\n");
    print("      UpdateDelay " . int(1000/$UPDATE_RATE) . "\n");
    &write_master_footer();
    close(MASTER);
    select($old_fd);

    return $master_cfg;
}

sub write_replica_config_file
{
    my $replica_cfg = "$Tools::tmp_dir/stream-replica.$$.cfg";
    open(REPLICA, ">$replica_cfg") ||
      die("Could not create replica configuration file ($replica_cfg).\n");

    my $old_fd = select(REPLICA);
    &write_tester_header();
    &write_tester_footer();

    close(REPLICA);
    select($old_fd);

    return $replica_cfg;
}

sub write_tester_config_file
{
    my $dobj_size     = shift(@_);
    my $update_size   = shift(@_);


    my $tester_cfg = "$Tools::tmp_dir/stream-tester.$$.cfg";
    open(TESTER, ">$tester_cfg") ||
      die("Could not create tester configuration file ($tester_cfg).\n");

    my $old_fd = select(TESTER);
    &write_tester_header();

    my $master = "\${MasterID}";
    &write_tester_stage($master, $update_size, $dobj_size);

    &write_tester_footer();
    close(TESTER);
    select($old_fd);

    return $tester_cfg;
}

sub write_master_header
{
    print <<'EOHeader';
<sandstorm>
    <!include Generic.hdr>
	
<stages>
  <StreamMicroBenchmarkMaster>
    class ostore.apps.test.StreamMicroBenchmarkMaster
    queueThreshold 1000
    <initargs>
        ResultLog ${InternalLog}
EOHeader
}

sub write_master_footer
{
    print <<'EOFooter';
    </initargs>
  </StreamMicroBenchmarkMaster>

  #  We need to have Network and Tapestry stages to generate the
  #  proper 'Ready' messages that the run-scripts look for.
	<!include Network.stg>
    
        <!include Router.stg>
	
	<!include StaticTClient.stg>

  # Leave these in so the master node can be a remote replica.
	<!include Replica.stg>

	<!include Update.stg>

	<!include DTreeNode.stg>

	<!include Requestor.stg>

	<!include MemoryCache.stg>

	<!include RecoalesceBlk.stg>
	
</stages>

</sandstorm>
EOFooter
}

sub write_tester_header
{
    print <<'EOHeader';
<sandstorm>
    <!include Generic.hdr>
    <stages>
	<!include Network.stg>
    
        <!include Router_fake_keys.stg>
	
	<!include StaticTClient.stg>
	    
	<!include Client.stg>

	<!include Replica.stg>

	<!include Read.stg>

	<!include RecoalesceBlk.stg>

	<!include Update.stg>

	<!include DTreeNode.stg>

	<!include Requestor.stg>

	<!include MemoryCache.stg>
	
	<!include ArchivalStorage.stg>

EOHeader
}

sub write_tester_stage
{
    my $master      = shift(@_);
    my $update_size = shift(@_);
    my $dobj_size   = shift(@_);

    print <<EOStage;
  <StreamMicroBenchmark>
    class ostore.apps.test.StreamMicroBenchmark
    queueThreshold 1000
    <initargs>
      Master           $master
      InitLength            10
      InitPauseTime         10
      SettleTime          3000
      TestLength           100
      TestPauseTime          0
      ReplaceSize $update_size
      ReplaceGranularity Bytes   # 'Bytes' of 'Blocks'
      DobjSize      $dobj_size
      rp_pkeyfile        \${RPPublicKey}
      principal_pkeyfile \${PrincipalPublicKey}
      principal_skeyfile \${PrincipalPrivateKey}
    </initargs>
  </StreamMicroBenchmark>
EOStage
}

sub write_tester_footer
{
    print <<EOFooter;
</stages>
</sandstorm>
EOFooter
}

# try to kill run-experiment nicely.
sub slow_death
{
    print("*** Caught SIG@_ ***\n");
    if (defined $exp_pid)
    {
	print("  Killing run-experiment...\n");
	kill 2, $exp_pid;
	# Now wait for it to die.
	&returned_zero($exp_pid);
    }
    &exit_all(1);
}
