/**
 * GraphDisplayPanel.java
 *
 * Copyright (c) 2002 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
package Sonar; 
 
import Sonar.util.*; 
import Sonar.Dialog.*;
import Sonar.event.EdgeEvent;
import Sonar.event.NodeEvent;
import Sonar.PacketAnalyzers.GridLocationAnalyzer;

import javax.swing.*; 
import java.beans.*; 
import java.awt.*; 
import java.util.*; 
 
/**
 * This class is essentially a JPanel with the paint() function overridden.
 * It cycles through the nodes, edges, nodePainters, edgePainters, and
 * screenPainters  and tells them all to draw on its graphics object. Then I 
 * add some code for scaling and zooming the screen and converting from node
 * coordinates to screen coordinates 
 *
 * @author:   Kamin Whitehouse <kamin@cs.berkeley.edu> 
 */ 
public class GraphDisplayPanel extends javax.swing.JPanel// implements java.io.Serializable 
{ 
    //the following are member variables needed only for painting 
    protected Image doubleBufferImage; 
    protected Dimension doubleBufferImageSize; 
    protected Graphics doubleBufferGraphic; 
    protected boolean fitToScreenAutomatically = true; 

    //the following are needed for scaling and zooming 
    protected double xScale=-1; 
    protected double xScaleIntercept = -1; 
    protected double yScale = -1; 
    protected double yScaleIntercept=-1; 
    protected double xMargin=10, yMargin=10;
    //value of 10 -> margin width of 10% of the screensize 
    protected boolean screenIsEmpty=true; 

    protected double scale_x = 1;
    protected double scale_y = 1;

    protected double original_width = 0;
    protected double original_height = 0;

    protected double scale_x_org = 1;
    protected double scale_y_org = 1;

    //SANDSTORM additions
    public boolean planetlab = false;
    public DisplayManager displayManager;

    //DisplayManager - assign color
    //GDP - position the node and paint it with the location analyzer
    public void handleNodeEvent (NodeEvent e) {
	switch(e.getEventType()) {
	    case NodeEvent.CREATE:
		displayManager.nodeCreated(e);
		displayManager.locationAnalyzer.nodeCreated(e);
		break;
	    case NodeEvent.DELETE:
		displayManager.nodeDeleted(e);
		displayManager.locationAnalyzer.nodeDeleted(e);
		break;
	    case NodeEvent.ID:
		displayManager.identifyNode(e.getNodeId(), e.getIdType());
	}
    } 

    //DisplayManager - assign color
    //GDP - position the edge and paint it with the location analyzer
    public void handleEdgeEvent (EdgeEvent e) {
	switch(e.getEventType()) {
	    case EdgeEvent.CREATE:
		displayManager.edgeCreated(e);
		displayManager.locationAnalyzer.edgeCreated(e);
		break;
	    case EdgeEvent.DELETE:
		displayManager.edgeDeleted(e);
		displayManager.locationAnalyzer.edgeDeleted(e);
		break;
	    case EdgeEvent.MODIFY_WIDTH:
		displayManager.modifyEdgeWidth(e.getEdgeInfo());
	}
    }
    
    public GraphDisplayPanel(MainFrame mainFrame) {
	//{{INIT_CONTROLS 
	setLayout(null); 
	Insets ins = getInsets(); 
	setSize(ins.left + ins.right + 430,ins.top + ins.bottom + 270); 

	displayManager = new DisplayManager(mainFrame);
	addMouseListener(displayManager); 
	addMouseMotionListener(displayManager); 
	//}} 

	//{{REGISTER_LISTENERS 
	//}} 
    }

    //---------------------------------------------------------------- 
    //PAINT 
    //recall that paint() is called the first time the window 
    //needs to be drawn, or if something damages the window, 
    //and in this case by refreshScreenNow()
    public double scaleX() {
	return scale_x;
    }
    public double scaleY() {
	return scale_y;
    }
    
    public void paint(Graphics g) { 
	super.paint(g);//first paint the panel normally 

	//the following block of code is used 
	//if this is the first time being drawn or if the window was resized 
	//Otherwise we don't creat a new buffer 
	Dimension d = getSize(); 
	
	//HACK
	if(original_height == 0) {
	    original_height = d.height;
	    original_width = d.width;
	    scale_x_org = (double) d.height / (double) 462;
	    scale_y_org = (double) d.width / (double) 701;

	    //TODO fix hack to PacketAnalyzers
	    //MainClass.out("height "+original_height+", scale "+scale_y_org);
	    //MainClass.out("width "+original_width+", scale "+scale_x_org);
	}

	if ((doubleBufferImage == null) || (d.width != doubleBufferImageSize.width) || (d.height != doubleBufferImageSize.height))//if this is the first time being drawn or if the window was resized 
	{ 
	    doubleBufferImage = createImage(d.width, d.height); 
	    doubleBufferImageSize = d; 
	    if (doubleBufferGraphic != null) { 
		doubleBufferGraphic.dispose(); 
	    } 
	    doubleBufferGraphic = doubleBufferImage.getGraphics(); 
	    doubleBufferGraphic.setFont(getFont()); 
	} 
    
    /*
     * Assumption made that we are using the GridLocationAnalyzer 
     */
    if(planetlab && displayManager.locationAnalyzer.getLayoutMode() == GridLocationAnalyzer.MATRIX)
    {
	ImageIcon ih = new ImageIcon("images_os/earth-map.gif",
			"images_os/earth-map.gif");
	doubleBufferGraphic.drawImage(ih.getImage(), 0, 0, d.width, d.height, null); 
	//adjustNodesToZoom();
	
	scale_x = d.width / original_width; //701;
	scale_y = d.height / original_height; //462;
    } else {
    	doubleBufferGraphic.setColor(Color.white); 
	doubleBufferGraphic.fillRect(0, 0, d.width, d.height); 
	if( (fitToScreenAutomatically) || (screenIsEmpty))  {
	    //if the user wants all the nodes on the screen, or if these 
	    //are the first nodes on the screen, fit all nodes to the visible area 
	    fitToScreen(); 
	}

    }

    //TODO replace?
    displayManager.paintOverScreen(doubleBufferGraphic); 

    //draw all the nodes 
    NodeInfo nodeDisplayInfo; 
    int xCoord, yCoord, imageWidth =0, imageHeight; 

    for(Enumeration nodes = displayManager.getNodeInfo();
	    nodes.hasMoreElements();) { 
	nodeDisplayInfo = (NodeInfo)nodes.nextElement(); 

	//figure out where to put the node on the screen 
        if(! (planetlab && displayManager.locationAnalyzer.getLayoutMode() == GridLocationAnalyzer.MATRIX)) {
	    xCoord = scaleNodeXCoordToScreenCoord(displayManager.locationAnalyzer.getX(
			nodeDisplayInfo.getNodeId())); 
	    yCoord = scaleNodeYCoordToScreenCoord(displayManager.locationAnalyzer.getY(
			nodeDisplayInfo.getNodeId())); 
	} else {
	    //scale nodes to zoom, size of screen if not default values
	    xCoord = (int) ( displayManager.locationAnalyzer.getX(nodeDisplayInfo.getNodeId()) * (d.width / original_width) * scale_x_org );
	    yCoord = (int) ( displayManager.locationAnalyzer.getY(nodeDisplayInfo.getNodeId()) * (d.height / original_height) * scale_y_org );     
	}


	//if that spot is not on the visible area, don't draw it at all 
	if((xCoord > this.getSize().getWidth()) || (yCoord > this.getSize().getHeight())) {
	    continue; 
	} 

    
	if(planetlab && displayManager.locationAnalyzer.getLayoutMode() == GridLocationAnalyzer.MATRIX) {
	    xScale = (d.width-2*(d.width/xMargin))/d.width; 
	    yScale = (d.height-2*(d.height/yMargin))/d.height; 
	    xScale = Math.min(xScale, yScale);//this is to make sure that the x and y coordinates are not warped 
	    yScale = xScale; 
	}
	imageWidth = (int)Math.max(8,xScale*nodeDisplayInfo.getImageWidth());
	//note that this requires that we initialized the size of the image to
	//be in node coordinates, but it allows the image to scale
	//automatically with the size of the network 
	imageHeight = (int)Math.max(8,yScale*nodeDisplayInfo.getImageHeight()); 

	//go through the list of things that want to paint this node and paint it 
	displayManager.paintAllNodes(nodeDisplayInfo.getNodeId(), xCoord-imageWidth/2, yCoord-imageHeight/2, xCoord+imageWidth/2, yCoord+imageHeight/2, doubleBufferGraphic); 
    } 

    
    int grid_distance = Math.max(1, Math.abs(scaleNodeXCoordToScreenCoord(displayManager.locationAnalyzer.getCenterX()) - scaleNodeXCoordToScreenCoord(displayManager.locationAnalyzer.getCenterX() + 0.5)));
    //System.err.println("grid dist:"+grid_distance);

    //TODO
    displayManager.locationAnalyzer.paintGrids (scaleNodeXCoordToScreenCoord(displayManager.locationAnalyzer.getCenterX()), scaleNodeYCoordToScreenCoord(displayManager.locationAnalyzer.getCenterY()), grid_distance , doubleBufferGraphic);

    //draw all the edges 
    EdgeInfo edgeDisplayInfo; 
    for(Enumeration edges = displayManager.getEdgeInfo(); edges.hasMoreElements();) { 
	edgeDisplayInfo = (EdgeInfo)edges.nextElement(); 
	//figure out the coordinates of the endpoints of the edge 
	double x1_d = displayManager.locationAnalyzer.getX(edgeDisplayInfo.getSource()); 
	double y1_d = displayManager.locationAnalyzer.getY(edgeDisplayInfo.getSource()); 
	double x2_d = displayManager.locationAnalyzer.getX(edgeDisplayInfo.getDest()); 
	double y2_d = displayManager.locationAnalyzer.getY(edgeDisplayInfo.getDest()); 
	int x1, y1, x2, y2;
	if(planetlab && displayManager.locationAnalyzer.getLayoutMode() == GridLocationAnalyzer.MATRIX) {
	    x1 = (int) (x1_d * d.width / 701);
	    x2 = (int) (x2_d * d.width / 701);
	    y1 = (int) (y1_d * d.height / 462);
	    y2 = (int) (y2_d * d.height / 462);
	} else {
	    x1 = scaleNodeXCoordToScreenCoord(x1_d);
	    y1 = scaleNodeYCoordToScreenCoord(y1_d);
	    x2 = scaleNodeXCoordToScreenCoord(x2_d);
	    y2 = scaleNodeYCoordToScreenCoord(y2_d);
	}

	//edgeDisplayInfo.paint(doubleBufferGraphic); 
	displayManager.paintAllEdges(edgeDisplayInfo.getSource(), edgeDisplayInfo.getDest(), x1, y1, x2, y2, doubleBufferGraphic); 
    }

    //draw all the typed edges 
    Hashtable typedEdges;
    Enumeration edgeTables = displayManager.getTypedEdgeInfo();
    while(edgeTables != null && edgeTables.hasMoreElements()) {
	typedEdges = (Hashtable) edgeTables.nextElement(); 
	if(typedEdges != null) {
	    for(Enumeration edgeElms = typedEdges.elements(); edgeElms.hasMoreElements();) {
		edgeDisplayInfo = (EdgeInfo) edgeElms.nextElement();
		//figure out the coordinates of the endpoints of the edge 
		double x1_d = displayManager.locationAnalyzer.getX(edgeDisplayInfo.getSource()); 
		double y1_d = displayManager.locationAnalyzer.getY(edgeDisplayInfo.getSource()); 
		double x2_d = displayManager.locationAnalyzer.getX(edgeDisplayInfo.getDest()); 
		double y2_d = displayManager.locationAnalyzer.getY(edgeDisplayInfo.getDest()); 
		int x1, y1, x2, y2;
	
		if(planetlab && displayManager.locationAnalyzer.getLayoutMode() == GridLocationAnalyzer.MATRIX) {
		    x1 = (int) (x1_d * d.width / 701);
		    x2 = (int) (x2_d * d.width / 701);
		    y1 = (int) (y1_d * d.height / 462);
		    y2 = (int) (y2_d * d.height / 462);
		} else {
		    x1 = scaleNodeXCoordToScreenCoord(x1_d);
		    y1 = scaleNodeYCoordToScreenCoord(y1_d);
		    x2 = scaleNodeXCoordToScreenCoord(x2_d);
		    y2 = scaleNodeYCoordToScreenCoord(y2_d);
		}

		//edgeDisplayInfo.paint(doubleBufferGraphic); 
		displayManager.paintAllTypedEdges(edgeDisplayInfo.getSource(), edgeDisplayInfo.getDest(), x1, y1, x2, y2, doubleBufferGraphic); 
	    }
	}
    }

    //draw things over the entire display 
    displayManager.paintOverScreen(doubleBufferGraphic); 

    //Make everything that was drawn visible 
    g.drawImage(doubleBufferImage, 0, 0, null); 

    } 
    //END OF PAINT 
    //---------------------------------------------------------------- 

    public void update(Graphics g) { 
	paint(g); 
    } 

    public void repaint(Graphics g) { 
	paint(g); 
    } 

               
  //******************************************************* 
  //******************************************************* 
  //the following code sets X and Y scaling factors for  
  //the X and Y coordinates of the nodes  
  //******************************************************* 
  //the norm is to set scaling factors to keep everything 
  //on the screen.  However, if the user attempts to zoom 
  //the scaling factors can be larger and will be held  
  //constant 
  //******************************************************* 
  //******************************************************* 
 
               
               
  //---------------------------------------------------------------- 
  //FIT TO SCREEN 
  //this function will set the scaling factors above such 
  //that all nodes are scaled to within the screen viewing area 
  public void fitToScreen() //do not synchronize 
  { 
    double largestXCoord = Double.MIN_VALUE; 
    double smallestXCoord = Double.MAX_VALUE; 
    double largestYCoord = Double.MIN_VALUE; 
    double smallestYCoord = Double.MAX_VALUE; 
    double x, y; 
    //find the largest and smallest coords for all nodes 
    NodeInfo currentDisplayInfo; 
    for(Enumeration nodes = displayManager.getNodeInfo(); nodes.hasMoreElements();) {
      currentDisplayInfo = (NodeInfo)nodes.nextElement(); 
      if(screenIsEmpty){currentDisplayInfo.setFitOnScreen(true);} 
      if( ((currentDisplayInfo.getDisplayThisNode() == true) && (currentDisplayInfo.getFitOnScreen() == true) ))//If the current node is displayed and old enough, or if they are the first nodes to be drawn 
	{ 
	  x = displayManager.locationAnalyzer.getX(currentDisplayInfo.getNodeId()); 
	  y = displayManager.locationAnalyzer.getY(currentDisplayInfo.getNodeId()); 
	  if(x > largestXCoord) {largestXCoord = x;} 
	  if(x < smallestXCoord){smallestXCoord = x;} 
	  if(y > largestYCoord){largestYCoord = y;} 
	  if(y < smallestYCoord){smallestYCoord = y;} 
	} 
    }	 
    //here we use the following equations to set the scaling factors: 
    // xScale*SmallestXCoord + XIntercept = 0 
    // xScale*LargestXCoord  + XIntercept = window.width(); 
    // 
    //And the same for the y scaling factors.   
    //Note that I want a border of <Margin>% of the screen on both sides 
     
    Dimension d = getSize(); 
     
    xScale = (d.width-2*(d.width/xMargin))/(largestXCoord - smallestXCoord); 
    yScale = (d.height-2*(d.height/yMargin))/(largestYCoord - smallestYCoord); 
 	
    xScale = Math.min(xScale, yScale);//this is to make sure that the x and y coordinates are not warped 
    yScale = xScale; 
 
    xScaleIntercept = -xScale*smallestXCoord + d.width/xMargin; 
    yScaleIntercept = -yScale*smallestYCoord + d.height/yMargin; 
    screenIsEmpty = false; 
    if(displayManager.proprietaryNodeInfo.isEmpty())//if there are no nodes and none of this function was executed 
      { 
	screenIsEmpty = true; 
	xScale = 1; 
	yScale = 1; 
	xScaleIntercept = 0; 
	yScaleIntercept = 0; 
      }
  }
  
  public int scaleNodeXCoordToScreenCoord(double pXCoord)//do not synchronize 
  {         //take the local coordinate system of the nodes and show it as a graphical coordinate system 
    Double xCoord = new Double(xScale*pXCoord+xScaleIntercept); 
    return xCoord.intValue(); 
  } 
	 
  public int scaleNodeYCoordToScreenCoord(double pYCoord)       //do not synchronize 
  {         //take the local coordinate system of the nodes and show it as a graphical coordinate system 
    Double yCoord = new Double(yScale*pYCoord+yScaleIntercept); 
    return yCoord.intValue(); 
  } 
 
  public Double scaleScreenXCoordToNodeCoord(double pXCoord)//do not synchronize 
  {         //take a graphical coordinate system and show it as the local coordinate system of the nodes  
    return new Double((pXCoord-xScaleIntercept)/xScale); 
  } 
	 
  public Double scaleScreenYCoordToNodeCoord(double pYCoord)       //do not synchronize 
  {         //take a graphical coordinate system and show it as the local coordinate system of the nodes  
    return new Double((pYCoord-yScaleIntercept)/yScale); 
  } 
	 
  //------------------------------------------------------------------ 
  //GET/SET 
  public boolean getFitToScreenAutomatically(){return fitToScreenAutomatically;} 
  public void setFitToScreenAutomatically(boolean p){fitToScreenAutomatically=p;} 
  public double getXScale(){return xScale;} 
  public double getYScale(){return yScale;} 
  //GET/SET 
  //------------------------------------------------------------------ 
  
  //{{DECLARE_CONTROLS 
  //}} 
 
}
